/* Copyright (C) 2000, 2012  The PARI group.

This file is part of the PARI/GP package.

PARI/GP is free software; you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free Software
Foundation. It is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY WHATSOEVER.

Check the License for details. You should have received a copy of it, along
with the package; see the file 'COPYING'. If not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */

/*
*  2018.2.23-2018.2.25 : pclgp() by Takashi Fukuda
*  2018.3.9-2018.3.10 : long p  --> t_INT p
*  2018.9.10 : abellambda(), abelstkpol(), quadlambda0(),quadstkpol().
*  2018.10.11 : rcncubcyclo()
*
*  GEN pclgp(GEN p, long f, GEN H, long flag)
*  tries to compute p-part A_F of the ideal class group of
*  an abelin number field F by handling all cyclic subfields K of F.
*
*  p : odd prime number not dividing [F:Q], t_INT or t_VEC
*  f : F subset Q(zeta_f), preferably equal to cond(F)
*  H : t_VECSMALL listing generators of sugbr H of G=(Z/fZ)^*
*      corresponding to F
*  flag controls behavior of pclgp().
*  (flag&1)!=0 ==> output plus part A_F^+
*  (flag&2)!=0 ==> output minus part A_F^-
*  (flag&4)!=0 ==> ignore proper subfields of F (for Coates conj).
*  (flag&8)!=0 ==> try to apply algorithm of Aoki-Fukuda.
*                  results are rigorous without GRH.
*  (flag&16)!=0 ==> output K which has non-trivial A_K
*  pclgp() return t_VEC z having 6 component.
*  z[1] : p
*  z[2] : A_F^+
*  z[3] : A_F^-
*  z[4] : G(F/Q)
*  z[5] : number of cyclic subfields K of F except for Q
*  z[6] : number of all representatives of inj. char. of G(K/Q)
*
*  z[2]=[3, [1, 1, 1]] ==> A_F^+ = (p,p,p)
*  z[2]=[3, []] ==> |A_F^+| = p^3, gr. str. is unknown
*  data of z[2] is always non-rigorous, but considered almost true.
*  If you run with flag=(1|8), then you get rigorous order and str. of A_F^+.
*  But it will take very long time.
*
*  z[3]=[3, [1, 1, 1]] ==> A_F^- = (p,p,p)
*  z[3]=[3, []] ==> |A_F^-| = p^3, gr. str. is unknown
*  data of z[3] is always rigorous without GRH.
*  If str. was unknown, you can run with flag=(2|8).
*  Then you get str. of A_F^-.
*
*  If (flag&1)==0, then z[2]=[]
*  If (flag&2)==0, then z[3]=[]
*  Default flag=3.
*
*****************************************************************************
*
*  GEN abellambda(long p, long f, GEN H)
*  computes Iwasawa lambda_p_minus invariant of abelian field F
*  corresponding to subgr H of (Z/fZ)^*
*  p is an odd prime number not dividing [F:Q]
*
*  GEN abelstkpol(long p, long f, GEN H, long n)
*  computes minus part of Iwasawa polynomial of abelian field F
*  corresponding to subgr H of (Z/fZ)^* using Stickelberger ele. xi_n.
*  p is an odd prime number not dividing [F:Q]
*
*  GEN quadlambda0(long p, long m)
*  computes Iwasawa lambda_p invariant of imag. quad. field F=Q(sqrt(m))
*  p is any prime number.
*  m is a negative square-free integer.
*  For example, quadlambda0(3,-956238) returns [14,[1,3,9,23]].
*  This means lambda_3(Q(sqrt(-956238)))=14, |A_0|=3, |A_1|=3^3,
*  |A_2|=3^9, |A_3|=3^23 and |A_{n+1}|/|A_n|=3^14 for all n>=2.
*
*  GEN quadstkpol0(long p, long m, long n, long flag)
*  computes Iwasawa polynomial of imag. quad. field F=Q(sqrt(m)) using xi_n.
*  p is any prime number.
*  m is a negative square-free integer.
*  n is an integer >= 1.
*  (flag&1)!=0 ==> output (1/2)*xi_n
*  (flag&2)!=0 ==> output power series
*  (flag&4)!=0 ==> output Iwasawa polynomial
*  Default flag=4.
*  xi_n is sometimes useful. If p does not split in F/Q and A_0 is cyclic
*  G(F/Q)-module, then A_n=Z[G(F_n/F_0)]/(xi_n).
*  quadstkpol0(3,-956238,3,1) enables us to see A_3=(Z/9Z)^9*(Z/3Z)^5.
*
*  Computing times of the above 4 ft. are proportional to f*p^(n+1) (n>=1).
*  So it is reasonable to assume f*p^2<2^32.
*  Simple codings were chosen assuming f<2^32 and p^2<2^32.
*****************************************************************************
*
*  GEN rel_class_num(long f, GEN H, long flag, long prec)
*  computes relative class number(i.e. 1st factor) of abelian number field F
*  corresponding to subgr H of (Z/fZ)^*
*  flag controls algorithms.
*  B_{1,chi}=(1/f)*sum_{1<= a <= f} a*chi(a)            (1)
*           =-(1/(2-psi(2)))*sum_{1<= a <= f/2} chi(a)  (2)
*            psi = complex conjugation of chi.
*  flag==1 ==> approximate calculation of (1)
*  flag==2 ==> approximate calculation of (2)
*  flag==3 ==> integral calculation of norm of (1) by resultant
*  flag==4 ==> integral calculation of norm of (2) by resultant
*  flag==5 ==> compute norm of (2) directly
*  flag==6 ==> compute norm of (2) using cyclic decomposition
*
*  prec is the decimal precision for flag==1,2.
*  If prec==0, then an appropriate precicion is settled automatically.
*  Default flag=6.
*
*  If F is not a full cyclotomic field, the unit index Q is difficult
*  to determine. So we assume Q=2 in many cases.
*/

#include "pari.h"
#include "paripriv.h"

/*
#define DEBUGLEVEL 6
#define VECSMALL_CHI
#define USE_LLL_BASIS
#define USE_GALOISSUBCYCLO
*/

#define FpX_translate1(xi,p0) FpX_translate(xi,gen_1,p0)
#define onevec(x) const_vec(x,gen_1)
#define cyc_mul(x, y, n) ZX_mod_Xnm1(ZX_mul(x, y), n)
#define nullvec() cgetg(1, t_VEC)

#define PLUS_PART       (1L<<0)
#define MINUS_PART      (1L<<1)
#define SKIP_PROPER     (1L<<2)
#define USE_MLL         (1L<<3)
#define VERBOSE_BIT     (1L<<4)
#define USE_BASIS       (1L<<5)
#define USE_FACTOR      (1L<<6)
#define USE_GALOIS_POL  (1L<<7)
#define USE_F           (1L<<8)
#define SAVE_MEMORY     (1L<<9)
#define DEBUG_BIT       (1L<<10)
#define DEBUG1          (1L<<11)
#define DEBUG2          (1L<<12)
#define DEBUG3          (1L<<13)
#define DEBUG4          (1L<<14)

/* R : sorted t_VEC of t_REAL
*  x : t_REAL
*  return R[i] nearest to x */
static GEN
nearr(GEN R, GEN x)
{
  long n = lg(R)-1, low = 0, high = n-1, mid, min = -1;
  GEN min_diff = itor(gen_1, lg(gel(R, 1))), diff, adiff;
  while(low<=high)
  {
   mid = (low+high)>>1;
   if(equalrr(x, gel(R, 1+mid))) return gel(R, 1+mid);
   diff = subrr(x, gel(R, 1+mid));
   adiff = absr(diff);
   if(cmprr(adiff, min_diff)<0) {min = mid;min_diff = adiff;}
   if(cmprs(diff, 0)<0) high = mid-1;
   else low = mid+1;
  }
  if(min<0) pari_err(e_MISC,"[nearr");
  return gel(R, 1+min);
}


/* return NULL if A mod el does not exist */
static GEN
RgM_to_Flm_safe(GEN A, ulong el)
{
  pari_sp av = avma;
  long i, j, col = lg(A), row = lg(gel(A, 1));
  GEN a = cgetg(col, t_MAT);
  for(j=1;j<col;j++)
  {
    GEN x = cgetg(row, t_VECSMALL), y = gel(A, j);
    gel(a, j) = x;
    for(i=1;i<row;i++)
    {
      GEN z = gel(y, i);ulong z1, z2;
      switch(typ(z))
      {
        case t_INT:
          x[i] = umodiu(z, el);
          break;
        case t_FRAC:
          z1 = umodiu(gel(z, 1), el);
          if(z1==0) {x[i] = 0;break;}
          z2 = umodiu(gel(z, 2), el);
          if(z2) {x[i] = Fl_div(z1, z2, el);break;}
          avma = av;return NULL;
        default:
          avma = av;return NULL;
      }
    }
  }
  return a;
}

/* return NULL if x mod el does not exist */
static GEN
RgX_to_Flx_safe(GEN x, ulong el)
{
  pari_sp av = avma;
  long i, lx = lg(x);
  GEN a = cgetg(lx, t_VECSMALL);
  a[1]=((ulong)x[1])&VARNBITS;
  for (i=2; i<lx; i++)
  {
     GEN z = gel(x, i);ulong z1, z2;
     switch(typ(z))
     {
       case t_INT:
         a[i] = umodiu(z, el);
         break;
       case t_FRAC:
         z1 = umodiu(gel(z, 1), el);
         if(z1==0) {a[i] = 0;break;}
         z2 = umodiu(gel(z, 2), el);
         if(z2) {a[i] = Fl_div(z1, z2, el);break;}
         avma = av;return NULL;
       default:
         avma = av;return NULL;
     }
  }
  return Flx_renormalize(a, lx);
}

static GEN
ary2vec(long *C, long n)
{
  long i;GEN z;
  z = cgetg(n+1, t_VEC);
  for(i=0;i<n;i++) z[i+1] = (long)stoi(C[i]);
  return z;
}

static GEN
ary2vec2(long *C, long n)
{
  GEN z;
  z = cgetg(3, t_VEC);
  gel(z, 1) = utoi(C[n]-C[n-1]);
  gel(z, 2) = ary2vec(C, n+1);
  return z;
}

static GEN
zv_to_ZX(GEN x, long sv)
{
  long i, l=lg(x);GEN z;
  while(--l>=1 && x[l]==0);
  if(l==0) return pol_0(sv);
  l+=2;
  z = cgetg(l, t_POL);z[1]=evalsigne(1) | evalvarn(sv);
  x--;
  for(i=2;i<l;i++) gel(z, i) = stoi(x[i]);
  return z;
}

/*
* create d*d cyclic matrix
* p is t_POL of deg d-1
* p[lg(p)-1]=p[d+1] may be 0
*/
static GEN
pol2mat(GEN p, long d)
{
  long i, j;GEN z;
  z = cgetg(d+1, t_MAT);
  for(i=0;i<d;i++)
  {
    gel(z, i+1) = cgetg(d+1, t_COL);
    for(j=0;j<d;j++) gmael(z, i+1, j+1) = gel(p, 2+(i+j)%d);
  }
  return z;
}

static GEN
#ifdef VECSMALL_CHI
set_A(GEN B, GEN chi)
#else
set_A(GEN B, int *chi)
#endif
{
  long a, i, j, B0, lA = lg(B)-1;GEN A;
#ifdef VECSMALL_CHI
  chi++;
#endif
  A = cgetg(lA+1, t_VECSMALL);
  for(a=0, j=1, B0=B[1];j<B0;j++) a+=chi[j];
  A[1] = a;
  for(i=1;i<lA;i++)
  {
    long Bi = B[i+1];
    for(a=A[i], j=B[i];j<Bi;j++) a+=chi[j];
    A[i+1] = a;
  }
  return A;
}

/*
* k=Q(sqrt(m)), A_n=p-class gr. of k_n, |A_n|=p^(e_n)
* return e_n-e_(n-1)
* essential assumption : m is not divisible by p
* Gold, Acta Arith. XXVI (1974), p.22 formula (3)
*/
static long
#ifdef VECSMALL_CHI
ediff(ulong p, long m, ulong n, GEN chi)
#else
ediff(ulong p, long m, ulong n, int *chi)
#endif
{
  long i, j,  s, d, y, g, lx, *px;
  ulong p1, pn, pn1, pn_1, phipn, phipn1;
  GEN A, B, x, gs, cs;
  pari_sp av = avma;

  d=((m-1)%4==0)?labs(m):4*labs(m);
  p1=p-1;pn_1=upowuu(p, n-1);pn=p*pn_1;pn1=p*pn;phipn=p1*pn_1;phipn1=p1*pn;
  lx=2*p1*phipn;
  y=Fl_inv(pn1%d, d);g=pgener_Zl(p);  /* pn1 may > d */
  /* printf("(d,y,g)=(%ld,%ld,%ld)\n", d,  y, g); */
  cs = cgetg(2+phipn, t_POL);cs[1]=evalsigne(1) | evalvarn(0);
  x = cgetg(1+lx, t_VECSMALL);
  gs = Fl_powers(g, phipn1-1, pn1); /* gs[1+i]=g^i(mod p^(n+1)), 0<=i<p^(n+1) */

  for(px=x,i=0;i<p1;i++)
  {
   long ipn=i*pn+1,ipnpn=ipn+phipn;
   for(s=0;s<phipn;s++)
   {
     *++px = (y*gs[s+ipn])%d;  /* gs[s+ipn] may > d */
     *++px = (y*gs[(s%pn_1)+ipnpn])%d;
   }
  }
  B = vecsmall_uniq(x);
  A = set_A(B, chi);
  for(s=0;s<phipn;s++)
  {
    long a=0, ipn=1, spn1=s%pn_1;
    for(i=0;i<p1;i++)
    {
     if((j=zv_search(B, (y*gs[s+ipn])%d))<=0)
        pari_err_BUG("zv_search failed\n");
     a+=A[j];
     if((j=zv_search(B, (y*gs[spn1+ipn+phipn])%d))<=0)
        pari_err_BUG("zv_search failed\n");
     a-=A[j];
     ipn+=pn;
    }
    gel(cs, 2+s) = stoi(a);
  }
  cs = ZX_renormalize(cs, 2+phipn);
  y = (lg(cs)==3) ? phipn*Z_lval(gel(cs, 2), p)
                  : Z_lval(resultant(cs, polcyclo(pn, 0)), p);
  avma=av;
  return y;
}

/*
* g_n(a)=g_n(b) <==> a^2=b^2 mod 2^(n+2) <==> a=b,-b mod 2^(n+2)
* g_n(a)=g_n(1+q0)^k <==> a=x(1+q0)^k x=1,-1
* gam[1+a]=k, k<0  ==> g_n(a)=0
*             k>=0 ==> g_n(a)^(-1)=gamma^k, gamma=g_n(1+q0)
*/
static GEN
set_gam2(long q01, long p, long n)
{
  long i, x, x1, pn, pn2;GEN gam;
  pn = (1L<<n);
  pn2 = (pn<<2);
  gam = const_vecsmall(pn2, -1);
  x=Fl_inv(q01, pn2);x1=1;
  for(i=0;i<pn;i++)
  {
    gam[1+x1] = gam[1+Fl_neg(x1, pn2)] = i;
    x1 = Fl_mul(x1, x, pn2);
  }
  return gam;
}

/*
* g_n(a)=g_n(b) <==> a^(p-1)=b^(p-1) mod p^(n+1) <==> a=xb x=<g^(p^n)>
* g_n(a)=g_n(1+q0)^k <==> a=x(1+q0)^k x=<g^(p^n)>
* gam[1+a]=k, k<0  ==> g_n(a)=0
*             k>=0 ==> g_n(a)^(-1)=gamma^k, gamma=g_n(1+q0)
*/
static GEN
set_gam(long q01, long p, long n)
{
  long i, j, g, g1, x, x1, p1, pn, pn1;GEN A, gam;
  p1 = p-1;pn = upowuu(p, n);pn1 = p*pn;
  gam = const_vecsmall(pn1, -1);
  g = pgener_Zl(p);g1 = Fl_powu(g, pn, pn1);
  A = Fl_powers(g1, p1-1, pn1);  /* A[i]=g^(ip^n) mod p^(n+1) */
  x = Fl_inv(q01, pn1);x1 = 1;
  for(i=0;i<pn;i++)
  {
    for(j=1;j<=p1;j++) gam[1+Fl_mul(x1, A[j], pn1)] = i;
    x1 = Fl_mul(x1, x, pn1);
  }
  return gam;
}

static GEN
#ifdef VECSMALL_CHI
quadstk(long p, long m, long n, GEN chi)
#else
quadstk(long p, long m, long n, int *chi)
#endif
{
  long i, j, j0, f, f0, pn, pn1, pn1f, q0, *C1;GEN C, z, gam;
#ifdef VECSMALL_CHI
  chi++;
#endif
  f = ((m-1)%4==0)?labs(m):4*labs(m);
  pn = upowuu(p, n);pn1 = p*pn;pn1f = pn1%f;
  q0 = (f%p)?f*p:f;
  f0 = (f%p)?f:f/p;
  C = const_vecsmall(pn, 0);C1=C+1;
  gam = set_gam((1+q0)%pn1, p, n);
  for(j=1;j<pn1;j++)
  {
   long ipn1;
   if(j%p==0) continue;
   j0 = gam[1+j];ipn1 = j%f;
   /*
   for(i=1;i<f0;i++) C1[j0]+=i*chi[(i*pn1+j)%f];
   */
   for(i=1;i<f0;i++)
   {
#ifdef VECSMALL_CHI
    long sgn;
#else
    int sgn;
#endif
    if((ipn1+=pn1f)>=f) ipn1-=f;
    if((sgn = chi[ipn1])>0) C1[j0]+=i;
    else if(sgn<0) C1[j0]-=i;
   }
  }
  if(f0!=4) f0<<=1;
  for(i=0;i<pn;i++)
  {
   if(C1[i]%f0) pari_err_BUG("stickel. ele. is not integral.\n");
   else C1[i]/=f0;
  }
  z = zv_to_ZX(C, 0);
  setvarn(z, fetch_user_var("s"));
  return z;
}

static GEN
#ifdef VECSMALL_CHI
quadstk2(long m, long n, GEN chi)
#else
quadstk2(long m, long n, int *chi)
#endif
{
  long i, j, j0, f, f0, pn, pn1, pn2, pn2f, q0, *C1;GEN C, z, gam;
#ifdef VECSMALL_CHI
  chi++;
#endif
  f = ((m-1)%4==0)?labs(m):4*labs(m);
  pn = 1L<<n;pn1 = pn<<1;pn2 = pn1<<1;pn2f = pn2%f;
  q0 = (f&1)?f*4:f;
  f0 = (f&1)?f:f/4;
  C = const_vecsmall(pn, 0);C1=C+1;
  gam = set_gam2((1+q0)%pn2, 2, n);
  for(j=1;j<pn2;j++)
  {
   long ipn2;
   if (!(j&1)) continue;
   j0 = gam[1+j];
   ipn2 = j%f;
   /*
   for(i=1;i<f0;i++) C1[j0]+=i*chi[(i*pn2+j)%f];
   */
   for(i=1;i<f0;i++)
   {
#ifdef VECSMALL_CHI
    long sgn;
#else
    int sgn;
#endif
    if((ipn2+=pn2f)>=f) ipn2-=f;
    if((sgn=chi[ipn2])>0) C1[j0]+=i;
    else if(sgn<0) C1[j0]-=i;
   }
  }
  for(f0<<=1, i=0;i<pn;i++)
  {
   if(C1[i]%f0) pari_err_BUG("stickel. ele. is not integral.\n");
   else C1[i]/=f0;
  }
  z = zv_to_ZX(C, 0);
  setvarn(z, fetch_user_var("s"));
  return z;
}

#ifdef DEBUG
/*
* slow
*/
static int*
set_quad_chi_1(long m)
{
  long a, d, f;int *chi;
  d=((m-1)%4==0)?m:4*m;f=labs(d);
  chi= (int*)stack_calloc(sizeof(int)*f);
  for(a=1;a<f;a++) chi[a]=kross(d, a);
  return chi;
}
#endif

/*
* chi[a]=kross(d, a)   0<=a<=f-1
* d=discriminant of Q(sqrt(m)), f=abs(d)
*
* Algorithm: m=-p1*p2*...*pr ==> kross(d,gi)=-1 (1<=i<=r), gi=proot(pi)
* set_quad_chi_1(m)=set_quad_chi_2(m) for all square-free m s.t. |m|<10^5.
*/
#ifdef VECSMALL_CHI
static GEN
#else
static int*
#endif
set_quad_chi_2(long m)
{
  pari_sp av;
  long i, j, a, d, f, r, g, gn, np, nm, pq[2][30];
  GEN fa, P, E;

  d = ((m-1)%4==0)?m:4*m;f = labs(d);
#ifdef VECSMALL_CHI
  GEN chi0 = const_vecsmall(f, 0), chi=chi0+1;
  av = avma;
  GEN plus0 = const_vecsmall(f, 0), plus=plus0+1, p0 = plus;
  GEN minus0 = const_vecsmall(f, 0), minus=minus0+1, p1 = minus;
#else
  int *chi0 = (int*)stack_calloc(sizeof(int)*f), *chi=chi0;
  av = avma;
  int *plus0 = (int*)stack_calloc(sizeof(int)*f), *plus=plus0, *p0=plus;
  int *minus0 = (int*)stack_calloc(sizeof(int)*f), *minus=minus0, *p1=minus;
#endif
  fa = factoru(f), P = gel(fa, 1), E = gel(fa, 2);r = lg(P)-1;
  for(i=0;i<r;i++)
  {
    long p, q, x, y;
    p=upowuu(P[1+i], E[1+i]);q=f/p;
    cbezout(p, q, &x, &y);
    /* printf("%d*(%d)+%d*(%d)\n", p, x, q, y); */
    pq[0][i] = smodss(p*x, f);
    pq[1][i] = smodss(q*y, f);
  }
  if(E[1]==2)       /* f=4*(-m) */
  {
    a=(1*pq[1][0]+pq[0][0])%f;
    *p0++=a;
    a=(3*pq[1][0]+pq[0][0])%f;
    *p1++=a;
    i=1;
  }
  else if(E[1]==3)  /* f=8*(-m) */
  {
    a = (1*pq[1][0]+pq[0][0])%f;
    *p0++ = a;
    a = (3*pq[1][0]+pq[0][0])%f;
    if(kross(d, a)>0) *p0++ = a;
    else *p1++ = a;
    a = (5*pq[1][0]+pq[0][0])%f;
    if(kross(d, a)>0) *p0++ = a;
    else *p1++ = a;
    a = (7*pq[1][0]+pq[0][0])%f;
    if(kross(d, a)>0) *p0++ = a;
    else *p1++ = a;
    i = 1;
  }
  else              /* f=-m */
  {*p0++ = 1;i=0;}
  for(;i<r;i++)
  {
     g = pgener_Fl(P[1+i]);
     g = (g*pq[1][i]+pq[0][i])%f;
     gn = g;
     np = p0-plus;
     nm = p1-minus;
CONT:for(j=0;j<np;j++) *p1++ = (plus[j]*gn)%f;
     for(j=0;j<nm;j++) *p0++ = (minus[j]*gn)%f;
     gn = (gn*g)%f;
     if(gn==1) continue;
     for(j=0;j<np;j++) *p0++ = (plus[j]*gn)%f;
     for(j=0;j<nm;j++) *p1++ = (minus[j]*gn)%f;
     gn = (gn*g)%f;
     if(gn!=1) goto CONT;
  }
  np = p0-plus;
  nm = p1-minus;
  for(i=0;i<np;i++) chi[plus[i]] = 1;
  for(i=0;i<nm;i++) chi[minus[i]] = -1;
  avma = av;
  return chi0;
}

/* return the minimun n>=1 s.t. x^n=1 (mod f), x < f */
static ulong
order_f_x(ulong f, ulong x)
{
  ulong n = 1, y;
  if ((x%=f)==0) return 1;
  for (y=x; y!=1; y=Fl_mul(y, x, f)) n++;
  return n;
}

static GEN
FpX_rescale_m1(GEN P, GEN p)
{
  long i, l = lg(P);
  GEN Q = cgetg(l, t_POL);
  for(i=l-1;i>=2;i--)
  {
    gel(Q, i) = gel(P, i);
    if (--i==1) break;
    gel(Q, i) = Fp_neg(gel(P, i), p);
  }
  Q[1] = P[1];return Q;
}

static long
FpX_srh_fac(GEN vn, GEN x, GEN pols, GEN el0)
{
  pari_sp av = avma;
  long i, l=lg(pols);
  for(i=1;i<l;i++)
  {
    if (vn[i]==0) continue;
    if (lgpol(FpX_rem(x, gel(pols, i), el0))==0) {avma = av;vn[i] = 0;return i;}
  }
  pari_err_BUG("FpX_factor_polcyclo");
  return 0;
}

static GEN
FpX_factor_polcyclo(long n, GEN el)
{
  pari_sp av = avma;
  long i, j, n_p=0, m=1, pp[100];
  GEN fa = factoru(n), P = gel(fa, 1), E = gel(fa, 2);long l = lg(P);
  GEN pol, pol0;

  for(i=1;i<l;i++) for(j=E[i];j;j--) pp[n_p++] = P[i];
  pol = pol_x(0);gel(pol, 2) = subiu(el, 1);  /* pol = x-1 */
  pol0=const_col(1, pol);
  while(--n_p>=0)
  {
    long i, n_f, p = pp[n_p], mp = m*p, elm=umodiu(el, m), elmp=umodiu(el, mp);
    long flag = eulerphiu(mp)/order_f_x(mp,elmp)-eulerphiu(m)/order_f_x(m,elm);
    GEN vn_g, pol1=const_vec(0, NULL);
    n_f = lg(pol0)-1;
    vn_g = identity_perm(n_f);
    for(i=1;i<=n_f;i++)
    {
      GEN z0, zp;
      z0 = gel(pol0, i);
      if ((m&1) && p==2)
        zp = FpX_rescale_m1(z0, el);
      else
      {
        zp = RgX_inflate(z0, p);
        if (m%p)
          zp = FpX_div(zp, gel(pol0, FpX_srh_fac(vn_g, zp, pol0, el)), el);
      }
      pol1 = (flag>0)? shallowconcat(pol1, gel(FpX_factor(zp, el), 1))
                     : vec_append(pol1, zp);
    }
    m=mp;pol0=pol1;
  }
  return gerepilecopy(av, pol0);
}

static GEN
Flx_rescale_m1(GEN P, ulong p)
{
  long i, l = lg(P);
  GEN Q = cgetg(l, t_VECSMALL);
  for(i=l-1;i>=2;i--)
  {
    Q[i] = P[i];
    if (--i==1) break;
    Q[i] = Fl_neg(P[i], p);
  }
  Q[1] = P[1];return Q;
}

static long
Flx_srh_fac(GEN vn, GEN x, GEN pols, long el)
{
  pari_sp av = avma;
  long i, l=lg(pols);
  for(i=1;i<l;i++)
  {
    if (vn[i]==0) continue;
    if (lgpol(Flx_rem(x, gel(pols, i), el))==0) {avma = av;vn[i] = 0;return i;}
  }
  pari_err_BUG("Flx_factor_polcyclo");
  return 0;
}

/*
*  factorize polcyclo(n,0) mod el inductively assuming el is prime to n.
*  faster than Flx_factor(ZX_to_Flx(polcyclo(n, 0), el), el)
*  when n is large and polcyclo(n, 0) has many irred. factors.
*/
static GEN
Flx_factor_polcyclo(long n, long el)
{
  pari_sp av = avma;
  long i, j, n_p=0, m=1, pp[100];
  GEN fa = factoru(n), P = gel(fa, 1), E = gel(fa, 2);long l = lg(P);
  GEN pol, pol0;
  /* pari_printf("(n,el)=(%ld,%ld)\n", n, el); */

  for(i=1;i<l;i++) for(j=E[i];j;j--) pp[n_p++] = P[i];
  pol = const_vecsmall(3, 0);
  pol[1] = evalsigne(1) | evalvarn(0);
  pol[2] = el-1;pol[3] = 1;  /* pol = x-1 */
  pol0 = const_vec(1, pol);
  while(--n_p>=0)
  {
    long i, n_f, p = pp[n_p], mp = m*p;
    long flag = eulerphiu(mp)/order_f_x(mp, el)-eulerphiu(m)/order_f_x(m, el);
    GEN vn_g, pol1=const_col(0, NULL);
    n_f = lg(pol0)-1;
    vn_g = identity_perm(n_f);
    for(i=1;i<=n_f;i++)
    {
      GEN z0, zp;
      z0 = gel(pol0, i);
      if ((m&1) && p==2)
        zp = Flx_rescale_m1(z0, el);
      else
      {
        zp = Flx_inflate(z0, p);
        if (m%p)
          zp = Flx_div(zp, gel(pol0, Flx_srh_fac(vn_g, zp, pol0, el)), el);
      }
      /*
      pari_printf("%ld/%ld m=%ld mp=%ld el=%ld flag=%ld\n",i,n_f,m,mp,el,flag);
      */
      pol1 = (flag>0)? shallowconcat(pol1, gel(Flx_factor(zp, el), 1))
                     : vec_append(pol1, zp);
    }
    m=mp;pol0=pol1;
  }
  return gerepilecopy(av, pol0);
}

static GEN
FpX_conj(GEN g, GEN T, GEN p0n, long d, long p)
{
  long i, l = lg(g);GEN z;
  z = cgetg(d+2, t_POL);z[1] = evalsigne(1) | evalvarn(0);
  for(i=0;i<d;i++) gel(z, 2+i) = gen_0;
  for(i=2;i<l;i++) gel(z, 2+(p*(i-2))%d) = gel(g, i);
  return FpX_rem(FpX_renormalize(z, lg(z)), T, p0n);
}

static GEN
FqX_conj(GEN g, GEN T, GEN p0n, long d, long p)
{
  long i, l = lg(g);GEN z;
  z = cgetg(l, t_POL);z[1] = evalsigne(1) | evalvarn(0);
  for(i=2;i<l;i++) gel(z, i) = FpX_conj(gel(g, i), T, p0n, d, p);
  return z;
}

static GEN
FqX_norm(GEN g, GEN T, GEN p0n, long d, long p)
{
  long i, d_chi = degpol(T);GEN z = g, z1 = g;
  for(i=1;i<d_chi;i++)
  {
    z1 = FqX_conj(z1, T, p0n, d, p);
    z = FqX_mul(z, z1, T, p0n);
  }
  return z;
}

static GEN
FpXX_translate1(GEN g, GEN p)
{    
  long i, l;GEN z;
  l = lg(g);
  z = cgetg(3, t_POL); z[1] = evalsigne(1) |  evalvarn(0);
  gel(z, 2) = gel(g, l-1);
  for(i=l-2; i>=2; i--)
  {
    z  =  FpXX_add(z, RgX_shift_shallow(z, 1), p);   /* z=z*(1+X) */
    gel(z, 2) = FpX_add(gel(z, 2), gel(g, i), p);
  }
  return z ;
}

static long
FpXX_weier_deg(GEN pol)
{
  long i, l = lg(pol);
  for(i=2;i<l && lgpol(gel(pol, i))==0;i++);
  return (i<l)?i-2:-1;
}

static long
FpX_weier_deg(GEN pol, GEN p)
{
  long i, l = lg(pol);
  for(i=2;i<l && dvdii(gel(pol, i), p);i++);
  return (i<l)?i-2:-1;
}

static GEN
FpXX_to_FpX_check(GEN g, GEN p0, long d)
{
  long i, l = lg(g);GEN z;
  z = cgetg(l,t_POL);z[1] = evalsigne(1) | evalvarn(0);
  for(i=2;i<l;i++)
  {
   if(degpol(gel(g, i))>0) pari_err_BUG("FqX_norm failed\n");
   /*
   gel(z, i) = truecoeff(gel(g, i),0);
   */
   gel(z, i) = constant_coeff(gel(g, i));
  }
  if(FpX_weier_deg(z, p0) != d) pari_err_BUG("FqX_norm : wrong degree\n");
  return z ;
}

static GEN
FpXX_to_FpX(GEN g)
{
  long i, l = lg(g);GEN z;
  z = cgetg(l,t_POL);z[1] = evalsigne(1) | evalvarn(0);
  for(i=2;i<l;i++) gel(z, i) = truecoeff(gel(g, i), 0);
  return z ;
}

/* assume 0 <= x[i], y[j] <= m-1 */
static GEN
FpV_shift_add(GEN x, GEN y, GEN m, long start, long end)
{
  long i, j;
  for(i=start, j=1;i<=end;i++, j++)
  {
    GEN z = addii(gel(x, i), gel(y, j));
    gel(x, i) = (cmpii(z, m)>=0)?subii(z, m):z;
  }
  return x;
}

/* assume 0 <= x[i], y[j] <= m-1 */
static GEN
FpV_shift_sub(GEN x, GEN y, GEN m, long start, long end)
{
  long i, j;
  for(i=start, j=1;i<=end;i++, j++)
  {
    GEN z = subii(gel(x, i), gel(y, j));
    gel(x, i) = (signe(z)<0)?addii(z, m):z;
  }
  return x;
}

/* assume 0 <= x[i], y[j] <= m-1 */
static GEN
Flv_shift_add(GEN x, GEN y, ulong m, long start, long end)
{
  long i, j;
  for(i=start, j=1;i<=end;i++, j++)
  {
    ulong xi = x[i], yj = y[j];
    x[i] = Fl_add(xi, yj, m);
  }
  return x;
}

/* assume 0 <= x[i], y[j] <= m-1 */
static GEN
Flv_shift_sub(GEN x, GEN y, ulong m, long start, long end)
{
  long i, j;
  for(i=start, j=1;i<=end;i++, j++)
  {
    ulong xi = x[i], yj = y[j];
    x[i] = Fl_sub(xi, yj, m);
  }
  return x;
}

static GEN
ZX_Z_divexact_check(GEN y, GEN x)
{
  long i, l = lg(y);
  GEN z = cgetg(l,t_POL); z[1] = y[1];
  for(i=2; i<l; i++)
  {
    if(dvdii(gel(y, i), x)==0) pari_err_BUG("xi_n for f==p failed.\n");
    gel(z, i) = diviiexact(gel(y, i), x);
  }
  return z;
}

static long
Flx_weier_deg(GEN pol, long p)
{
  long i, l = lg(pol);
  for(i=2;i<l && pol[i]%p==0;i++);
  return (i<l)?i-2:-1;
}

static GEN
Flxn_shift_mul(GEN g, long n, GEN p, long d, long m)
{
  return Flx_shift(Flxn_mul(g, p, d, m), n);
}

static GEN
Flx_divuexact(GEN g, ulong p)
{
  long i, l = lg(g);GEN z;
  z = cgetg(l,t_VECSMALL);
  z[1] = evalvarn(0);
  for(i=2;i<l;i++)
  {
   if(g[i]%p) pari_err_BUG("Flx_divuexact failed\n");
   z[i] = g[i]/p;
  }
  return z;
}

/* z[1+i]=x^i (mod pol, el), 0<=i<=n */
static GEN
Flx_powers(long n, GEN pol, long el)
{
  long i;GEN z, x = pol1_Flx(0);
  z = cgetg(n+2, t_VEC);
  gel(z, 1) = x;
  for(i=1;i<=n;i++)
  {
    x = Flx_rem(Flx_shift(x, 1), pol, el);
    gel(z, 1+i) = x;
  }
  return z;
}

/*
*  Flx_translate1(g, pn) becomes slow when degpol(g)>1000.
*  So I wrote Flxn_translate1().
*  I need lambda to truncate pol.
*  But I need to translate T --> 1+T to know lambda.
*  Though the code has a little overhead, it is still fast.
*/
static GEN
Flxn_translate1(GEN g, long p, long n)
{
  long i, j, r, d, lam, pn;GEN z;
  pn = upowuu(p, n);
  for(lam=10;lam;lam<<=1)  /* first expected upper bound is 10 */
  {
    for(r=1;upowuu(p, r)<lam;r++);    /* r is min int s.t. lam<=p^r */
    if((d = (n-r+2)*lam+1)>=pn) d = pn;
    z = const_vecsmall(d, 0);  /* z[1],...,z[d] <--> a_0,...,a_{d-1} */
    for(i=degpol(g);i>=0;i--)
    {
      for(j=d;j>1;j--) z[j] = Fl_add(z[j], z[j-1], pn);  /* z = z*(1+T) */
      z[1] = Fl_add(z[1], g[2+i], pn);
    }
    z = Flv_to_Flx(z, 0);
    if (Flx_weier_deg(z, p) <= lam) return z;
  }
  return NULL; /* dummy */
}

/*
* lam==-1 ==> return 0 (lam can't be determined)
* lam== 0 ==> return 1
* lam>= 1 ==> return dist. poly. of degree lam.
*/
static GEN
Flx_ser_to_pol(GEN g, long p, long n)
{
  long i, r, d, dg = degpol(g), lam, pn, t;
  GEN U, UINV, P, PU, g0, g1, gp, gU;
  if((lam = Flx_weier_deg(g, p))<=0) return(Fl_to_Flx(lam+1, 0));
  pn = upowuu(p, n);
  if(lam>=pn) pari_err(e_MISC,"Flx_ser_to_pol: less precision. Increase n!");
  for(r=1;upowuu(p, r)<lam;r++);    /* r is min int s.t. lam<=p^r */
  if((d = (n-r+2)*lam+1)>=dg) d = dg;  /* d=min((n-r+2)*lam+1, dg) */
  if(d<=lam) d=1+lam;
  /* printf("lam=%ld r=%ld d=%ld\n", lam, r, d); */
  g = Flxn_red(g, d);
  t = Fl_inv(g[2+lam],  pn);
  g = Flx_Fl_mul(g, t, pn);  /* normalized so as g[2+lam]=1 */
  U = Flx_shift(g, -lam);
  UINV = Flxn_inv(U, d, pn);
  P = Flx_divuexact(Flxn_red(g, lam), p);
  PU = Flxn_mul(P, UINV, d, pn);
  gU = Flxn_mul(g, UINV, d, pn);
  g0 = pol1_Flx(0);
  g1 = pol1_Flx(0);
  for(i=1;i<n;i++)
  {
   g1 = Flxn_shift_mul(g1, -lam, PU, d, pn);
   gp = Flx_Fl_mul(g1, upowuu(p, i), pn);
   g0 = (i&1)?Flx_sub(g0, gp, pn):Flx_add(g0, gp, pn);
  }
  g0 = Flxn_mul(g0, gU, lam+1, pn);
  g0 = Flx_red(g0, upowuu(p, (p==2)?n-r:n+1-r));
  return g0;
}

/* compute e_n from the Stickelberger element when kro(m,p)==0.
*  Coates-Lichtenbaum, Ann. Math. vol.98 No.3 (1973), 498-550, Lemma 2.15
*  If kro(m,p)!=1, then orders of two groups coincide.
*/
static long
#ifdef VECSMALL_CHI
get_en_stk(long p, long m, long n, GEN chi)
#else
get_en_stk(long p, long m, long n, int *chi)
#endif
{ return Z_lval(ZM_det(pol2mat(quadstk(p, m, n, chi), upowuu(p, n))), p);}

/*
*  xi_n and Iwasawa pol. for Q(sqrt(m)) and p
*
* (flag&1)!=0 ==> output xi_n
* (flag&2)!=0 ==> output power series
* (flag&4)!=0 ==> output Iwasawa polynomial
*/
static GEN
quadstkpol0(long p, long m, long n, long flag)
{
  long pn;GEN stk, stk2, ser, ser2, pol, pol2;
  pari_timer ti;
  if(p==2 && (m==-1 || m==-2 || m==-3 || m==-6))
    pari_err(e_MISC,"(p,m)=(2,%ld) is excluded.\n",m);
  if(p==3 && m==-3) pari_err(e_MISC,"(p,m)=(3,-3) is excluded.\n");
  if(p==2 && m%2==0) m/=2;
#ifdef VECSMALL_CHI
  GEN chi = set_quad_chi_2(m);
#else
  int *chi = set_quad_chi_2(m);
#endif
  stk = (p==2)? quadstk2(m, n, chi)
              : quadstk(p, m, n, chi);
  if(flag==1) return stk;

  pn = upowuu(p, n);
  stk2 = ZX_to_Flx(stk, pn);
  if (DEBUGLEVEL >= 6) timer_start(&ti);
  if((flag&2) || pn<100)
    ser2 = Flx_translate1(stk2, pn);  /* faster than FpX_translate() */
  else
    ser2 = Flxn_translate1(stk2, p, n);
  if (DEBUGLEVEL >= 6) timer_printf(&ti, "Flx_translate1");
  ser = Flx_to_ZX(ser2);
  setvarn(ser, fetch_user_var("T"));
  if(flag==2) return ser;
  else if(flag==3) return mkvec2(stk, ser);
  pol2 = Flx_ser_to_pol(ser2, p, n);
  pol = Flx_to_ZX(pol2);
  setvarn(pol, fetch_user_var("T"));
  if(flag==4) return pol;
  else if(flag==5) return mkvec2(stk, pol);
  else if(flag==6) return mkvec2(ser, pol);
  else return mkvec3(stk, ser, pol);
}

/* m is negative square-free integer
*  Kida, Tohoku Math. J. vol.31 (1979), 91-96, Theorem 1.
*/
static GEN
quadlambda2(long m)
{
  long lam;
  m = -m;
  if((m&1)==0) m>>=1;  /* lam_2(Q(sqrt(-m)))=lam_2(Q(sqrt(-2*m))) */
  if(m<=3) lam = 0;
  else
  {
    GEN fa = factoru(m), p = gel(fa, 1);
    long i, l = lg(p);
    for(lam=-1,i=1;i<l;i++)
      lam+=1L<<(-3+vals(p[i]-1)+vals(p[i]+1));
  }
  return mkvecs(lam);
}

/*
* Iwasawa lambda invariant of Q(sqrt(m)) (m<0) for p
* |A_n|=p^(e[n])
* kross(m,p)!=1 : e[n]-e[n-1]<eulerphi(p^n)  ==> lambda=e[n]-e[n-1]
* kross(m,p)==1 : e[n]-e[n-1]<=eulerphi(p^n) ==> lambda=e[n]-e[n-1]
* Gold, Acta Arith. XXVI (1974), p.25, Cor. 3
* Gold, Acta Arith. XXVI (1975), p.237, Cor.
*/
static GEN
quadlambda0(long p, long m)
{
  long flag, n, phipn, e[30];
  if(p==2) return(quadlambda2(m));
  else if(p==3 && m==-3) {e[0] = e[1] = 0;return ary2vec2(e, 1);}
  flag = kross(m, p);
  e[0] = Z_lval(quadclassno(quaddisc(stoi(m))), p);
  if(flag!=1 && e[0]==0) {e[1] = 0;return ary2vec2(e, 1);}
#ifdef VECSMALL_CHI
  GEN chi = set_quad_chi_2(m);
#else
  int *chi = set_quad_chi_2(m);
#endif
  phipn = p-1;  /* phipn=phi(p^n) */
  for(n=1;n;n++)
  {
   e[n] = flag?e[n-1]+ediff(p, m, n, chi):get_en_stk(p, m, n, chi);
   if((flag!=1 && (e[n]-e[n-1]<phipn))||
      (flag==1 && (e[n]-e[n-1]<=phipn))) break;
   phipn*=p;
  }
  return ary2vec2(e, n);
}

/* factor n-th cyclotomic polynomial mod p^r and return a minimal
*  polynomial of zeta_n over Q_p */
static GEN
set_minpol(ulong n, GEN p, ulong r, long flag)
{
  GEN z, v, pol, pr;
  pari_timer ti;
  if (umodiu(p, n)==1) /* zeta_n in Z_p, faster than polcyclo() */
  {
    GEN prm1 = powiu(p, r-1), pr = mulii(prm1, p); /* pr=p^r */
    GEN prn = diviuexact(subii(pr, prm1), n);      /* prn=phi(p^r)/n */
    z = Fp_pow(pgener_Fp(p), prn, pr);
    return deg1pol_shallow(gen_1, Fp_neg(z, pr), 0);
  }
  pr = powiu(p, r);
  pol = FpX_red(polcyclo(n, 0), pr); /* separable mod p */
  if (DEBUGLEVEL >= 6) timer_start(&ti);
  z = (n>500)? gel(FpX_factor_polcyclo(n, p), 1)
             : gcoeff(FpX_factor(pol, p), 1, 1);
  if (DEBUGLEVEL >= 6) timer_printf(&ti, "FpX_factor:n=%ld  ", n);
  if (degpol(z) == degpol(pol)) return pol;
  v = ZpX_liftfact(pol, mkvec2(z, FpX_div(pol, z, p)), pr, p, r);
  return gel(v, 1);
}

static GEN
set_minpol_teich(ulong g_K, GEN p, ulong r)
{
  GEN prm1 = powiu(p, r-1), pr = mulii(prm1, p), z;
  z = Fp_pow(Fp_inv(utoi(g_K), p), prm1, pr);
  return deg1pol_shallow(gen_1, Fp_neg(z, pr), 0);
}

static long
srh_1(GEN H)
{
  GEN bits = gel(H, 3);
  ulong f = bits[1];
  return F2v_coeff(bits, f-1);
}

/*
* aH=g^iH ==> chi(a)=zeta_n^(-i); Chi[g^iH]=i
* Since Chi[0] never be accessed, Chi is implemented by vecsmall.
*/
static GEN
get_chi(GEN H1)
{
  GEN H = gel(H1, 3), H1data = gel(H1, 2);
  long d_K = H1data[1], f = H1data[2], h = H1data[3], g = H1data[5];
  long i, j, gi;
  GEN Chi = const_vecsmall(f-1, -1);

  for (j=1; j<=h; j++) Chi[H[j]] = 0; /* i = 0 */
  for (i = 0, gi = 1; i<d_K; i++)
  {
    for (j=1; j<=h; j++) Chi[Fl_mul(gi, H[j], f)] = i;
    gi = Fl_mul(gi, g, f);
  }
  return Chi;
}

/*
*  (1/f)sum_{1<=a<=f}a*chi^{-1}(a) = -(1/(2-chi(a)))sum_{1<=a<=f/2} chi^{-1}(a)
*  does not overflow
*/
static GEN
ber_s2(GEN Chi, long f, long d, long j)
{
  GEN B1 = const_vecsmall(d, 0), B2 = const_vecsmall(d, 0);
  long i, f2=f/2, *B11= B1+1, *B22=B2+1;

  for (i = 1; i <= f2; i++)
    if (Chi[i] >= 0) B11[(Chi[i]*j)%d] ++;

  if (Chi[2]>=0) B22[Fl_neg((Chi[2]*j)%d, d)] = 1;
  B22[0]-=2;
  return mkvec2(zv_to_ZX(B1, 0), zv_to_ZX(B2, 0));
}

/* The computation is fast when p^n and el=1+k*f*p^n are less than 2^64
*  for m <= n <= M
*  We believe M>=3 is enough when f%p=0 and M>=2 is enough for other case
*  because we expect that p^2 does not divide |A_{K,psi}| for a large p.
*  FIXME: M should be set according as p and f. */
static void
set_p_f(GEN pp, ulong f, long *pm, long *pM)
{
  ulong p = itou_or_0(pp);
  if (!p || p >= 2000000) { *pm=2; *pM = dvdui(f, pp)? 3: 2; }
  else if (p == 3)      { *pm=5; *pM=20;}
  else if (p == 5)      { *pm=5; *pM=13;}
  else if (p == 7)      { *pm=5; *pM=11;}
  else if (p == 11)     { *pm=5; *pM=9; }
  else if (p == 13)     { *pm=5; *pM=8; }
  else if (p < 400)     { *pm=5; *pM=7; }
  else if (p < 5000)    { *pm=3; *pM=5; }
  else if (p < 50000)   { *pm=2; *pM=4; }
  else                  { *pm=2; *pM=3; }
}

static GEN
subgp2ary(GEN H)
{
  GEN v = gel(H, 3);
  long i, j, n = v[1];
  GEN w = cgetg(n+1, t_VECSMALL);
  for (i = 1, j = 0; i <= n; i++)
    if (F2v_coeff(v,i)) w[++j] = i;
  return vecsmall_shorten(w, j);
}

/* \prod g[i]^h[i] mod q */
static ulong
Flv_factorback(GEN g, GEN h, ulong q)
{
  long i, l = lg(g);
  ulong t;
  if (l == 1) return 1UL;
  t = Fl_powu(g[1], h[1], q);
  for (i = 2; i < l; i++) t = Fl_mul(t, Fl_powu(g[i], h[i], q), q);
  return t;
}

static GEN
FlvV_factorback(GEN g, GEN H, ulong q)
{
  long i, l = lg(H);
  GEN v = cgetg(l, t_VECSMALL);
  for (i = 1; i < l; i++) v[i] = Flv_factorback(g, gel(H,i), q);
  return v;
}

/* lift chi character on G/H to character on G */
static GEN
zncharlift(GEN chi, GEN ncycGH, GEN U, GEN cycG)
{
  GEN nchi = char_normalize(chi, ncycGH);
  GEN c = ZV_ZM_mul(gel(nchi, 2), U), d = gel(nchi, 1);
  return char_denormalize(cycG, d, c);
}

/* 0 <= c[i] < d, i=1..r; (c[1],...,c[r], d) = 1; find e[i] such that
 * sum e[i]*c[i] = 1 mod d */
static GEN
Flv_extgcd(GEN c, ulong d)
{
  long i, j, u, f, l = lg(c);
  GEN e = zero_zv(l-1);
  if (l == 1) return e;
  for (f = d, i = 1; f != 1 && i < l; i++)
  {
    f = cbezout(f, itou(gel(c,i)), &u, &e[i]);
    if (!e[i]) continue;
    e[i] = umodsu(e[i], d);
    u = umodsu(u, d);
    if (u != 1) for (j = 1; j < i; j++) e[j] = Fl_mul(e[j], u, d);
  }
  return e;
}

static long
srh_x(GEN T, long n, long x)
{
  for (; x<n; x++) if (!T[x]) return x;
  return -1;
}

/*
* G is a cyclic group of order d. hat(G)=<chi>.
* chi, chi^p, ... , chi^(p^(d_chi-1)) are conjugate.
* {chi^j | j in C} are repre. of Q_p-congacy classes of inj. chars.
*/
static GEN
set_C(long p, long d, long d_chi, long n_conj)
{ long i, j, x, y, pmodd = p%d;
  GEN T = const_vecsmall(d, 0)+1;
  GEN C = cgetg(1+n_conj, t_VECSMALL);
  for (i=0, x=1; x >= 0; x = srh_x(T, d, x))
  {
    if (cgcd(x, d)==1) C[++i] = x;
    for (j=0, y=x; j<d_chi; j++) T[y=Fl_mul(y, pmodd, d)] = 1;
  }
  return C;
}

/* f!=p, coef is const. if d_chi==1
* return mod p for lam_chi()
* return mod p^n for pol_chi()
*/
static GEN
get_xi_1(long p, long f, long n, long d, long k, GEN K, long minpolpow)
{ long i, j, j0, f0, d_chi, pn, pn1, deg, pn1f, **xi;
  GEN p0 = utoi(p), Chi = gel(K, 1), MinPol = gel(K, 2), z;
  long *Gam = (long *)gel(K, 3)+1;

  f0 = (f%p)?f:f/p;
  d_chi = order_f_x(d, p%d);   /* equal to deg(MinPol) */
  pn = upowuu(p, n);pn1 = p*pn;pn1f = pn1%f;
  xi = (long**)stack_malloc(sizeof(long)*pn);
  for(i=0;i<pn;i++) xi[i] = (long *)stack_calloc(sizeof(long)*d);
  for(j=1;j<pn1;j++)  /* xi is exact */
    {
     long ipn1,*xij0;
     if((j0 = Gam[j])<0) continue;
     ipn1 = j%f;xij0 = xi[j0];
     for(i=1;i<f0;i++)
       {
        if((ipn1 += pn1f) >= f) ipn1 -= f;
        if(ipn1==0 || (deg = Chi[ipn1])<0) continue;
        xij0[(deg*k)%d] += i;
       }
    }
  z = cgetg(pn+2,t_POL);z[1] = evalsigne(1) | evalvarn(0);
  for(i=0;i<pn;i++)  /* z is exact */
     {gel(z, 2+i) = cgetg(d+2, t_POL);
      gmael(z, 2+i, 1) = (GEN)(evalsigne(1) | evalvarn(0)) ;
      for(j=0;j<d;j++) gmael(z, 2+i, 2+j) = utoi(xi[i][j]);}
  /* Now z is approx. */
  for(i=0;i<pn;i++) gel(z,2+i)=FpX_rem(gel(z,2+i), MinPol, powiu(p0,minpolpow));
  z = FpXX_renormalize(z, lg(z));  /* important */
  /* pari_printf("d_chi=%ld xi_rem=%Ps\n", d_chi, z); */
  if(d_chi==1) z = FpXX_to_FpX(z);
  return z;
}

/* f==p, coef is always const.
* return mod p for lam_chi()
* return mod p^n for pol_chi()
*/
static GEN
get_xi_2(long p, long f, long n, long d, long k, GEN K, long minpolpow)
{ long a, amodf, i, j, j0, pn, pn1, deg, **xi;
  GEN p0 = utoi(p), Chi = gel(K, 1), MinPol = gel(K, 2), z;
  long *Gam = (long *)gel(K, 3)+1;

  pn = upowuu(p, n);pn1 = p*pn;
  xi = (long**)stack_malloc(sizeof(long)*pn);
  for(i=0;i<pn;i++) xi[i] = (long *)stack_calloc(sizeof(long)*d);
  for(a=1,amodf=0;a<pn1;a++)  /* xi is exact */
    {
     if(++amodf>=f) amodf -= f;
     if((j0=Gam[a])<0 || amodf==0 || (deg=Chi[amodf])<0) continue;
     xi[j0][(deg*k)%d] += a;
    }
  z = cgetg(pn+2,t_POL);z[1] = evalsigne(1) | evalvarn(0);
  for(i=0;i<pn;i++)  /* z is exact */
    {
     gel(z, 2+i) = cgetg(d+2, t_POL);
     gmael(z, 2+i, 1) = (GEN)(evalsigne(1) | evalvarn(0)) ;
     for(j=0;j<d;j++) gmael(z, 2+i, 2+j) = utoi(xi[i][j]);
    }
  /* Now z is approx. */
  for(i=0;i<pn;i++) gel(z,2+i)=FpX_rem(gel(z,2+i), MinPol, powiu(p0,minpolpow));
  z = FpXX_renormalize(z, lg(z));  /* important */
  z = FpXX_to_FpX(z);
  z = ZX_Z_divexact_check(z, powiu(p0, n+1));
  return z;
}

static GEN
pol_chi(GEN p0, GEN H1, GEN K, long j, long n)
{ 
  pari_sp av = avma;
  long pn, d_chi, x, wd, p = itou(p0);GEN z;
  GEN H1data = gel(H1, 2), xi;
  long d = H1data[1], f = H1data[2];
  GEN Chi = gel(K, 1), MinPol = gel(K, 2);
  GEN p0n = powiu(p0, n), ser, ser2, pol, pol2;

  d_chi = order_f_x(d, p%d);
  pn = upowuu(p, n);
  /*
  pari_printf("pol_chi:p=%ld f=%ld d=%ld d_chi=%ld n=%ld Chi[%ld]=%ld\n",
               p,f,d,d_chi,n,p,Chi[p%f]);
  */
  if (f == d+1 && p == f && j == 1) /* Teichmuller */
    return pol_1(0);

  if(Chi[p%f])   /* try to determine using xi_0 */
  {
    GEN y, B1, B2, p02 = powiu(p0, 2);
    y = ber_s2(Chi, f, d, j);
    B1 = FpX_rem(gel(y, 1), MinPol, p02);
    B2 = FpX_rem(gel(y, 2), MinPol, p02);
    if (degpol(B1) >= 0)
    {
      x = ZX_pval(B1, p0) - ZX_pval(B2, p0);
      if (x < 0) pari_err_BUG("ipolsubcyclo [Bernoulli number]");
      if (x==0) return pol_1(0); /* trivial */
    }
  }

  /*
  * compute Stickelberger element
  * xi is FpX if d_chi==1 and FpXX if d_chi>1
  */
  xi = (f==p)?get_xi_2(p,f,n,d,j,K,2*n+1):get_xi_1(p,f,n,d,j,K,n);
  z = (d_chi==1)?FpX_translate1(xi, p0):FpXX_translate1(xi, p0);
  wd = (d_chi==1)?FpX_weier_deg(z, p0):FpXX_weier_deg(z);
  if(wd<0) pari_err(e_MISC,"Weierstrass degree is not finite. Increase n!\n");
  else if(wd==0) return pol_1(0);

  /* wd>0, convert to dist. poly. */
  ser = (d_chi==1)?FpX_translate1(xi, p0n):FpXX_translate1(xi, p0n);
  if(d_chi>1)
   {
     ser = FqX_norm(ser, MinPol, p0n, d, p);
     ser = FpXX_to_FpX_check(ser, p0, wd*d_chi);
   }
  ser2 = ZX_to_Flx(ser, pn);
  pol2 = Flx_ser_to_pol(ser2, p, n);
  pol = Flx_to_ZX(pol2);
  setvarn(pol, fetch_user_var("T"));
#ifdef DEBUG
  if(wd>0 && d_chi>1)
    err_printf("(wd,d_chi,p,f,d,j,H)=(%ld,%ld,%ld,%ld,%ld,%ld,%Ps)\n",
       wd,d_chi,p,f,d,j,gmael(H1, 1, 1));
#endif
  return gerepilecopy(av, pol);
}

static GEN
imag_cyc_pol(GEN p0, GEN H1, long n, GEN MinPol)
{ 
  pari_sp av = avma;
  GEN H1data = gel(H1, 2), MinPol2, Kdata;
  long i, j, q0, pn1, minpolpow;
  long p = itou(p0);
  long d = H1data[1], f = H1data[2];
  long d_chi, n_conj;
  GEN z = nullvec(), Gam, C, Chi = get_chi(H1);

  q0 = (f%p)?f*p:f;
  d_chi = order_f_x(d, p%d);n_conj = eulerphiu(d)/d_chi;
  C = set_C(p, d, d_chi, n_conj);
  pn1 = upowuu(p, n+1);
  Gam = set_gam((1+q0)%pn1, p, n);
  minpolpow = (f==p)?2*n+1:n;  /* for xi_n */
  MinPol2 = FpX_red(MinPol, powiu(p0, minpolpow));
  Kdata = mkvec3(Chi, MinPol2, Gam);
  for(i=1;i<=n_conj;i++)
  {GEN z1;
   j = C[i];
   z1 = pol_chi(p0, H1, Kdata, j, n);
   if(degpol(z1)) z = vec_append(z, z1);
  }
  return gerepilecopy(av, z);
}


/* HH a t_VECSMALL listing group generators */
static GEN
abelstkpol(long p, long f, GEN HH, long n)
{ 
  pari_sp av = avma;
  long OrderG, OrderH, DegF, start_pow, max_pow;
  long fH, i, s, n_f, h, gen, lgH, expG;
  GEN gG, gH, gGH, cycGH, cycG, ncycGH, vChar, P, U, Ui, vH1, G, vData;
  GEN z = nullvec(), p0 = utoi(p);

  /* pari_printf("p=%ld f=%ld HH=%Ps n=%ld\n", p, f, HH, n); */
  HH = znstar_generate(f, HH); /* group elements */
  fH = znstar_conductor(HH);
  if (fH == 1) return gen_0;
  else if (fH != f) { HH = znstar_reduce_modulus(HH, fH); f = fH; }

  OrderH = h = zv_prod(gel(HH, 2));
  OrderG = eulerphiu(f);
  DegF = OrderG / OrderH;
  s = srh_1(HH);
  if (DegF == 1 || s == 1) return gen_0;

  G = znstar0(utoipos(f), 1);
  gH = gel(HH, 1); lgH = lg(gH); /* gens. of H */
  P = cgetg(lgH, t_MAT);
  for (i = 1; i < lgH; i++) gel(P,i) = Zideallog(G, utoi(gH[i]));

  /* group structure of G/H */
  cycG = znstar_get_cyc(G);
  expG = itou(gel(cycG, 1));
  /* gG generators of G, gGH generators of G/H: gGH = g.Ui, g = gGH.U */
  cycGH = ZM_snf_group(hnfmodid(P, cycG), &U, &Ui);
  ncycGH = cyc_normalize(cycGH);
  gG = ZV_to_Flv(znstar_get_gen(G), f); /* gens. of G */
  /* generators of G/H */
  gGH = FlvV_factorback(gG, ZM_to_Flm(Ui, expG), f);
  vChar = chargalois(cycGH, NULL); n_f = lg(vChar)-2;
  vH1 = cgetg(n_f+1, t_VEC);
  for (i = 1; i <= n_f; i++)
  { /* skip trivial character */
    GEN chi = gel(vChar,i+1), nchi = char_normalize(chi, ncycGH);
    GEN chiG, E, H1, C = gel(nchi, 2);
    long e, he, d = itou(gel(nchi, 1));
    /* chi(prod g[i]^e[i]) = e(sum e[i]*C[i] / d), chi has order d = #(G/H1)*/
    E = Flv_extgcd(C, d); /* \sum C[i]*E[i] = 1 in Z/dZ */

    chiG = zncharlift(chi, ncycGH, U, cycG);
    H1 =  charker(cycG, chiG); /* H1 < G with G/H1 cyclic */
    e = itou( zncharconductor(G, chiG) ); /* cond H1 = cond chi */
    H1 = FlvV_factorback(gG, ZM_to_Flm(H1, expG), e);
    gen = Flv_factorback(gGH, E, e); /* G/H1 = <gen>, chi(gen) = e(1/d) */
    H1 = znstar_generate(e, H1);
    he = eulerphiu(e) / d;
    /* G/H1 = <gen> cyclic of index d, e = cond(H1) */
    gel(vH1,i) = mkvec3(H1, mkvecsmall5(d,e,he,srh_1(H1), gen), subgp2ary(H1));
  }

  if (DegF%p==0) pari_err(e_MISC,"p=%ld divides [F:Q]=%ld\n", p, DegF);
  set_p_f(p0, f, &start_pow, &max_pow);
  if (2*n+1>max_pow) max_pow = 2*n+1;
  vData = const_vec(DegF, NULL);
  for (i=1; i<=n_f; i++) /* prescan. set Teichmuller */
  {
    GEN H1 = gel(vH1,i), H1data = gel(H1, 2);
    long d_K = H1data[1], f_K = H1data[2], g_K = H1data[5];

    if (f_K == d_K+1 && p == f_K)  /* found K=Q(zeta_p) */
    {
      gel(vData, d_K) = set_minpol_teich(g_K, p0, max_pow);
      break;
    }
  }

  for (i=1; i<=n_f; i++)
  {
   GEN H1 = gel(vH1,i), H1data = gel(H1, 2), MinPol, z1;
   long d_K=H1data[1], s = H1data[4];

   if(s) continue;  /* F is real */
#ifdef DEBUG
   err_printf("  handling %s cyclic subfield K, deg(K)=%ld, cond(K)=%ld\n",
               s? "a real": "an imaginary", d_K, H1data[2]);
#endif
   if (!gel(vData, d_K))
     gel(vData, d_K) = set_minpol(d_K, p0, max_pow, 0);
   MinPol = gel(vData, d_K);
   if(lg(z1 = imag_cyc_pol(p0, H1, n, MinPol)) > 1) z = shallowconcat(z, z1);
  }
  avma = av;
  return gerepilecopy(av, z);
}

/*
* return lam_psi (psi=chi^j) if lam_psi is determined.
* else return -1.
*/
static long
lam_chi(GEN p0, GEN H1, GEN K, long j, long n)
{ 
  pari_sp av = avma;
  long d_chi, x, wd, p = itou(p0);GEN z;
  GEN H1data = gel(H1, 2), xi;
  long d = H1data[1], f = H1data[2];
  GEN Chi = gel(K, 1), MinPol = gel(K, 2);

  d_chi=order_f_x(d, p%d);
  /*
  pari_printf("lam_chi:p=%ld f=%ld f0=%ld d=%ld n=%ld Chi[%ld]=%ld\n",
                p,f,f0,d,n,p,Chi[p%f]);
  */
  if (f == d+1 && p == f && j == 1) /* Teichmuller */
    return 0;

  if(n==0)   /* compute by Bernoulli number */
    {
      GEN  y, B1, B2, p02=powiu(p0, 2);
      y = ber_s2(Chi, f, d, j);
      B1 = FpX_rem(gel(y, 1), MinPol, p02);
      B2 = FpX_rem(gel(y, 2), MinPol, p02);
      if (degpol(B1)<0 || degpol(B2)<0) return -1; /* 0 mod p^2 */
      x = ZX_pval(B1, p0) - ZX_pval(B2, p0);
      if (x<0) pari_err_BUG("lambdasubcyclo [Bernoulli number]");
      avma = av;
      return (x==0)?0:-1;
     }

  /*
  * n>=1, compute by Stickelberger element
  * xi is FpX if d_chi==1 and FpXX if d_chi>1
  */
  xi = (f==p)?get_xi_2(p,f,n,d,j,K,n+2):get_xi_1(p,f,n,d,j,K,1);
  z = (d_chi==1)?FpX_translate1(xi, p0):FpXX_translate1(xi, p0);
  wd = (d_chi==1)?FpX_weier_deg(z, p0):FpXX_weier_deg(z);
#ifdef DEBUG
  if(wd>0 && d_chi>1)
    err_printf("(wd,d_chi,p,f,d,j,H)=(%ld,%ld,%ld,%ld,%ld,%ld,%Ps)\n",
       wd,d_chi,p,f,d,j,gmael(H1, 1, 1));
#endif
  avma = av;
  return (wd<0)?-1:wd*d_chi;
}

/*
* K is an imaginary cyclic extension of degree d contained in Q(zeta_f)
* H is the subgr of G=(Z/fZ)^* corresponding to K
* h=|H|, d*h=phi(f)
* G/H=<g> i.e. g^d \in H
* d_chi=[Q_p(zeta_d):Q_p], i.e. p^d_chi=1 (mod d)
* An inj. char. of G(K/Q) is automatically imaginary.
*
* G(K/Q)=G/H=<g>, chi:G(K/Q)-->overline{Q_p} s.t. chi(g)=zeta_d^(-1)
* Chi[a]=k, k<0  ==> chi(a)=0
*           k>=0 ==> chi(a)=zeta_d^(-k)
* psi=chi^j, j in C : repre. of inj. odd char.
* psi(p)==1 <==> chi(p)^j==0 <==> j*Chi[p]=0 (mod d) <==> Chi[p]==0
*/
static long
imag_cyc_lam(GEN p0, GEN H1, GEN MinPol, long max_pow)
{
  pari_sp av = avma;
  GEN H1data = gel(H1, 2), MinPol2, Kdata;
  long i, q0, pn1, n, lam = 0, minpolpow;
  long p = itou(p0);
  long d = H1data[1], f = H1data[2];
  long d_chi, n_conj, n_done = 0;
  GEN Gam=NULL, C, Lam, Chi = get_chi(H1);

  q0 = (f%p)?f*p:f;
  d_chi = order_f_x(d, p%d);n_conj = eulerphiu(d)/d_chi;
  C = set_C(p, d, d_chi, n_conj);
  Lam = const_vecsmall(n_conj, -1);
  for(n=Chi[p%f]?0:1;n>=0;n++)
  {
    pn1 = upowuu(p, n+1);
    if(n) Gam = set_gam((1+q0)%pn1, p, n);
    if(n==0) minpolpow = (f%p)?1:2;  /* for B_{1,chi} */
    else minpolpow = (f==p)?n+2:1;   /* for xi_n */
    MinPol2 = (minpolpow > max_pow)? set_minpol(d, p0, minpolpow, 0)
                                   : FpX_red(MinPol, powiu(p0, minpolpow));
    Kdata = mkvec3(Chi, MinPol2, Gam);
    for(i=1;i<=n_conj;i++)
    {
     if(Lam[i]<0 && (Lam[i] = lam_chi(p0, H1, Kdata, C[i], n)) >= 0) n_done++;
    }
    if(n_conj==n_done) break;
  }
  for(i=1;i<=n_conj;i++) lam+=Lam[i];
  avma = av;
  return lam;
}

/* HH a t_VECSMALL listing group generators */
static GEN
abellambda(long p, long f, GEN HH)
{ 
  pari_sp av = avma;
  long OrderG, OrderH, DegF, lam = 0, start_pow, max_pow;
  long fH, i, s, n_f, h, gen, lgH, expG;
  long n_sub = 0, n_chi = 0;
  GEN p0 = utoi(p);
  GEN gG, gH, gGH, cycGH, cycG, ncycGH, vChar, P, U, Ui, vH1, G, vData;

  HH = znstar_generate(f, HH); /* group elements */
  fH = znstar_conductor(HH);
  if (fH == 1) return gen_0;
  else if (fH != f) { HH = znstar_reduce_modulus(HH, fH); f = fH; }

  OrderH = h = zv_prod(gel(HH, 2));
  OrderG = eulerphiu(f);
  DegF = OrderG / OrderH;
  s = srh_1(HH);
  if (DegF == 1 || s == 1) return gen_0;
#ifdef DEBUG
  err_printf("F is %s abelian field, deg(F)=%ld, ",
                         s?"a real":"an imaginary", DegF);
#endif

  G = znstar0(utoipos(f), 1);
  gH = gel(HH, 1); lgH = lg(gH); /* gens. of H */
  P = cgetg(lgH, t_MAT);
  for (i = 1; i < lgH; i++) gel(P,i) = Zideallog(G, utoi(gH[i]));

  /* group structure of G/H */
  cycG = znstar_get_cyc(G);
  expG = itou(gel(cycG, 1));
  /* gG generators of G, gGH generators of G/H: gGH = g.Ui, g = gGH.U */
  cycGH = ZM_snf_group(hnfmodid(P, cycG), &U, &Ui);
  ncycGH = cyc_normalize(cycGH);
  gG = ZV_to_Flv(znstar_get_gen(G), f); /* gens. of G */
  /* generators of G/H */
  gGH = FlvV_factorback(gG, ZM_to_Flm(Ui, expG), f);
  vChar = chargalois(cycGH, NULL); n_f = lg(vChar)-2;
  vH1 = cgetg(n_f+1, t_VEC);
  for (i = 1; i <= n_f; i++)
  { /* skip trivial character */
    GEN chi = gel(vChar,i+1), nchi = char_normalize(chi, ncycGH);
    GEN chiG, E, H1, C = gel(nchi, 2);
    long e, he, d = itou(gel(nchi, 1));
    /* chi(prod g[i]^e[i]) = e(sum e[i]*C[i] / d), chi has order d = #(G/H1)*/
    E = Flv_extgcd(C, d); /* \sum C[i]*E[i] = 1 in Z/dZ */

    chiG = zncharlift(chi, ncycGH, U, cycG);
    H1 =  charker(cycG, chiG); /* H1 < G with G/H1 cyclic */
    e = itou( zncharconductor(G, chiG) ); /* cond H1 = cond chi */
    H1 = FlvV_factorback(gG, ZM_to_Flm(H1, expG), e);
    gen = Flv_factorback(gGH, E, e); /* G/H1 = <gen>, chi(gen) = e(1/d) */
    H1 = znstar_generate(e, H1);
    he = eulerphiu(e) / d;
    /* G/H1 = <gen> cyclic of index d, e = cond(H1) */
    gel(vH1,i) = mkvec3(H1, mkvecsmall5(d,e,he,srh_1(H1), gen), subgp2ary(H1));
  }

  if (DegF%p==0) pari_err(e_MISC,"p=%ld divides [F:Q]=%ld\n", p, DegF);
  set_p_f(p0, f, &start_pow, &max_pow);
  vData = const_vec(DegF, NULL);
  for (i=1; i<=n_f; i++) /* prescan. set Teichmuller */
  {
    GEN H1 = gel(vH1,i), H1data = gel(H1, 2);
    long d_K = H1data[1], f_K = H1data[2], g_K = H1data[5];

    if (f_K == d_K+1 && p == f_K)  /* found K=Q(zeta_p) */
    {
      gel(vData, d_K) = set_minpol_teich(g_K, p0, max_pow);
      break;
    }
  }

  for (i=1; i<=n_f; i++)
  {
    pari_sp av;
    GEN H1 = gel(vH1,i), H1data = gel(H1, 2), MinPol;
    long d_K = H1data[1], s = H1data[4];

    if(s) continue;   /* F is real */
    n_sub++;n_chi+=eulerphiu(d_K)/order_f_x(d_K, p%d_K);
#ifdef DEBUG
   err_printf("  handling %s cyclic subfield K, deg(K)=%ld, cond(K)=%ld\n",
               s? "a real": "an imaginary", d_K, H1data[2]);
#endif
    if (!gel(vData, d_K))
      gel(vData, d_K) = set_minpol(d_K, p0, max_pow, 0);
    MinPol = gel(vData, d_K);
    av = avma;
    lam += imag_cyc_lam(p0, H1, MinPol, max_pow);
    avma = av;
  }
  return gerepilecopy(av, utoi(lam));
}

static GEN
ary2mat(GEN x, long n)
{
  long i, j;GEN z;
  z = cgetg(n+1,t_MAT);
  for(i=1;i<=n;i++)
  {
    gel(z,i) = cgetg(n+1,t_COL);
    for(j=1;j<=n;j++) gmael(z, i, j) = utoi(x[(i-1)*n+j-1]);
  }
  return z;
}

static long
is_cyclic(GEN x)
{
  GEN y = gel(x, 2);long i, l = lg(y), n = 0;
  for(i=1;i<l;i++) if(!isintzero(gel(y,i))) n++;
  return (n>1)?0:1;
}

static GEN
make_p_part(GEN y, GEN p0, long d_pow)
{
  long i, l=lg(y);GEN z = const_vecsmall(l-1, 0);
  for(i=1;i<l;i++) z[i] = isintzero(gel(y,i))?d_pow:Z_pval(gel(y,i), p0);
  return z;
}

static GEN
structure_MLL(GEN y, GEN p0, long d_pow)
{
  long i, l=lg(y), y0;GEN x = gen_0, z = const_vec(2, NULL);
  gel(z, 2) = const_vec(l-1, NULL);
  for(i=1;i<l;i++)
  {
    if((y0=d_pow-y[i])<0) y0=0;
    x = addiu(x, y0);
    gmael(z, 2, l-i) = utoi(y0);
  }
  gel(z, 1) = x;
  return z;
}

static GEN
find_del_el(GEN old, GEN new, long n, long n_el, long d_chi, long flag)
{
  if(n_el==1) return mkvec2(gen_1, old);
  if(cmpis(gmael(new, 2, 1), n_el)==0) return mkvec2(utoi(n), old);
  if (cmpii(gel(old, 1), gel(new, 1))>=0) return mkvec2(utoi(n), old);
  else if (n>1 && is_cyclic(new)) return mkvec2(utoi(n-1), new);
  else if(n==n_el) return mkvec2(utoi(n), old);
  if(cmpis(gel(new, 1), n*d_chi)<0) return mkvec2(utoi(n), old);
  else return NULL;
}

static GEN
subgr2vecsmall(GEN H, long h, long f)
{
  long i;GEN z = const_vecsmall(f-1, 0); /* f=lg(z) */
  for(i=1;i<=h;i++) z[H[i]] = 1; /* H[i]!=0 */
  return z;
}

static long
H0_iscontained_He(GEN H0, long e)
{
  long i, f=lg(H0);
  for(i=2;i<f;i++)
    if(H0[i] && i%e!=1) return 0;
  return 1;
}

static long
root_of_1(GEN H0, long d, long f)
{
  long e;
  for(e=f;e>1;e--)
    if(f%e==0 && (d%eulerphiu(e))==0 && H0_iscontained_He(H0, e)) break;
  return (e&1)?(e<<1):e;
}

static long
find_ele(GEN H)
{
  long i, f=lg(H);
  for(i=1;i<f;i++) if(H[i]) return i;
  return 0;
}

static void
delete_ele(GEN H, long j, long el)
{
  long f = lg(H), x = 1;
  do H[(j*x)%f] = 0;
  while((x=(x*el)%f)!=1);
}

static GEN
get_coset(GEN H, long h, long f, long el)
{
  long i, j, k = h/order_f_x(f, el%f);GEN H2, coset = const_vecsmall(k, 0);
  H2 = subgr2vecsmall(H, h, f);
  for(i=0;(j=find_ele(H2))>0;i++)
  {
    coset[1+i] = j;
    delete_ele(H2, j, el);
  }
  if (i != k) pari_err_BUG("failed to find coset\n");
  return coset;
}

static long
srh_pol(GEN xpows, GEN vn, GEN pols, long el, long k, long f)
{
  pari_sp av = avma;
  long i, j, l = lg(pols), d = degpol(gel(pols, 1));GEN pol = gel(pols, 1);

  for(i=1;i<l;i++)
  {
    GEN x, y, z;
    if (vn[i]==0) continue;
    y = gel(pols, vn[i]);
    z = pol0_Flx(0);
    for(j=0;j<=d;j++)
      z = Flx_add(z, Flx_Fl_mul(gel(xpows, 1+Fl_mul(j, k, f)), y[2+j], el), el);
    x = Flx_rem(z, pol, el);
    if (lgpol(x)==0)
      {avma = av;vn[i] = 0;return i;}  /* pols[i] is min pol of zeta_f^k */
  }
  pari_err_BUG("pclgpsubcyclo [minimal polinomial]");
  return 0;  /* to suppress warning */
}

/* el=1 (mod f) */
static long
chk_el_real_f(GEN K, ulong p, ulong d_pow, ulong el, ulong j0)
{
  pari_sp av = avma;
  GEN TR = gel(K, 1), H1 = gel(K, 3), H1data = gel(H1, 2), H = gel(H1, 3);
  ulong d_K = H1data[1], f = H1data[2], h = H1data[3], g_K = H1data[5];
  ulong  i, j, gi, d = upowuu(p, d_pow), dp = d*p, *e_chi;
  ulong g_el, z_f, flag = 0, el1f = (el-1)/f, el1dp = (el-1)/dp;
  GEN vz_f, xi_el = cgetg(d_K+1, t_VECSMALL);

  e_chi=(ulong *)stack_malloc(sizeof(long)*d_K);
  for(i=0;i<d_K;i++) e_chi[i]=umodiu(gel(TR, Fl_mul(i, j0, d_K)), dp);
  g_el = pgener_Fl(el);
  z_f = Fl_powu(g_el, el1f, el);
  vz_f = Fl_powers(z_f, f-1, el);

  for (gi = 1, i = 0; i < d_K; i++)
  {
    ulong x = 1;
    for (j = 1; j <= h; j++)
    {
      ulong y = Fl_mul(H[j], gi, f);
      x = Fl_mul(x, vz_f[1+y]-1, el); /* vz_f[1+y] = z_f^y  */
    }
    gi = Fl_mul(gi, g_K, f);
    xi_el[1+i] = x;  /* xi_el[1+i] = xi^{g_K^i} mod el */
  }
  for(i=0;i<d_K;i++)
  {
    ulong x = 1;
    for(j=0;j<d_K;j++)
      x = Fl_mul(x, Fl_powu(xi_el[1+j], e_chi[(i+j)%d_K], el), el);
    if ((x = Fl_powu(x, el1dp, el))!=1) flag = 1;
    if (Fl_powu(x, p, el)!=1) {avma = av;return 0;}
  }
  avma = av;
  return flag?1:0;
}

/*
*  theta = Tr_{Q(zeta_f)/K}(zeta_f)
*  computes all conjugates of theta
*/
static GEN
theta_f_approx(GEN K, long prec)
{
  pari_sp av = avma;
  GEN H1 = gel(K, 3), H1data = gel(H1, 2), H = gel(H1, 3);
  long d_K = H1data[1], f = H1data[2], h = H1data[3], g_K = H1data[5];
  GEN roots = cgetg(d_K+1, t_VEC), Z = grootsof1(f, prec);
  long i, k, h2 = h>>1;

  for(k=0;k<d_K;k++)
  {
    long gk = Fl_powu(g_K, k, f);
    GEN z = gen_0;
    for(i=1;i<=h2;i++)
    {
      long y = Fl_mul(H[i], gk, f);
      z = gadd(z, gmael(Z, 1+y, 1));
    }
    gel(roots, 1+k) = gmul2n(z, 1);
  }
  return gerepilecopy(av, roots);
}

static long
check_prec(GEN T)
{
  long i, l=lg(T);
  for(i=2;i<l;i++)
  {
    GEN z = gel(T, i);
    if(typ(z) == t_INT) continue;
    if(expo(z)>(lg(z)-2)*BITS_IN_LONG) return 0; /* insufficient precision */
    if(300+expo(subri(z, roundr(z)))>0) return 0; /* err > 1/10^100 */
  }
  return 1;
}

static GEN
polish_roots(GEN Roots, GEN T, long prec)
{
  pari_sp av = avma;
  long i, j, n, oldprec = lg(gel(Roots, 1)), d=degpol(T);
  GEN Roots2 = cgetg(d+1, t_VEC), T2 = RgX_deriv(T);

  for(n=1;oldprec<prec;n++) oldprec<<=1;
  for(i=1;i<=d;i++)
  {
    GEN z = rtor(gel(Roots, i), prec);
    for(j=0;j<n;j++) z = subrr(z, divrr(poleval(T, z), poleval(T2, z)));
    gel(Roots2, i) = z;
  }
  return gerepilecopy(av, Roots2);
}

/*
*  For a cyclic field K contained in Q(zeta_f),
*  computes minimal polynomial T of theta=Tr_{Q(zeta_f)/K}(zeta_f) over Q
*  and conjugates of theta.
*
*  Perhaps same as galoissubcyclo, but we need conjugates of theta with
*  specified galois action.
*/
static GEN
minpol_theta(GEN K)
{
  pari_sp av = avma;
  long i, l, prec;GEN Roots, Roots2, T, T2;
  pari_timer ti;

  if (DEBUGLEVEL >= 6) timer_start(&ti);
#ifdef USE_GALOISSUBCYCLO
  GEN H1 = gel(K, 3), HH = gmael(H1,1,1);
  T2 = galoissubcyclo(utoi(f), HH, 0, 0);
  prec = 10;
  Roots = theta_f_approx(K, prec);
#else
  for(prec=10;prec;prec<<=1)
  {
    Roots = theta_f_approx(K, prec);
    T = roots_to_pol_r1(Roots, 0, lg(Roots)-1);
    if(check_prec(T)) break;
  }
  l = lg(T);
  T2 = cgetg(l, t_POL);T2[1] = evalsigne(1) | evalvarn(0);
  for(i=2;i<l;i++) gel(T2, i) = mpround(gel(T, i));
#endif
  prec<<=1;
  Roots2 = polish_roots(Roots, T2, prec);
  if (DEBUGLEVEL >= 6) timer_printf(&ti, "minpol_theta:prec=%ld",prec);
  return gerepilecopy(av, mkvec3(utoi(prec), Roots2, T2));
}

/* extracted from mftrace.c */
static GEN
const_mat(long n, GEN x)
{
  long j, l = n+1;
  GEN A = cgetg(l, t_MAT);
  for (j = 1; j < l; j++) gel(A, j) = const_col(n, x);
  return A;
}

static GEN
mat_basis(GEN B)
{
  long i, j, n=lg(B)-1;GEN A = const_mat(n, gen_0);
  for(j=1;j<=n;j++)
  {
    GEN col = gel(A, j), pol = gel(B, j);
    long d=degpol(pol);
    for(i=0;i<=d;i++) gel(col, 1+i) = gel(pol, 2+i);
  }
  return A;
}

static GEN
xi_f_approx(GEN K, long prec)
{
  pari_sp av = avma;
  GEN H1 = gel(K, 3), H1data = gel(H1, 2), H = gel(H1, 3);
  long d_K = H1data[1], f = H1data[2], h = H1data[3], g_K = H1data[5];
  GEN x = cgetg(d_K+1, t_COL), Z = grootsof1(f, prec);
  long i, k, h2 = h>>1;
  for(k=0;k<d_K;k++)
  {
    pari_sp av = avma;
    long gk = Fl_powu(g_K, k, f);
    GEN z = utor(1, prec);
    for(i=1;i<=h2;i++)
    {
      long y = Fl_mul(H[i], gk, f);
      z = mulrr(z, shiftr(subsr(1, gmael(Z, 1+y, 1)), 1));
    }
    gel(x, 1+k) = gerepilecopy(av, z);
  }
  return gerepilecopy(av, x);
}

static long
xi_f_approx_pre(GEN K)
{
  pari_sp av = avma;
  long prec = 3;
  GEN H1 = gel(K, 3), H1data = gel(H1, 2), H = gel(H1, 3);
  long f = H1data[2], h = H1data[3];
  GEN Z = grootsof1(f, prec), z = utor(1, prec);
  long e, i, h2 = h>>1;

  for(i=1;i<=h2;i++)
    z = mulrr(z, shiftr(subsr(1, gmael(Z, 1+H[i], 1)), 1));
  z = absr(z);
  if(cmprs(z, 1)<0) z = invr(z);
  e = (3*expo(z))/BITS_IN_LONG;
  avma = av;
  return (e<LOWDEFAULTPREC)?LOWDEFAULTPREC:e;
}

static GEN
vec_to_col(GEN z)
{
  z[0] = evallg(lg(z)) | evaltyp(t_COL);
  return z;
}

/*
*  K is a cyclic field of conductor f with degree d=d_K
*  xi = Norm_{Q(zeta_f)/K}(1-zeta_f)
*  1: T, min poly of a=Tr_{Q(zeta_f)/K}(zeta_f) over Q
*  2: B, power basis of K with respect to a
*  3: A, rational matrix s.t. t(v_1,...v_d) = A*t(1,a,...,a^{d-1})
*  4: Xi, integer matrix s.t. t(xi^(1),...,xi^(d)) = Xi*t(v_1,...,v_d)
*/
static GEN
xi_data_basis(GEN K, long flag)
{
  pari_sp av = avma;
  GEN H1data = gmael(K, 3, 2);long d_K = H1data[1];
  GEN T0 = minpol_theta(K), Roots = gel(T0, 2), T = gel(T0, 3);
  GEN A, M, R = cgetg(d_K+1, t_MAT), Xi = cgetg(d_K+1, t_MAT);
  GEN V, x, y, z = cgetg(d_K+1, t_VEC);
  long i, j, prec, newprec, oldprec = itou(gel(T0, 1));

  if(flag&DEBUG_BIT) err_printf("T[x]=%Ps\n", T);
#ifdef USE_LLL_BASIS
  GEN B = gel(nfinit(T, DEFAULTPREC), 7);  /* B/gmael(B,1,2) is a real basis */
#else
  GEN D, B = nfbasis(T, &D, NULL);
#endif
  A = mat_basis(B);
  M = gmael(A, 1, 1);
  if(!equali1(M)) A = RgM_Rg_div(A, M);
  if(flag&DEBUG_BIT) err_printf("B=%Ps\nA=%Ps\n", B, A);
  newprec = xi_f_approx_pre(K);
  if(oldprec<newprec)
  {
    Roots = polish_roots(Roots, T, newprec);
    prec = newprec;
  }
  else
    prec = oldprec;
  for(i=1;i<=d_K;i++) gel(R, i) = vec_to_col(gpowers(gel(Roots, i), d_K-1));
  V = RgM_inv(RgM_mul(gtrans(R), A));
  x = xi_f_approx(K, prec);
  for(i=0;i<d_K;i++)
  {
    for(j=0;j<d_K;j++) gel(z, 1+j) = gel(x, 1+(i+j)%d_K);
    y = RgM_RgC_mul(V, z);
    gel(Xi, 1+i) = cgetg(d_K+1, t_COL);
    for(j=1;j<=d_K;j++) gmael(Xi, 1+i, j) = roundr(gel(y, j));
  }
  if(flag&DEBUG_BIT) err_printf("Xi[1]=%Ps\n", gel(Xi, 1));
  A = gtrans(A);Xi = gtrans(Xi);
  return gerepilecopy(av, mkvec4(T, B, A, Xi));
}

/*
*  When factorization of polcyclo mod el is difficult, one can try to
*  check the condition of el using an integral basis of K.
*  This is useful when d_K is small.
*/
static long
chk_el_real_basis(GEN K, long p, long d_pow, long el, long j0, long flag)
{
  pari_sp av = avma;
  GEN xi = gel(K, 7), T = gel(xi, 1), A = gel(xi, 3), Xi = gel(xi, 4);
  GEN TR = gel(K, 1), H1 = gel(K, 3), H1data = gel(H1, 2);
  long d_K = H1data[1], d = upowuu(p, d_pow), dp = d*p, el1dp = (el-1)/dp;
  GEN F, v_el, xi_el, xi_e_chi = cgetg(d_K+1, t_VECSMALL);
  long i, j, root_el, found=0, *e_chi=(long *)stack_malloc(sizeof(long)*d_K);

  A = RgM_to_Flm_safe(A, el); /* RgM_to_Flm causes Fl_inv error if failed */
  if(A==NULL) {avma = av;return 0;}
  F = Flx_factor(ZX_to_Flx(T, el), el);
  root_el = Fl_neg(gmael(F, 1, 1)[2], el);
  v_el = Flm_Flc_mul(A, Fl_powers(root_el, d_K-1, el), el);
  xi_el = Flm_Flc_mul(RgM_to_Flm(Xi, el), v_el, el);
  if(flag&DEBUG_BIT) err_printf("el=%ld xi_el=%Ps\n", el, xi_el);
  for(i=0;i<d_K;i++) e_chi[i]=umodiu(gel(TR, Fl_mul(i, j0, d_K)), dp);
  for(i=0;i<d_K;i++)
  {
    long z = 1;
    for(j=0;j<d_K;j++)
      z = Fl_mul(z, Fl_powu(xi_el[1+j], e_chi[(i+j)%d_K], el), el);
    xi_e_chi[1+i] = z;
  }
  if(flag&DEBUG_BIT) err_printf("xi_e_chi=%Ps\n", xi_e_chi);
  for(i=1;i<=d_K;i++)
  {
    long x = Fl_powu(xi_e_chi[i], el1dp, el);
    if (x!=1) found = 1;
    if (Fl_powu(x, p, el)!=1) {avma = av;return 0;}
  }
  avma = av;
  return found ? 1 : 0;
}

static GEN
minpol_xi(GEN K)
{
  pari_sp av = avma;
  GEN H1 = gel(K, 3), H1data = gel(H1, 2);
  long d = H1data[1], l = d+3, i, prec;
  GEN Roots, Roots2, T, T2;
  pari_timer ti;

  if (DEBUGLEVEL >= 6) timer_start(&ti);
  /*
  for(prec=xi_f_approx_pre(K);prec;prec<<=1)
  */
  for(prec=10;prec;prec<<=1)
  {
    Roots = xi_f_approx(K, prec);
    T = roots_to_pol_r1(Roots, 0, lg(Roots)-1);
    if(check_prec(T)) break;
  }
  T2 = cgetg(l, t_POL);T2[1] = evalsigne(1) | evalvarn(0);
  for(i=2;i<l;i++) gel(T2, i) = mpround(gel(T, i));
  prec<<=1;
  Roots2 = polish_roots(Roots, T2, prec);
  if (DEBUGLEVEL >= 6) timer_printf(&ti, "minpol_xi:prec=%ld",prec);
  return gerepilecopy(av, mkvec3(utoi(prec), Roots2, T2));
}

/*
*  G = H_1*H_2*...*H_m is cyclic of order n, H_i=<perms[i]>
*  G is not necessarily a direct product.
*  If p^e || n, then p^e || |H_i| for some i.
*  return a generator of G.
*/
static GEN
find_gen(GEN perms, long n)
{
  long i, j, l = lg(perms);
  GEN fa = factoru(n), P = gel(fa, 1), E = gel(fa, 2);
  long r = lg(P);
  GEN gen = cgetg(r, t_VEC), orders = cgetg(l, t_VECSMALL), perm;

  for(i=1;i<l;i++) orders[i] = perm_order(gel(perms, i));
  for(i=1;i<r;i++)
  {
    long pe = upowuu(P[i], E[i]);
    for(j=1;j<l;j++) if(orders[j]%pe==0) break;
    gel(gen, i) = perm_pow(gel(perms, j), orders[j]/pe);
  }
  perm = gel(gen, 1);
  for(i=2;i<l;i++) perm = perm_mul(perm, gel(gen, i));
  return perm;
}

static long
find_conj_el(GEN K, GEN pol)
{
  pari_sp av = avma;
  GEN H1 = gel(K, 3), H1data = gel(H1, 2), H = gel(H1, 3);
  long d_K = H1data[1], f = H1data[2], h = H1data[3], g = H1data[5];
  long j, k, el, g_el, z_f, xi = 1, xi_g = 1;
  GEN T = NULL, vz_f;

  for(el=f+1;el;el+=f)
    if(uisprime(el) && (T = RgX_to_Flx_safe(pol, el))) break;
  g_el = pgener_Fl(el);
  z_f = Fl_powu(g_el, (el-1)/f, el);
  vz_f = Fl_powers(z_f, f-1, el);
  for(j=1;j<=h;j++)
    xi = Fl_mul(xi, vz_f[1+H[j]]-1, el);
  for(j=1;j<=h;j++)
    xi_g = Fl_mul(xi_g, vz_f[1+Fl_mul(H[j], g, f)]-1, el);
  for(k=1;k<=d_K;k++)
  {
    xi = Flx_eval(T, xi, el);
    if(xi == xi_g) break;
  }
  if(xi != xi_g) pari_err(e_MISC,"find_conj_el");
  avma = av;
  return k;
}

static long
find_conj_approx(GEN R, GEN T, GEN pol, long d_K)
{
  pari_sp av = avma;
  long i, min_i = -1, prec, prec_xi = lg(gel(R, 1));
  GEN R2, x1, x2, min;

  prec = (expo(gsupnorm(R, 0))*d_K*1.2)/BITS_IN_LONG;
  if(prec_xi < prec) R = polish_roots(R, T, prec);
  R2 = sort(R);
  x1 = gel(R, 1); x2 = gel(R, 2);min = itor(gen_1, lg(gel(R, 1)));
  for(i=1;i<=d_K;i++)
  {
    GEN z;
    x1 = RgX_Rg_eval_bk(pol, x1);
    x1 = nearr(R2, x1);
    z = absr(subrr(x1, x2));
    if(cmprr(z, min)<0) {min_i = i;min = z;}
  }
  if(min_i < 0) pari_err(e_MISC,"find_conj_approx");
  avma = av;
  return min_i;
}

/*
*  R is the roots of T. R[1+i] = R[1]^(g_K^i), 0 <= i <= d_K-1
*  1: min poly T of xi over Q
*  2: F(x)\in Q[x] s.t. xi^(g_K)=F(xi)
*/
static GEN
xi_data_galois(GEN K, long flag)
{
  pari_sp av = avma;
  const long MAX_F = 100000;
  GEN H1data = gmael(K, 3, 2);
  GEN P = minpol_xi(K), R = gel(P, 2), T = gel(P, 3);
  GEN G, perms, perm, pol, pol2;
  long k, d_K = H1data[1], f = H1data[2], prec_xi = itou(gel(P, 1));
  pari_timer ti;

  if(flag&DEBUG_BIT) err_printf("prec=%ld T[x]=%Ps\n", prec_xi, T);
  if (DEBUGLEVEL >= 6) timer_start(&ti);
  G = galoisinit(T, NULL);
  if (DEBUGLEVEL >= 6) timer_printf(&ti, "galoisinit");
  perms = gal_get_gen(G);
  perm = (lg(perms)==2)?gel(perms, 1):find_gen(perms, d_K);
  if (DEBUGLEVEL >= 6) timer_start(&ti);
  pol = galoispermtopol(G, perm);
  if (DEBUGLEVEL >= 6) timer_printf(&ti, "galoispermtopol");
  if (DEBUGLEVEL >= 6) timer_start(&ti);
  k = (f<MAX_F)?find_conj_el(K, pol):find_conj_approx(R, T, pol, d_K);
  if (DEBUGLEVEL >= 6) timer_printf(&ti,"find_conj_%s",(f<MAX_F)?"el":"approx");
  pol2 = galoispermtopol(G, perm_pow(perm, k));
  if(flag&DEBUG_BIT) err_printf("F[x]=%Ps\n",pol2);
  return gerepilecopy(av, mkvec2(T, pol2));
}

/* 
*  If g(X)\in Q[X] s.t. g(xi)=xi^{g_K} was found,
*  then we fix an integer x_0 s.t. xi=x_0 (mod el) and construct x_i
*  s.t. xi^{g_K^i}=x_i (mod el) using g(X).
*/
static long
chk_el_real_galois(GEN K, long p, long d_pow, long el, long j0, long flag)
{
  pari_sp av = avma;
  GEN xi = gel(K, 7), T = gel(xi, 1), F = gel(xi, 2);
  GEN TR = gel(K, 1), H1 = gel(K, 3), H1data = gel(H1, 2);
  long d_K = H1data[1], d = upowuu(p, d_pow), dp = d*p, el1dp = (el-1)/dp;
  GEN xi_el = cgetg(d_K+1, t_VECSMALL), xi_e_chi = cgetg(d_K+1, t_VECSMALL);
  GEN Fel, fa;
  long i, j, x, found=0, *e_chi=(long *)stack_malloc(sizeof(long)*d_K);

  Fel = RgX_to_Flx_safe(F, el);
  if(Fel==NULL) {avma = av;return 0;}
  fa = Flx_factor(ZX_to_Flx(T, el), el);
  x = Fl_neg(gmael(fa, 1, 1)[2], el);
  for(i=1;i<=d_K;i++)
  {
    xi_el[i] = x;
    x = Flx_eval(Fel, x, el);
  }
  if(flag&DEBUG_BIT) err_printf("el=%ld xi_el=%Ps\n", el, xi_el);
  for(i=0;i<d_K;i++) e_chi[i]=umodiu(gel(TR, Fl_mul(i, j0, d_K)), dp);
  for(i=0;i<d_K;i++)
  {
    long z = 1;
    for(j=0;j<d_K;j++)
      z = Fl_mul(z, Fl_powu(xi_el[1+j], e_chi[(i+j)%d_K], el), el);
    xi_e_chi[1+i] = z;
  }
  if(flag&DEBUG_BIT) err_printf("xi_e_chi=%Ps\n", xi_e_chi);
  for(i=1;i<=d_K;i++)
  {
    long x = Fl_powu(xi_e_chi[i], el1dp, el);
    if (x!=1) found = 1;
    if (Fl_powu(x, p, el)!=1) {avma = av;return 0;}
  }
  avma = av;
  return found ? 1 : 0;
}

/*
*  checks the condition of el using the irreducible
*  polynomial G_K(X) of zeta_f over K.
*  G_K(X) mod el is enough for our purpose and it is obtained by factoring
*  polcyclo(f) mod el.
*/
static long
chk_el_real_factor(GEN K, long p, long d_pow, long el, long j0, long flag)
{
  pari_sp av = avma;
  GEN TR = gel(K, 1), H1 = gel(K, 3), H1data = gel(H1, 2);
  long d_K = H1data[1], f = H1data[2], h = H1data[3], g_K = H1data[5];
  GEN H = gel(H1, 3);
  long  i, j, k, d = upowuu(p, d_pow), dp = d*p, *e_chi, found = 0;
  GEN pols, coset, vn_g, polnum, xpows, G_K;
  long el1dp = (el-1)/dp, n_coset, n_g, gi;
  pari_timer ti;

  e_chi=(long *)stack_malloc(sizeof(long)*d_K);
  for(i=0;i<d_K;i++) e_chi[i]=umodiu(gel(TR, Fl_mul(i, j0, d_K)), dp);
  coset = get_coset(H, h, f, el);
  if (DEBUGLEVEL >= 6)
  {timer_start(&ti);err_printf("factoring polcyclo(%ld) (mod %ld)\n",f, el);}
  if (f<1000)
    pols = gel(Flx_factor(ZX_to_Flx(polcyclo(f, 0), el), el), 1); /* t_COL */
  else
    pols = Flx_factor_polcyclo(f, el);  /* t_VEC --> t_COL */
  if (DEBUGLEVEL >= 6) timer_printf(&ti, "Flx_factor:f=%ld",f);
  n_coset = lg(coset)-1;
  n_g = lg(pols)-1;
  vn_g = identity_perm(n_g);
  /* d_g = order_f_x(f, el%f); */  /* degree(pols[i]) */
  /* err_printf("n_coset=%ld n_g=%ld d_g=%ld\n", n_coset, n_g, d_g); */

  polnum = const_vec(d_K, NULL);
  for(i=1;i<=d_K;i++) gel(polnum, i) = const_vecsmall(n_coset, 0);
  xpows = Flx_powers(f-1, gel(pols, 1), el);
  for(gi=1,i=1;i<=d_K;i++)
  {
    for(j=1;j<=n_coset;j++)
    {
      long x, conj = Fl_mul(gi, coset[j], f);
      x = srh_pol(xpows, vn_g, pols, el, conj, f);
      gel(polnum, i)[j] = x;
    }
    gi = Fl_mul(gi, g_K, f);
  }
  G_K = const_vec(d_K, NULL);
  for(i=1;i<=d_K;i++)
  {
    GEN z = pol1_Flx(0);
    for(j=1;j<=n_coset;j++) z = Flx_mul(z, gel(pols, gel(polnum,i)[j]), el);
    gel(G_K, i) = z;
  }

  for(k=0;k<d_K;k++)
  {
    long x = 1;
    for (i = 0; i < d_K; i++)
    {
      long x0, t;
      x0 = Flx_eval(gel(G_K, 1+i), 1, el);
      t = Fl_powu(x0, e_chi[(i+k)%d_K], el);
      x = Fl_mul(x, t, el);
    }
    x = Fl_powu(x, el1dp, el);
    if (x!=1) found = 1;
    if (Fl_powu(x, p, el)!=1) {avma = av;return 0;}
  }
  avma = av;
  return found?1:0;
}

static long
chk_el_real_chi(GEN K, long p, long d_pow, long el, long j0, long flag)
{
  GEN H1data = gmael(K, 3, 2);
  long f = H1data[2];

  if(el%f==1)
    return chk_el_real_f(K, p, d_pow, el, j0); /* must be faster */
  else if(flag&USE_BASIS)
    return chk_el_real_basis(K, p, d_pow, el, j0, flag);
  else if(flag&USE_GALOIS_POL)
    return chk_el_real_galois(K, p, d_pow, el, j0, flag);
  else
    return chk_el_real_factor(K, p, d_pow, el, j0, flag);
}

static long
chk_ell_real(GEN K, long d2, GEN ell, long j0)
{
  pari_sp av = avma;
  GEN TR = gel(K, 1), H1 = gel(K, 3), H1data = gel(H1, 2), H = gel(H1, 3);
  long d_K = H1data[1], f = H1data[2], h = H1data[3], g_K = H1data[5];
  long  i, j, *e_chi, gi;
  GEN g_ell, z_f, vz_f, xi_el = cgetg(d_K+1, t_VEC);
  GEN ell1d2=diviuexact(subiu(ell, 1), d2), ell1f=diviuexact(subiu(ell, 1), f);

  e_chi=(long *)stack_malloc(sizeof(long)*d_K);
  for(i=0;i<d_K;i++) e_chi[i]=umodiu(gel(TR, Fl_mul(i, j0, d_K)), d2);
  g_ell = pgener_Fp(ell);
  z_f = Fp_pow(g_ell, ell1f, ell);
  vz_f = Fp_powers(z_f, f-1, ell);
  for(gi=1, i=0;i<d_K;i++)
  {
    GEN x = gen_1;
    for (j = 1; j <= h; j++)
    {
      ulong y = Fl_mul(H[j], gi, f);
      x = Fp_mul(x, subiu(gel(vz_f, 1+y), 1), ell); /* vz_f[1+y] = z_f^y  */
    }
    gi = Fl_mul(gi, g_K, f);
    gel(xi_el, 1+i) = x;  /* xi_el[1+i]=xi^{g_K^i} */
  }
  for(i=0;i<d_K;i++)
  {
    GEN x = gen_1;
    for(j=0;j<d_K;j++)
      x = Fp_mul(x, Fp_powu(gel(xi_el, 1+j), e_chi[(i+j)%d_K], ell), ell);
    x = Fp_pow(x, ell1d2, ell);
    if (!equali1(x)) {avma = av;return 0;}
  }
  avma = av;
  return 1;
}

static GEN
next_el_real(GEN K, long p, long d_pow, GEN elg, long j0, long flag)
{
  GEN Chi = gel(K, 5), H1 = gel(K, 3), H1data = gel(H1, 2);
  ulong f = H1data[2], h = H1data[3];
  ulong el = elg[1], d = upowuu(p, d_pow), d2 = d*d, d22 = d2<<1, d2f = d2*f;

  el = (flag&USE_F)? el+d2f : el+d22;
  if (1000<h && el<h)  /* O(el*h) --> O(el*log(el)) by FFT */
  {
    long y = (flag&USE_F)? d2f:d22, x = (h/y)*y+1;
    el = (x<h)?x+y:x;
  }
  if (flag&USE_F) goto F;  /* el=1 (mod f) */
  for(;1;el+=d22)
    if(uisprime(el) && Chi[el%f]==0 &&
       chk_el_real_chi(K, p, d_pow, el, j0, flag)) break;
  return mkvecsmall2(el, pgener_Fl(el));

F:for(;1;el+=d2f)
    if(uisprime(el) && chk_el_real_f(K, p, d_pow, el, j0)) break;
  return mkvecsmall2(el, pgener_Fl(el));
}

static GEN
next_ell_real(GEN K, GEN ellg, long d2, GEN df0l, long j0)
{
  GEN ell = gel(ellg, 1);
  for(ell = addii(ell, df0l);ell;ell = addii(ell, df0l))
    if(isprime(ell) && chk_ell_real(K, d2, ell, j0)) break;
  return mkvec2(ell, pgener_Fp(ell));
}

static void
delete_el_real(GEN K, long p, long d_pow, GEN velg,
               long n, long f, long j0, long flag)
{
  long i, l;GEN elg;
  if (flag&DEBUG_BIT) err_printf("deleting %ld ...\n", gmael(velg, n, 1));
  for(l=lg(velg)-1;l>=1;l--) if(gel(velg, l)!=NULL) break;
  elg = next_el_real(K, p, d_pow, gel(velg, l), j0, flag);
  for(i=n;i<l;i++) gel(velg, i) = gel(velg, i+1);
  gel(velg, l) = elg;
}

/* velg has n components */
static GEN
set_ell_real(GEN K, GEN velg, long n, long d_chi, long d2, long f0, long j0)
{
  long i, n_ell = n*d_chi;GEN z = const_vec(n_ell, NULL);
  GEN df0l = muluu(d2, f0), ellg = mkvec2(gen_1, gen_1);
  for(i=1;i<=n;i++) df0l = muliu(df0l, gel(velg, i)[1]);
  for(i=1;i<=n_ell;i++) ellg = gel(z, i)= next_ell_real(K, ellg, d2, df0l, j0);
  return z;
}

static GEN
G_K_vell(GEN K, GEN vellg, ulong gk)
{
  pari_sp av = avma;
  GEN H1 = gel(K, 3), H1data = gel(H1, 2), H = gel(H1, 3);
  ulong f = H1data[2], h = H1data[3];
  GEN z_f, vz_f, A, P, M, z =  cgetg(h+1, t_VEC);
  ulong i, lv = lg(vellg);

  A=cgetg(lv, t_VEC);
  P=cgetg(lv, t_VEC);
  for(i=1;i<lv;i++)
  {
    GEN ell = gmael(vellg, i, 1), g_ell = gmael(vellg, i, 2);
    GEN ell_1 = subiu(ell, 1), ell_1_f = diviuexact(ell_1, f);
    gel(A, i) = Fp_pow(g_ell, ell_1_f, ell);
    gel(P, i) = ell;
  }
  z_f = ZV_chinese(A, P, &M);
  vz_f = Fp_powers(z_f, f-1, M)+1;
  for(i=1;i<=h;i++) gel(z, i) = gel(vz_f, Fl_mul(H[i], gk, f));
  return gerepilecopy(av, FpV_roots_to_pol(z, M, 0));
}

/*
*  f=cond(K), M=product of ell in vell, G(K/Q)=<g_K>
*  G_K[1+i]=minimal polynomial of zeta_f^{g_k^i} over K mod M, 0 <= i <= d_K-1
*/
static GEN
make_G_K(GEN K, GEN vellg)
{
  GEN H1data = gmael(K, 3, 2);
  ulong d_K = H1data[1], f = H1data[2], g_K = H1data[5];
  GEN G_K = cgetg(d_K+1, t_VEC);
  ulong i, g = 1;

  for(i=0;i<d_K;i++)
  {
    gel(G_K, 1+i) = G_K_vell(K, vellg, g);
    g = Fl_mul(g, g_K, f);
  }
  return G_K;
}

static GEN
G_K_p(GEN K, GEN ellg, ulong gk)
{
  pari_sp av = avma;
  GEN H1 = gel(K, 3), H1data = gel(H1, 2), H = gel(H1, 3);
  ulong i, f = H1data[2], h = H1data[3];
  GEN ell = gel(ellg, 1), g_ell = gel(ellg, 2);
  GEN ell_1 = subiu(ell, 1), ell_1_f = diviuexact(ell_1, f);
  GEN z_f, vz_f, z = cgetg(h+1, t_VEC);

  z_f = Fp_pow(g_ell, ell_1_f, ell);
  vz_f = Fp_powers(z_f, f-1, ell)+1;
  for(i=1;i<=h;i++) gel(z, i) = gel(vz_f, Fl_mul(H[i], gk, f));
  return gerepilecopy(av, FpV_roots_to_pol(z, ell, 0));
}

static GEN
G_K_l(GEN K, GEN ellg, ulong gk)
{
  pari_sp av = avma;
  ulong ell = itou(gel(ellg, 1)), g_ell = itou(gel(ellg, 2));
  GEN H1 = gel(K, 3), H1data = gel(H1, 2), H = gel(H1, 3);
  ulong f = H1data[2], h = H1data[3];
  ulong i, z_f, ell_1 = ell-1, ell_1_f = ell_1/f;
  GEN vz_f, z = cgetg(h+1, t_VEC);

  z_f = Fl_powu(g_ell, ell_1_f, ell);
  vz_f = Fl_powers(z_f, f-1, ell)+1;
  for(i=1;i<=h;i++) z[i] = vz_f[Fl_mul(H[i], gk, f)];
  return gerepilecopy(av, Flv_roots_to_pol(z, ell, 0));
}

static GEN
vz_el_vell(GEN elg, GEN vellg, GEN *pM)
{
  long i, el = elg[1], lv = lg(vellg);
  GEN A = cgetg(lv, t_VEC), P = cgetg(lv, t_VEC), M, z_el, vz_el;

  for(i=1;i<lv;i++)
  {
    GEN ell = gmael(vellg, i, 1), g_ell = gmael(vellg, i, 2);
    GEN ell_1 = subiu(ell, 1), ell_1_el = diviuexact(ell_1, el);
    gel(A, i) = Fp_pow(g_ell, ell_1_el, ell);
    gel(P, i) = ell;
  }
  z_el = ZV_chinese(A, P, &M);*pM = M;
  vz_el = Fp_powers(z_el, el-1, M);
  return vz_el;
}

static GEN
vz_2f_vell(long f, GEN vellg, GEN *pM)
{
  ulong i, f2 = f<<1, lv = lg(vellg);
  GEN A = cgetg(lv, t_VEC), P = cgetg(lv, t_VEC), M, z_2f, vz_2f;

  for(i=1;i<lv;i++)
  {
    GEN ell = gmael(vellg, i, 1), g_ell = gmael(vellg, i, 2);
    GEN ell_1 = subiu(ell, 1);
    gel(A, i) = Fp_pow(g_ell, diviuexact(ell_1, f2), ell);
    gel(P, i) = ell;
  }
  z_2f = ZV_chinese(A, P, &M);*pM = M;
  vz_2f = Fp_powers(z_2f, f2-1, M);
  return vz_2f;
}

static GEN
D_xi_el_vell_FFT(GEN K, GEN elg, GEN vellg, ulong d, ulong j0, GEN vG_K)
{
  pari_sp av = avma;
  GEN TR = gel(K, 1), H1data = gmael(K, 3, 2);
  ulong d_K = H1data[1], h = H1data[3], d_chi = gel(K, 6)[1];
  ulong el = elg[1], g_el = elg[2], el_1 = el-1;
  ulong i, j, i2, k, *e_chi, dwel;
  GEN u = cgetg(el+2, t_POL) , v = cgetg(h+3, t_POL);
  GEN w = cgetg(el+1, t_VEC), ww;
  GEN M, vz_el, G_K, z = const_vec(d_chi, gen_1);

  vz_el = vz_el_vell(elg, vellg, &M);
  u[1] = evalsigne(1) | evalvarn(0);
  v[1] = evalsigne(1) | evalvarn(0);
  e_chi=(ulong *)stack_malloc(sizeof(long)*d_K);
  for(i=0;i<d_K;i++) e_chi[i]=umodiu(gel(TR, Fl_mul(i, j0, d_K)), d);

  for(i=i2=0;i<el;i++)
  {
    ulong j2 = i2?el-i2:i2; /* i2=(i*i)%el */
    gel(u, 2+i) = gel(vz_el, 1+j2);
    if ((i2+=i+i+1)>=el) i2%=el;
  }
  for(k=0;k<d_K;k++)
  {
    pari_sp av = avma;
    pari_timer ti;
    G_K = gel(vG_K, 1+k);
    for(i=i2=0;i<=h;i++)
    {
      gel(v, 2+i) = Fp_mul(gel(G_K, 2+i), gel(vz_el, 1+i2), M);
      if ((i2+=i+i+1)>=el) i2%=el;
    }
    if (DEBUGLEVEL >= 6) timer_start(&ti);
    ww = ZX_mul(u, v);
    if (DEBUGLEVEL >= 6)
      timer_printf(&ti, "ZX_mul:%ld/%ld h*el=%ld*%ld", k, d_K, h, el);
    dwel = degpol(ww)-el;
    for(i=0;i<=dwel;i++) gel(w, 1+i) = addii(gel(ww, 2+i), gel(ww, 2+i+el));
    for(;i<el;i++) gel(w, 1+i) = gel(ww, 2+i);
    for(i=i2=1;i<el;i++)  /* w[i]=G_K(z_el^(2*i)) */
    {
      gel(w, i) = Fp_mul(gel(w, 1+i), gel(vz_el, 1+i2), M);
      if ((i2+=i+i+1)>=el) i2%=el;
    }

    long gd = Fl_powu(g_el, d, el);  /* a bit faster */
    long gi = g_el;
    GEN x1 = gen_1;
    for(i=1;i<d;i++)
    {
      GEN xi = gen_1;
      long gdi = gi;
      for(j=0;i+j<el_1;j+=d)
      {
        xi = Fp_mul(xi, gel(w, (gdi+gdi)%el), M);
        gdi = Fl_mul(gdi, gd, el);
      }
      gi = Fl_mul(gi, g_el, el);
      xi = Fp_powu(xi, i, M);
      x1 = Fp_mul(x1, xi, M);
    }
    for(i=1;i<=d_chi;i++)
    {
      GEN x2 = Fp_powu(x1, e_chi[(k+i-1)%d_K], M);
      gel(z, i) = Fp_mul(gel(z, i), x2, M);
    }
    z = gerepilecopy(av, z);
  }
  return gerepilecopy(av, z);
}

static GEN
D_xi_el_vell(GEN K, GEN elg, GEN vellg, ulong d, ulong j0)
{
  pari_sp av = avma;
  GEN TR = gel(K, 1), H1 = gel(K, 3), H1data = gel(H1, 2), H = gel(H1, 3);
  ulong d_K = H1data[1], f = H1data[2], h = H1data[3], g_K = H1data[5];
  GEN z_f, z_el, vz_f, vz_el;
  ulong el = elg[1], g_el = elg[2], el_1 = el-1;
  ulong i, j, k, lv = lg(vellg), *e_chi, d_chi = gel(K, 6)[1];
  GEN A, B, P, M, z = const_vec(d_chi, gen_1);

  A=cgetg(lv, t_VEC);
  B=cgetg(lv, t_VEC);
  P=cgetg(lv, t_VEC);
  for(i=1;i<lv;i++)
  {
    GEN ell = gmael(vellg, i, 1), g_ell = gmael(vellg, i, 2);
    GEN ell_1 = subiu(ell, 1);
    GEN ell_1_f = diviuexact(ell_1, f), ell_1_el = diviuexact(ell_1, el);
    gel(A, i) = Fp_pow(g_ell, ell_1_f, ell);
    gel(B, i) = Fp_pow(g_ell, ell_1_el, ell);
    gel(P, i) = ell;
  }
  z_f = ZV_chinese(A, P, &M);
  z_el = ZV_chinese(B, P, &M);

  e_chi=(ulong *)stack_malloc(sizeof(long)*d_K);
  for(i=0;i<d_K;i++) e_chi[i] = umodiu(gel(TR, Fl_mul(i, j0, d_K)), d);
  vz_f = Fp_powers(z_f, f-1, M);
  vz_el = Fp_powers(z_el, el-1, M);
  for(k=0;k<d_K;k++)
  {
    pari_sp av = avma;
    GEN x0 = gen_1;
    long gk = Fl_powu(g_K, k, f);
    for(i=1;i<el_1;i++)
    {
      long gi = Fl_powu(g_el, i, el);
      GEN x1 = gen_1;
      GEN x2 = gel(vz_el, 1+gi);
      for(j=1;j<=h;j++)
      {
        long y = Fl_mul(H[j], gk, f);
        x1 = Fp_mul(x1, Fp_sub(x2, gel(vz_f, 1+y), M), M);
      }
      x1 = Fp_powu(x1, i%d, M);
      x0 = Fp_mul(x0, x1, M);
    }
    for(i=1;i<=d_chi;i++)
    {
      GEN x2 = Fp_powu(x0, e_chi[(k+i-1)%d_K], M);
      gel(z, i) = Fp_mul(gel(z, i), x2, M);
    }
    z = gerepilecopy(av, z);
  }
  return gerepilecopy(av, z);
}

static GEN
D_xi_el_ZX_mul(GEN K, GEN elg, GEN ellg, GEN vG_K, ulong d, ulong j0)
{
  pari_sp av = avma;
  GEN TR = gel(K, 1), H1data = gmael(K, 3, 2);
  ulong d_K = H1data[1], f = H1data[2], h = H1data[3], g_K = H1data[5];
  ulong el = elg[1], g_el = elg[2], el_1 = el-1, d_chi = gel(K, 6)[1];
  GEN ell = gel(ellg, 1), g_ell = gel(ellg, 2);
  GEN ell_1 = subiu(ell, 1), ell_1_el = diviuexact(ell_1, el);
  GEN u = cgetg(el+2, t_POL) , v = cgetg(h+3, t_POL);
  GEN w = cgetg(el+1, t_VEC), ww;
  GEN z_el, vz_el, G_K, z = const_vec(d_chi, gen_1);
  ulong i, j, i2, k, *e_chi, dwel;

  u[1] = evalsigne(1) | evalvarn(0);
  v[1] = evalsigne(1) | evalvarn(0);
  e_chi=(ulong *)stack_malloc(sizeof(long)*d_K);
  for(i=0;i<d_K;i++) e_chi[i]=umodiu(gel(TR, Fl_mul(i, j0, d_K)), d);
  z_el = Fp_pow(g_ell, ell_1_el, ell);
  vz_el = Fp_powers(z_el, el_1, ell);

  for(i=i2=0;i<el;i++)
  {
    ulong j2 = i2?el-i2:i2; /* i2=(i*i)%el */
    gel(u, 2+i) = gel(vz_el, 1+j2);
    if ((i2+=i+i+1)>=el) i2%=el;
  }
  for(k=0;k<d_K;k++)
  {
    pari_sp av = avma;
    pari_timer ti;
    long gk = Fl_powu(g_K, k, f);
    if (DEBUGLEVEL >= 6) timer_start(&ti);
    G_K = (vG_K==NULL) ? G_K_p(K, ellg, gk):RgX_to_FpX(gel(vG_K, 1+k), ell);
    if (DEBUGLEVEL >= 6) timer_printf(&ti, "G_K_p");
    for(i=i2=0;i<=h;i++)
    {
      gel(v, 2+i) = Fp_mul(gel(G_K, 2+i), gel(vz_el, 1+i2), ell);
      if ((i2+=i+i+1)>=el) i2%=el;
    }
    if (DEBUGLEVEL >= 6) timer_start(&ti);
    ww = ZX_mul(u, v);
    if (DEBUGLEVEL >= 6)
      timer_printf(&ti, "ZX_mul:%ld/%ld h*el=%ld*%ld", k, d_K, h, el);
    dwel = degpol(ww)-el;
    for(i=0;i<=dwel;i++) gel(w, 1+i) = addii(gel(ww, 2+i), gel(ww, 2+i+el));
    for(;i<el;i++) gel(w, 1+i) = gel(ww, 2+i);
    for(i=i2=1;i<el;i++)  /* w[i]=G_K(z_el^(2*i)) */
    {
      gel(w, i) = Fp_mul(gel(w, 1+i), gel(vz_el, 1+i2), ell);
      if ((i2+=i+i+1)>=el) i2%=el;
    }

#if 0
    long gi = 1;
    GEN x1 = gen_1;
    for(gi = 1, x1 = gen_1, i=1;i<el_1;i++)
    {
      gi = Fl_mul(gi, g_el, el);
      i2 = (2*gi)%el;
      x1 = Fp_mul(x1, Fp_powu(gel(w, i2), i%d, ell), ell);
    }
#endif

    long gd = Fl_powu(g_el, d, el);  /* a bit faster */
    long gi = g_el;
    GEN x1 = gen_1;
    for(i=1;i<d;i++)
    {
      GEN xi = gen_1;
      long gdi = gi;
      for(j=0;i+j<el_1;j+=d)
      {
        xi = Fp_mul(xi, gel(w, (gdi+gdi)%el), ell);
        gdi = Fl_mul(gdi, gd, el);
      }
      gi = Fl_mul(gi, g_el, el);
      xi = Fp_powu(xi, i, ell);
      x1 = Fp_mul(x1, xi, ell);
    }
    for(i=1;i<=d_chi;i++)
    {
      GEN x2 = Fp_powu(x1, e_chi[(k+i-1)%d_K], ell);
      gel(z, i) = Fp_mul(gel(z, i), x2, ell);
    }
    z = gerepilecopy(av, z);
  }
  return gerepilecopy(av, z);
}

static GEN
D_xi_el_Flx_mul(GEN K, GEN elg, GEN ellg, GEN vG_K, ulong d, ulong j0)
{
  pari_sp av = avma;
  GEN TR = gel(K, 1), H1data = gmael(K, 3, 2);
  ulong d_K = H1data[1], f = H1data[2], h = H1data[3], g_K = H1data[5];
  ulong el = elg[1], g_el = elg[2], el_1 = el-1, d_chi = gel(K, 6)[1];
  ulong ell = itou(gel(ellg, 1)), g_ell = itou(gel(ellg, 2)), ell_1 = ell-1;
  ulong ell_1_el = ell_1/el, z_el;
  GEN u = cgetg(el+2, t_VECSMALL), v = cgetg(h+3, t_VECSMALL);
  GEN w = cgetg(el+1, t_VECSMALL), ww;
  GEN vz_el, G_K, z = const_vecsmall(d_chi, 1);
  ulong i, j, i2, k, *e_chi, dwel;

  u[1] = evalsigne(1) | evalvarn(0);
  v[1] = evalsigne(1) | evalvarn(0);
  e_chi=(ulong *)stack_malloc(sizeof(long)*d_K);
  for(i=0;i<d_K;i++) e_chi[i]=umodiu(gel(TR, Fl_mul(i, j0, d_K)), d);
  z_el = Fl_powu(g_ell, ell_1_el, ell);
  vz_el = Fl_powers(z_el, el_1, ell);

  for(i=i2=0;i<el;i++)
  {
    ulong j2 = i2?el-i2:i2;
    u[2+i] = vz_el[1+j2];
    if ((i2+=i+i+1)>=el) i2%=el;  /* i2=(i*i)%el */
  }
  for(k=0;k<d_K;k++)
  {
    pari_sp av = avma;
    pari_timer ti;
    ulong gk = Fl_powu(g_K, k, f);
    if (DEBUGLEVEL >= 6) timer_start(&ti);
    G_K = (vG_K==NULL)?G_K_l(K, ellg, gk):ZX_to_Flx(gel(vG_K, 1+k), ell);
    if (DEBUGLEVEL >= 6) timer_printf(&ti, "G_K_l");
    for(i=i2=0;i<=h;i++)
    {
      v[2+i] = Fl_mul(G_K[2+i], vz_el[1+i2], ell);
      if ((i2+=i+i+1)>=el) i2%=el;  /* i2=(i*i)%el */
    }
    if (DEBUGLEVEL >= 6) timer_start(&ti);
    ww = Flx_mul(u, v, ell);
    if (DEBUGLEVEL >= 6)
      timer_printf(&ti, "Flx_mul:%ld/%ld h*el=%ld*%ld", k, d_K, h, el);
    dwel=degpol(ww)-el; /* dwel=h-1 */
    for(i=0;i<=dwel;i++) w[1+i] = Fl_add(ww[2+i], ww[2+i+el], ell);
    for(;i<el;i++) w[1+i] = ww[2+i];
    for(i=i2=1;i<el;i++)  /* w[i]=G_K(z_el^(2*i)) */
    {
      w[i] = Fl_mul(w[1+i], vz_el[1+i2], ell);
      if ((i2+=i+i+1)>=el) i2%=el;  /* i2=(i*i)%el */
    }
    long gd = Fl_powu(g_el, d, el);  /* a bit faster */
    long x1 = 1, gi = g_el;
    for(i=1;i<d;i++)
    {
      long xi = 1, gdi = gi;
      for(j=0;i+j<el_1;j+=d)
      {
        xi = Fl_mul(xi, w[(gdi+gdi)%el], ell);
        gdi = Fl_mul(gdi, gd, el);
      }
      gi = Fl_mul(gi, g_el, el);
      xi = Fl_powu(xi, i, ell);
      x1 = Fl_mul(x1, xi, ell);
    }
    for(i=1;i<=d_chi;i++)
    {
      long x2 = Fl_powu(x1, e_chi[(k+i-1)%d_K], ell);
      z[i] = Fl_mul(z[i], x2, ell);
    }
    avma = av;
  }
  return gerepilecopy(av, Flv_to_ZV(z));
}

static GEN
D_xi_el_ss(GEN K, GEN elg, GEN ellg, ulong d, ulong j0)
{
  pari_sp av = avma;
  GEN TR = gel(K, 1), H1 = gel(K, 3), H1data = gel(H1, 2), H = gel(H1, 3);
  ulong d_K = H1data[1], f = H1data[2], h = H1data[3], g_K = H1data[5];
  ulong el = elg[1], g_el = elg[2], el_1 = el-1;
  ulong ell = itou(gel(ellg, 1)), g_ell = itou(gel(ellg, 2)), ell_1 = ell-1;
  ulong ell_1_f = ell_1/f, ell_1_el = ell_1/el;
  ulong i, j, k, *e_chi, z_f, z_el, d_chi = gel(K, 6)[1];
  GEN vz_f, vz_el, z = const_vecsmall(d_chi, 1);

  e_chi=(ulong *)stack_malloc(sizeof(long)*d_K);
  for(i=0;i<d_K;i++) e_chi[i] = umodiu(gel(TR, Fl_mul(i, j0, d_K)), d);
  z_f = Fl_powu(g_ell, ell_1_f, ell);
  z_el = Fl_powu(g_ell, ell_1_el, ell);
  vz_f = Fl_powers(z_f, f-1, ell);
  vz_el = Fl_powers(z_el, el-1, ell);
  for(k=0;k<d_K;k++)
  {
    pari_sp av = avma;
    long x0 = 1;
    long gk = Fl_powu(g_K, k, f);
    for(i=1;i<el_1;i++)
    {
      long gi = Fl_powu(g_el, i, el);
      long x1 = 1;
      long x2 = vz_el[1+gi];
      for(j=1;j<=h;j++)
      {
        long y = Fl_mul(H[j], gk, f);
        x1 = Fl_mul(x1, Fl_sub(x2, vz_f[1+y], ell), ell);
      }
      x1 = Fl_powu(x1, i%d, ell);
      x0 = Fl_mul(x0, x1, ell);
    }
    for(i=1;i<=d_chi;i++)
    {
      long x2 = Fl_powu(x0, e_chi[(k+i-1)%d_K], ell);
      z[i] = Fl_mul(z[i], x2, ell);
    }
    avma = av;
  }
  return gerepilecopy(av, Flv_to_ZV(z));
}

static GEN
D_xi_el_sl(GEN K, GEN elg, GEN ellg, ulong d, ulong j0)
{
  pari_sp av = avma;
  GEN TR = gel(K, 1), H1 = gel(K, 3), H1data = gel(H1, 2), H = gel(H1, 3);
  ulong d_K = H1data[1], f = H1data[2], h = H1data[3], g_K = H1data[5];
  ulong el = elg[1], g_el = elg[2], el_1 = el-1, d_chi = gel(K, 6)[1];
  GEN ell = gel(ellg, 1), g_ell = gel(ellg, 2), ell_1 = subiu(ell, 1);
  GEN ell_1_f = diviuexact(ell_1, f), ell_1_el = diviuexact(ell_1, el);
  GEN z_f, z_el, vz_f, vz_el, z = const_vec(d_chi, gen_1);
  ulong i, j, k, *e_chi;

  e_chi=(ulong *)stack_malloc(sizeof(long)*d_K);
  for(i=0;i<d_K;i++) e_chi[i] = umodiu(gel(TR, Fl_mul(i, j0, d_K)), d);
  z_f = Fp_pow(g_ell, ell_1_f, ell);
  z_el = Fp_pow(g_ell, ell_1_el, ell);
  vz_f = Fp_powers(z_f, f-1, ell);
  vz_el = Fp_powers(z_el, el-1, ell);
  for(k=0;k<d_K;k++)
  {
    pari_sp av = avma;
    GEN x0 = gen_1;
    long gk = Fl_powu(g_K, k, f);
    for(i=1;i<el_1;i++)
    {
      pari_sp av = avma;
      long gi = Fl_powu(g_el, i, el);
      GEN x1 = gen_1;
      GEN x2 = gel(vz_el, 1+gi);
      for(j=1;j<=h;j++)
      {
        long y = Fl_neg(Fl_mul(H[j], gk, f), f);
        x1 = Fp_mul(x1, Fp_sub(x2, gel(vz_f, 1+y), ell), ell);
      }
      avma = av;
      x1 = Fp_powu(x1, i%d, ell);
      x0 = Fp_mul(x0, x1, ell);
    }
    for(i=1;i<=d_chi;i++)
    {
      GEN x2 = Fp_powu(x0, e_chi[(k+i-1)%d_K], ell);
      gel(z, i) = Fp_mul(gel(z, i), x2, ell);
    }
    z = gerepilecopy(av, z);
  }
  return gerepilecopy(av, z);
}

static long
get_y(GEN z, GEN ellg, long d)
{
  GEN ell = gel(ellg, 1), g_ell = gel(ellg, 2);
  GEN ell_1 = subiu(ell, 1), elld = diviuexact(ell_1, d);
  GEN g_elld = Fp_pow(g_ell, elld, ell);
  GEN x = Fp_pow(modii(z, ell), elld, ell);
  long k;
  for(k=0;k<d;k++)
  {
    if(equali1(x)) break;
    x = Fp_mul(x, g_elld, ell);
  }
  if(k==0) k=d;
  else if(d<=k) pari_err_BUG("pclgpsubcyclo [MLL]");
  return k;
}

static void
real_MLLn(long *y, GEN K, ulong p, ulong d_pow, ulong n,
         GEN velg, GEN vellg, GEN vG_K, ulong j0, long flag)
{
  pari_sp av = avma;
  ulong i, j, k, d = upowuu(p, d_pow), h = gmael(K, 3, 2)[3];
  ulong row = lg(vellg)-1;
  for(i=1;i<=n;i++)
  {
    pari_sp av = avma;
    GEN elg = gel(velg, i), z;
    ulong el = elg[1], lz;
    pari_timer ti;
    if (DEBUGLEVEL >= 6) timer_start(&ti);
    if(flag&DEBUG1)
      z = D_xi_el_vell(K, elg, vellg, d, j0);
    else
      z = (h<el) ? D_xi_el_vell_FFT(K, elg, vellg, d, j0, vG_K)
                 : D_xi_el_vell(K, elg, vellg, d, j0);
    if (DEBUGLEVEL >= 6) timer_printf(&ti, "pclgpsubcyclo:[D_xi_el]");
    if(flag&DEBUG_BIT) err_printf("z=%Ps\n", z); 
    lz = lg(z)-1;
    for(k=1;k<=lz;k++)
    {
      for(j=1;j<=row;j++)
        y[(j-1)*row+(i-1)*lz+k-1] = get_y(gel(z, k), gel(vellg, j), d);
    }
    avma = av;
  }
  avma = av;
  return;
}

static void
real_MLL1(long *y, GEN K, ulong p, ulong d_pow,
         GEN velg, GEN vellg, ulong j0, long flag)
{
  pari_sp av = avma;
  ulong h = gmael(K, 3, 2)[3], d = upowuu(p, d_pow);
  GEN elg = gel(velg, 1), ellg = gel(vellg, 1), ell = gel(ellg, 1), z;
  ulong el = elg[1];
  pari_timer ti;

  if (DEBUGLEVEL >= 6) timer_start(&ti);
  if(flag&DEBUG1)
    z = (lgefint(ell)<=3)?D_xi_el_ss(K, elg, ellg, d, j0)
                         :D_xi_el_sl(K, elg, ellg, d, j0);
  else
  {
    if (h<el)
      z = (lgefint(ell)<=3)?D_xi_el_Flx_mul(K, elg, ellg, NULL, d, j0)
                           :D_xi_el_ZX_mul(K, elg, ellg, NULL, d, j0);
    else
      z = (lgefint(ell)<=3)?D_xi_el_ss(K, elg, ellg, d, j0)
                           :D_xi_el_sl(K, elg, ellg, d, j0);
  }
  if (DEBUGLEVEL >= 6) timer_printf(&ti, "pclgpsubcyclo:[D_xi_el]");
  if(flag&DEBUG_BIT) err_printf("z=%Ps\n", z); 
  y[0] = get_y(gel(z, 1), ellg, d);
  avma = av;
  return;
}

static void
real_MLL(long *y, GEN K, long p, long d_pow, long n,
         GEN velg, GEN vellg, GEN vG_K, long j0, long flag)
{
  pari_sp av = avma;
  long i, j, k, h = gmael(K, 3, 2)[3], d = upowuu(p, d_pow), row = lg(vellg)-1;
  for(j=1;j<=row;j++)
  {
    GEN ellg = gel(vellg, j), ell = gel(ellg, 1);
    for(i=1;i<=n;i++)
    {
      pari_sp av = avma;
      GEN elg = gel(velg, i), z;
      ulong el = elg[1], lz;
      pari_timer ti;
      if (DEBUGLEVEL >= 6) timer_start(&ti);
      if(flag&DEBUG1)
      {
        z = (lgefint(ell)<=3)?D_xi_el_ss(K, elg, ellg, d, j0)
                             :D_xi_el_sl(K, elg, ellg, d, j0);
      }
      else
      {
        if (h<el)
        {
          z = (lgefint(ell)<=3)?D_xi_el_Flx_mul(K, elg, ellg, vG_K, d, j0)
                               :D_xi_el_ZX_mul(K, elg, ellg, vG_K, d, j0);
        }
        else
        {
          z = (lgefint(ell)<=3)?D_xi_el_ss(K, elg, ellg, d, j0)
                               :D_xi_el_sl(K, elg, ellg, d, j0);
        }
      }
      if (DEBUGLEVEL >= 6) timer_printf(&ti, "pclgpsubcyclo:[D_xi_el]");
      if(flag&DEBUG_BIT) err_printf("z=%Ps\n", z); 
      lz = lg(z)-1;
      for(k=1;k<=lz;k++)
        y[(j-1)*row+(i-1)*lz+k-1] = get_y(gel(z, k), ellg, d);
      avma = av;
    }
  }
  avma = av;
  return;
}

static long
use_basis(long d_K, long f)
{
  if(d_K<=10 || (d_K<=30 && f<=5000)) return 1;
  else return 0;
}

static long
use_factor(ulong f)
{
  GEN fa = factoru(f), P = gel(fa, 1);
  return (P[lg(P)-1]<500)?1:0;
}

static GEN
cyc_real_MLL(GEN K, long p, long d_pow, long j0, long flag)
{
  pari_sp av = avma;
  GEN H1 = gel(K, 3), H1data = gel(H1, 2);
  ulong d_K = H1data[1], f = H1data[2], d_chi = gel(K, 6)[1];
  ulong i, n, f0, n_el = d_pow, d = upowuu(p, d_pow), rank = n_el*d_chi;
  GEN velg = const_vec(n_el, NULL), vellg = NULL, vG_K, z, p0 = utoi(p);
  GEN str = nullvec(), old = mkvec2(gen_0, NULL), new = mkvec2(gen_0, NULL);
  long *y0 = pari_calloc(sizeof(long)*rank*rank);

  if (flag&DEBUG_BIT)
    err_printf("cyc_real_MLL:p=%ld d_pow=%ld deg(K)=%ld cond(K)=%ld g_K=%ld\n",
                p, d_pow, d_K, f, H1data[5]);
 
  gel(K, 5) = get_chi(H1);
  if((flag&(USE_BASIS|USE_FACTOR|USE_GALOIS_POL|USE_F))==0)
  { 
    if(f-1 <= (d_K<<1)) flag |= USE_F;
    else if(use_basis(d_K, f)) flag |= USE_BASIS;
    else if(use_factor(f)) flag |= USE_FACTOR;
    else flag |= USE_GALOIS_POL;
  }
  if(flag&USE_BASIS) K = vec_append(K, xi_data_basis(K, flag));
  else if(flag&USE_GALOIS_POL) K = vec_append(K, xi_data_galois(K, flag));
  f0 = f%p?f:f/p;
  gel(velg, 1) = next_el_real(K, p, d_pow, mkvecsmall2(1, 1), j0, flag);

  for(n=1;n<=n_el;n++) /* loop while structure is unknown */
  {
    GEN y, m;long n_ell;
    pari_timer ti;
    if (DEBUGLEVEL >= 6) timer_start(&ti);
    vellg = set_ell_real(K, velg, n, d_chi, d*d, f0, j0);
    n_ell = lg(vellg) -1; /* equal to n*d_chi */
    if (DEBUGLEVEL >= 6) timer_printf(&ti, "set_ell_real");
    if (flag&DEBUG_BIT) err_printf("vel=%Ps\nvell=%Ps\n", velg, vellg);
    if(n_ell==1)
      real_MLL1(y0, K, p, d_pow, velg, vellg, j0, flag);
    else
    {
      if (DEBUGLEVEL >= 6) timer_start(&ti);
      vG_K = make_G_K(K, vellg);
      if (DEBUGLEVEL >= 6) timer_printf(&ti, "make_G_K");
      if(lgefint(gmael(vellg, n_ell, 1))<=3 || (flag&SAVE_MEMORY))
        real_MLL(y0, K, p, d_pow, n, velg, vellg, vG_K, j0, flag);
      else
        real_MLLn(y0, K, p, d_pow, n, velg, vellg, vG_K, j0, flag);
    }
    y=ary2mat(y0, n_ell);
    if (flag & DEBUG_BIT) err_printf("y=%Ps\n", y);
    y=ZM_snf(y);
    if (flag & DEBUG_BIT) err_printf("y=%Ps\n", y);
    y=make_p_part(y, p0, d_pow);
    if (flag & DEBUG_BIT)  err_printf("y=%Ps\n", y); 
    new = structure_MLL(y, p0, d_pow);
    if (flag & DEBUG_BIT)
      err_printf("d_pow=%ld d_chi=%ld old=%Ps new=%Ps\n",d_pow,d_chi,old,new); 
    if (cmpsi(d_pow*d_chi, gel(new, 1))==0) break;
    if ((m=find_del_el(old, new, n, n_el, d_chi, flag))!=NULL)
      {delete_el_real(K, p, d_pow, velg, itou(gel(m, 1)), f, j0, flag);
       n--;new = gel(m, 2);}
    else
       gel(velg, n+1) = next_el_real(K, p, d_pow, gel(velg, n), j0, flag);
    old=new;
  }
  pari_free(y0);
  z = gel(new, 2);n = lg(vellg)-1;
  for(i=1;i<=n;i++) if (lgefint(gel(z, i))>2) str = vec_append(str, gel(z, i));
  return gerepilecopy(av, mkvec3(utoi(d_pow*d_chi), str, gen_0));
}

static GEN
cyc_buch(GEN K, GEN p0, long d_pow, long flag)
{
  GEN H1data = gmael(K, 3, 2), Kdata = gel(K, 6);
  long f = H1data[2], d_chi = Kdata[1], i, rank;
  GEN z, h, gr, D, str = nullvec();

  D = flag?stoi(-f):stoi(f);
  z = Buchquad(D, 0.0, 0.0, 0);
  h = gel(z, 1);
  gr = gel(z, 2);
  if (Z_pval(h, p0) != d_pow) pari_err_BUG("pclgpsubcclo [Buchquad]");
  rank = lg(gr)-1;
  for(i=1;i<=rank;i++)
  {
    long x;
    if ((x = Z_pval(gel(gr, i), p0))) str = vec_append(str, utoi(x));
  }
  return mkvec3(utoi(d_pow*d_chi), str, flag?gen_1:gen_0);
}

static void
verbose_output(GEN K, GEN p, long pow, long j)
{
  GEN H1 = gel(K, 3), H1data = gel(H1, 2), Kdata = gel(K, 6);
  long d = H1data[1], f = H1data[2], s = H1data[4], d_chi = Kdata[1];
  err_printf("detected |A_K_psi|=%Ps^%ld, psi=chi^%ld, d_psi=%ld, %s,\n\
  [K:Q]=%ld, [f, H]=[%ld, %Ps]\n",
  p,pow*d_chi,j,d_chi,s?"real":"imaginary",d,f,vecsmall_to_vec(gmael(H1,1,1)));
}

/* return vec[-1,[],0], vec[0,[],0], vec[1,[1],0], vec[2,[1,1],0] etc */
static GEN
cyc_real_ss(GEN K, long p, long j, long pow, long el,
            GEN vz, ulong pn, long flag)
{
  pari_sp av = avma;
  GEN TR = gel(K, 1), H1 = gel(K, 3), H1data = gel(H1, 2), H = gel(H1, 3);
  long d_K = H1data[1], f_K = H1data[2], h = H1data[3], g = H1data[5];
  long i, m, *e_chi, x = 1, d_chi = gel(K, 6)[1];

  e_chi=(long *)stack_malloc(sizeof(long)*d_K);
  for(i=0;i<d_K;i++) e_chi[i]=umodiu(gel(TR, Fl_mul(i, j, d_K)), pn);
  for (i = 0; i < d_K; i++)
  {
    ulong si = Fl_powu(g, i, f_K), x0 = 1, t;
    for (m = 1; m <= h; m++)
    {
      ulong y = Fl_mul(H[m], si, f_K);
      x0 = Fl_mul(x0, vz[y+1]-1, el); /* vz[y+1] = z^y, ord(z) = f */
    }
    t = Fl_powu(x0, e_chi[i], el);
    x = Fl_mul(x, t, el);
  }
  x = Fl_powu(x, (el-1)/pn, el);
  avma = av;
  if (x==1) return mkvec3(gen_m1, nullvec(), gen_0); /* not determined */
  else if(--pow==0) return mkvec3(gen_0, onevec(0), gen_0); /* trivial */
  if (pow && (flag&VERBOSE_BIT)) verbose_output(K, utoi(p), pow, j);
  if (flag&USE_MLL)
    return (d_K==2)?cyc_buch(K,utoi(p), pow,0):cyc_real_MLL(K, p, pow, j, flag);
  else if (pow==1) return mkvec3(utoi(d_chi), onevec(d_chi), gen_0);
  else return mkvec3(utoi(pow*d_chi), onevec(0), gen_0);
}

static GEN
cyc_real_ll(GEN K, GEN p, long j, long pow, GEN el,
            GEN vz, GEN pn, long flag)
{
  pari_sp av = avma;
  GEN TR = gel(K, 1), H1 = gel(K, 3), H1data = gel(H1, 2), H = gel(H1, 3);
  long d_K = H1data[1], f_K = H1data[2], h = H1data[3], g = H1data[5];
  long i, m, d_chi = gel(K, 6)[1];
  GEN x = gen_1;

  for (i=0; i < d_K; i++)
  {
    long si = Fl_powu(g, i, f_K);
    GEN t, x0 = gen_1;
    for (m = 1; m <= h; m++)
    {
      long y = Fl_mul(H[m], si, f_K);
      x0 = Fp_mul(x0, subiu(gel(vz, y+1), 1), el); /* x0 *= (z^y-1)%el */
    }
    t = Fp_pow(x0, gel(TR, Fl_mul(i,j,d_K)), el); /* t=(x0^T[(i*j)%d])%el */
    x = gerepileuptoleaf(av, Fp_mul(x, t, el)); /* x *= t %el */
  }
  x = Fp_pow(x, diviiexact(subiu(el, 1), pn), el); /* x = x^(el-1)/pn mod el */
  x = gerepilecopy(av, x);
  if(equali1(x)) return mkvec3(gen_m1, nullvec(), gen_0); /* not determined */
  else if(--pow==0) return mkvec3(gen_0, onevec(0), gen_0); /* trivial */
  if (pow && (flag&VERBOSE_BIT)) verbose_output(K, p, pow, j);
  if (flag&USE_MLL)
    pari_err(e_MISC,"flag=%ld for large prime is not implemented\n", USE_MLL);
  else if (pow==1) return mkvec3(utoi(d_chi), onevec(d_chi), gen_0);
  return mkvec3(utoi(pow*d_chi), onevec(0), gen_0);
}

static GEN
pclgp_cyc_real(GEN K, GEN p, long max_pow, long flag)
{
  pari_sp av = avma;
  const long NUM_EL = 20;
  GEN H1 = gel(K, 3), C = gel(K, 4), Kdata = gel(K, 6), H1data = gel(H1, 2);
  long f_K = H1data[2], n_conj = Kdata[2];
  long i, pow, n_el, n_done;
  GEN gr = nullvec(), Done = const_vecsmall(n_conj, 0);

  for (n_done = 0, pow=1; pow<=max_pow; pow++)
  {
    GEN pn = powiu(p, pow), fpn = muliu(pn, f_K), el = addiu(fpn, 1);
    for (n_el = 0; n_el < NUM_EL; el = addii(el, fpn))
    {
      ulong uel;
      GEN vz;
      if (!isprime(el)) continue;
      uel = (lgefint(el) > 3)? 0: itou(el);
      if (uel)
      { /* ord(z) = f_K */
        ulong z = Fl_powu(pgener_Fl(uel), (uel-1)/f_K, uel);
        vz = Fl_powers(z, f_K, uel);
      }
      else
      { /* ord(z) = f_K */
        GEN z = Fp_pow(pgener_Fp(el), diviuexact(subiu(el, 1), f_K), el);
        vz = Fp_powers(z, f_K, el);
      }
      n_el++;
      for (i=1; i<=n_conj; i++)
      {
        GEN z;
        if (Done[i]) continue;
        if (uel)
          z = cyc_real_ss(K, itou(p), C[i], pow, uel, vz, itou(pn), flag);
        else
          z = cyc_real_ll(K, p, C[i], pow, el, vz, pn, flag);
        if (equalim1(gel(z, 1))) continue;
        Done[i]=1;
        if (!isintzero(gel(z, 1))) gr=vec_append(gr, z);
        if (++n_done == n_conj) return gerepilecopy(av, gr);
      }
    }
  }
  pari_err_BUG("pclgp_cyc_real: max_pow is not enough");
  return gen_0; /* dummy */
}

/* return (el, g_el) */
static GEN
next_el_imag(GEN elg, long f)
{
  long el = elg[1];
  if(f&1) f<<=1;
  while(uisprime(el+=f)==0);
  return mkvecsmall2(el, pgener_Fl(el));
}

/* return (ell, g_ell) */
static GEN
next_ell_imag(GEN ellg, GEN df0l)
{
  GEN ell = gel(ellg, 1);
  while(isprime(ell = addii(ell, df0l))==0);
  return mkvec2(ell, pgener_Fp(ell));
}

static GEN
set_ell_imag(GEN velg, long n, long d_chi, GEN df0)
{
  long i, n_ell = n*d_chi;GEN z = const_vec(n_ell, NULL);
  GEN df0l = shifti(df0, 1), ellg = mkvec2(gen_1, gen_1);
  for(i=1;i<=n;i++) df0l = muliu(df0l, gel(velg, i)[1]);
  for(i=1;i<=n_ell;i++) ellg = gel(z, i)= next_ell_imag(ellg, df0l);
  return z;
}

/*
*  U(X)=u(x)+u(X)*X^f+...+f(X)*X^((m-1)f) or u(x)-u(X)*X^f+...
*  U(X)V(X)=u(X)V(X)(1+X^f+...+X^((m-1)f))
*          =w_0+w_1*X+...w_{f+el-3}*X^(f+el-3)
*  w_i (1 <= i <= f+el-2) are needed.
*  w_{f+el-2}=0 if el-1 == f.
*  W_i = w_i + w_{i+el-1} (1 <= i <= f-1).
*/
static GEN
gauss_Flx_mul(ulong f, GEN elg, GEN ellg)
{
  pari_sp av = avma;
  ulong el = elg[1], g_el= elg[2];
  ulong el_1 = el-1, f1 = f-1, f2 = f<<1, lv = el_1, lu = f, m = el_1/f;
  ulong ell = itou(gel(ellg, 1)), g_ell = itou(gel(ellg, 2)), ell_1 = ell-1;
  ulong z_2f = Fl_powu(g_ell, ell_1/f2, ell);
  ulong z_el = Fl_powu(g_ell, ell_1/el, ell);
  ulong i, i2, gi;
  GEN W = cgetg(f+1, t_VECSMALL), vz_2f, vz_el;
  GEN u = cgetg(lu+2, t_VECSMALL), v = cgetg(lv+2, t_VECSMALL), w0;

  u[1] = evalsigne(1);
  v[1] = evalsigne(1);
  vz_2f = Fl_powers(z_2f, f2-1, ell);
  vz_el = Fl_powers(z_el, el_1, ell);
  for(i=i2=0;i<lu;i++)
  {
    long j2; /* i2=(i*i)%f2, gi=g_el^i */
    j2 = i2?f2-i2:i2;
    u[2+i] = vz_2f[1+j2];
    if ((i2+=i+i+1)>=f2) i2%=f2;
  }
  for(gi=1,i=i2=0;i<lv;i++)
  {
    v[2+i] = Fl_mul(vz_2f[1+i2], vz_el[1+gi], ell);
    gi = Fl_mul(gi, g_el, el);
    if ((i2+=i+i+1)>=f2) i2%=f2;
  }
  w0 = Flx_mul(u, v, ell);
  if(m==1)
  {
    for(i=1;i<el_1;i++) W[i] = Fl_add(w0[1+i], w0[1+i+lv], ell);
    W[el_1] = w0[1+el_1];  /* el_1=1+f1 */
  }
  else
  {
    long start = 1+f, end = f+el-1;
    GEN w = cgetg(end+1, t_VECSMALL);
    for(i=1;i<end;i++) w[i] = w0[1+i];
    w[end] = 0;
    for(i=1;i<m;i++, start+=f)
    {
      w = (f & i & 1)?Flv_shift_sub(w, w0+1, ell, start, end)
                     :Flv_shift_add(w, w0+1, ell, start, end);
    }
    for(i=0;i<=f1;i++) W[1+i] = Fl_add(w[1+i], w[1+i+lv], ell);
  }
  for(i=i2=1;i<f;i++)
  {
    W[i]=Fl_mul(W[1+i], vz_2f[1+i2], ell);
    if ((i2+=i+i+1)>=f2) i2%=f2;
  }
        /* W[r]=tau_{LL}^{sigma_r}, 1<= r <= f-1 */
  return gerepilecopy(av, Flv_to_ZV(W));
}

/*
*  simple but consumes memory and slow
*/
static GEN
gauss_Flx_mul2(ulong f, GEN elg, GEN ellg)
{
  pari_sp av = avma;
  ulong el = elg[1], g_el = elg[2], el_1 = el-1;
  ulong f2 = f<<1, lv=el_1, lu=f;
  ulong ell = itou(gel(ellg, 1)), g_ell = itou(gel(ellg, 2)), ell_1 = ell-1;
  ulong z_2f = Fl_powu(g_ell, ell_1/f2, ell);
  ulong z_el = Fl_powu(g_ell, ell_1/el, ell);
  ulong i, i2, gi;
  GEN W = cgetg(el, t_VECSMALL), vz_2f, vz_el;
  GEN u = cgetg(lu+2, t_VECSMALL), v = cgetg(lv+2, t_VECSMALL), w;

  u[1] = evalsigne(1) | evalvarn(0);
  v[1] = evalsigne(1) | evalvarn(0);
  vz_2f = Fl_powers(z_2f, f2-1, ell);
  vz_el = Fl_powers(z_el, el_1, ell);
  for(gi=1,i=i2=0;i<lu;i++)
  {
    long j2; /* i2=(i*i)%f2, gi=g_el^i */
    j2 = i2?f2-i2:i2;
    u[2+i] = vz_2f[1+j2];
    v[2+i] = Fl_mul(vz_2f[1+i2], vz_el[1+gi], ell);
    gi = Fl_mul(gi, g_el, el);
    if ((i2+=i+i+1)>=f2) i2%=f2;
  }
  w = Flx_mul(u, v, ell);
  for(i=1;i<el_1;i++) W[i] = Fl_add(w[1+i], w[1+i+lv], ell);
  W[el_1] = w[1+el_1];
  for(i=i2=1;i<f;i++)
  {
    W[i]=Fl_mul(W[1+i], vz_2f[1+i2], ell);
    if ((i2+=i+i+1)>=f2) i2%=f2;
  }
    /* W[r]=tau_{LL}^{sigma_r}, 1<= r <= f-1 */
  return gerepilecopy(av, Flv_to_ZV(W));
}

static GEN
gauss_ZX_mul(ulong f, GEN elg, GEN ellg)
{
  pari_sp av = avma, btop;
  ulong el = elg[1], g_el = elg[2], el_1 = el-1;
  ulong f2 = f<<1, f1 = f-1, lv=el_1, lu=f, m=el_1/f;
  GEN ell = gel(ellg, 1), g_ell = gel(ellg, 2), ell_1 = subiu(ell, 1);
  GEN z_2f = Fp_pow(g_ell, diviuexact(ell_1, f2), ell);
  GEN z_el = Fp_pow(g_ell, diviuexact(ell_1, el), ell);
  GEN W = cgetg(f+1, t_VEC), vz_2f, vz_el;
  GEN u = cgetg(lu+2, t_POL), v = cgetg(lv+2, t_POL), w0;
  long i, i2, gi;

  u[1] = evalsigne(1) | evalvarn(0);
  v[1] = evalsigne(1) | evalvarn(0);
  vz_2f = Fp_powers(z_2f, f2-1, ell);
  vz_el = Fp_powers(z_el, el_1, ell);
  for(gi=1,i=i2=0;i<lu;i++)
  {
    long j2; /* i2=(i*i)%f2, gi=g_el^i */
    j2 = i2?f2-i2:i2;
    gel(u, 2+i) = gel(vz_2f, 1+j2);
    if ((i2+=i+i+1)>=f2) i2%=f2;
  }
  for(gi=1,i=i2=0;i<lv;i++)
  {
    gel(v, 2+i) = Fp_mul(gel(vz_2f, 1+i2), gel(vz_el, 1+gi), ell);
    gi = Fl_mul(gi, g_el, el);
    if ((i2+=i+i+1)>=f2) i2%=f2;
  }
  w0 = FpX_mul(u, v, ell);
  btop = avma;
  if(m==1)
  {
    for(i=1;i<el_1;i++) gel(W, i) = Fp_add(gel(w0, 1+i), gel(w0, 1+i+lv), ell);
    gel(W, el_1) = gel(w0, 1+el_1);
  }
  else
  {
    long start = 1+f, end = f+el-1;
    GEN w = cgetg(end+1, t_VEC);
    for(i=1;i<end;i++) gel(w, i) = gel(w0, 1+i);
    gel(w, end) = gen_0;
    for(i=1;i<m;i++, start+=f)
    {
      w = (f & i & 1)?FpV_shift_sub(w, w0+1, ell, start, end)
                     :FpV_shift_add(w, w0+1, ell, start, end);
      gerepileall(btop, 1, &w);
    }
    for(i=0;i<=f1;i++) gel(W, 1+i) = addii(gel(w, 1+i), gel(w, 1+i+lv));
  }
  for(i=i2=1;i<f;i++)
  {
    gel(W, i)=Fp_mul(gel(W, 1+i), gel(vz_2f, 1+i2), ell);
    if ((i2+=i+i+1)>=f2) i2%=f2;
  }
  return gerepilecopy(av, W);  /* W[r]=tau_{LL}^{sigma_r}, 1<= r <= f-1 */
}

static GEN
gauss_ZX_mul2(ulong f, GEN elg, GEN ellg)
{
  pari_sp av = avma;
  ulong el = elg[1], g_el = elg[2], el_1 = el-1;
  ulong lv=el_1, lu=el_1, f2 = f<<1;
  GEN ell = gel(ellg, 1), g_ell = gel(ellg, 2), ell_1 = subiu(ell, 1);
  GEN z_2f = Fp_pow(g_ell, diviuexact(ell_1, f2), ell);
  GEN z_el = Fp_pow(g_ell, diviuexact(ell_1, el), ell);
  GEN W = cgetg(el, t_VEC), vz_2f, vz_el;
  GEN u = cgetg(lu+2, t_POL), v = cgetg(lv+2, t_POL), w;
  long i, i2, gi;

  u[1] = evalsigne(1) | evalvarn(0);
  v[1] = evalsigne(1) | evalvarn(0);
  vz_2f = Fp_powers(z_2f, f2-1, ell);
  vz_el = Fp_powers(z_el, el_1, ell);
  for(gi=1,i=i2=0;i<lu;i++)
  {
    long j2; /* i2=(i*i)%f2, gi=g_el^i */
    j2 = i2?f2-i2:i2;
    gel(u, 2+i) = gel(vz_2f, 1+j2);
    gel(v, 2+i) = Fp_mul(gel(vz_2f, 1+i2), gel(vz_el, 1+gi), ell);
    gi = Fl_mul(gi, g_el, el);
    if ((i2+=i+i+1)>=f2) i2%=f2;
  }
  w = ZX_mul(u, v);
  for(i=1;i<el_1;i++) gel(W, i) = addii(gel(w, 1+i), gel(w, 1+i+lv));
  gel(W, el_1) = gel(w, 1+el_1);
  for(i=i2=1;i<f;i++)
  {
    gel(W, i)=Fp_mul(gel(W, 1+i), gel(vz_2f, 1+i2), ell);
    if ((i2+=i+i+1)>=f2) i2%=f2;
  }
  return gerepilecopy(av, W);  /* W[r]=tau_{LL}^{sigma_r}, 1<= r <= f-1 */
}

/*
*  fast but consumes memory
*/
static GEN
gauss_el_vell(ulong f, GEN elg, GEN vellg, GEN vz_2f)
{
  pari_sp av = avma;
  ulong el = elg[1], g_el = elg[2], el_1 = el-1;
  ulong lv=el_1, f2=f<<1, lu=f, m=el_1/f, f1=f-1;
  GEN W = cgetg(f+1, t_VEC), vz_el;
  GEN u = cgetg(lu+2, t_POL), v = cgetg(lv+2, t_POL), w0, M;
  ulong i, i2, gi;

  vz_el = vz_el_vell(elg, vellg, &M);
  u[1] = evalsigne(1) | evalvarn(0);
  v[1] = evalsigne(1) | evalvarn(0);
  for(i=i2=0;i<lu;i++)
  {
    long j2; /* i2=(i*i)%f2, gi=g_el^i */
    j2 = i2?f2-i2:i2;
    gel(u, 2+i) = gel(vz_2f, 1+j2);
    if ((i2+=i+i+1)>=f2) i2%=f2;
  } 
  for(gi=1,i=i2=0;i<lv;i++)
  {
    gel(v, 2+i) = Fp_mul(gel(vz_2f, 1+i2), gel(vz_el, 1+gi), M);
    gi = Fl_mul(gi, g_el, el);
    if ((i2+=i+i+1)>=f2) i2%=f2;
  }
  w0 = FpX_mul(u, v, M);
  pari_sp btop = avma;
  if(m==1)
  {
    for(i=1;i<el_1;i++) gel(W, i) = Fp_add(gel(w0, 1+i), gel(w0, 1+i+lv), M);
    gel(W, el_1) = gel(w0, 1+el_1);  /* el_1=1+f1 */
  }
  else
  {
    long start = 1+f, end = f+el-1;
    GEN w = cgetg(end+1, t_VEC);
    for(i=1;i<end;i++) gel(w, i) = gel(w0, 1+i);
    gel(w, end) = gen_0;
    for(i=1;i<m;i++, start+=f)
    {
      w = (f & i & 1)?FpV_shift_sub(w, w0+1, M, start, end)
                     :FpV_shift_add(w, w0+1, M, start, end);
      gerepileall(btop, 1, &w);
    }
    for(i=0;i<=f1;i++) gel(W, 1+i) = Fp_add(gel(w, 1+i), gel(w, 1+i+lv), M);
  }
  for(i=i2=1;i<f;i++)
  {
    gel(W, i)=Fp_mul(gel(W, 1+i), gel(vz_2f, 1+i2), M);
    if ((i2+=i+i+1)>=f2) i2%=f2;
  }
  return gerepilecopy(av, W);  /* W[r]=tau_{LL}^{sigma_r}, 1<= r <= f-1 */
}

static GEN
gauss_brute(ulong f, GEN elg, GEN ellg)
{
  pari_sp av = avma;
  ulong el = elg[1], g_el = elg[2], el_1 = el-1, i, r;
  GEN ell = gel(ellg, 1), g_ell = gel(ellg, 2), ell_1 = subiu(ell, 1);
  GEN z_2f = Fp_pow(g_ell, diviuexact(ell_1, f<<1), ell);
  GEN z_el = Fp_pow(g_ell, diviuexact(ell_1, el), ell);
  GEN z_f = Fp_sqr(z_2f, ell), w = const_vec(f-1, NULL), vg_el, vz_f, vz_el;

  vg_el = Fl_powers(g_el, el-2, el);
  vz_f = Fp_powers(z_f, f-1, ell);
  vz_el = Fp_powers(z_el, el_1, ell);
  for(r=1;r<f;r++)
  {
    pari_sp av = avma;
    GEN x = gen_0, y;
    for(i=0;i<el_1;i++)
    {
      y = Fp_mul(gel(vz_f, 1+Fl_mul(r, i, f)), gel(vz_el, 1+vg_el[1+i]), ell);
      x = Fp_add(x, y, ell);
    }
    gel(w, r) = gerepilecopy(av, x);
  }
  return gerepilecopy(av, w);  /* w[r]=tau_{LL}^{sigma_r}, 1<= r <= f-1 */
}

static GEN
norm_chi(GEN K, GEN TAU, GEN p0, long d_pow, GEN ell, long j0)
{
  pari_sp av = avma;
  GEN TR = gel(K, 1), H1 = gel(K, 3), H1data = gel(H1, 2), H = gel(H1, 3);
  long d_K = H1data[1], f_K = H1data[2], h = H1data[3], g_K = H1data[5];
  long i, j, gi, pd = upowuu(itou(p0), d_pow), *e_chi, d_chi = gel(K, 6)[1];
  GEN z = const_vec(d_chi, gen_1);

  e_chi=(long *)stack_malloc(sizeof(long)*d_K);
  for(i=0;i<d_K;i++) e_chi[i]=umodiu(gel(TR, Fl_mul(i, j0, d_K)), pd);
  for(gi=1, i=0;i<d_K;i++)
  {
    GEN y = gen_1;
    for(j=1;j<=h;j++)
      y = Fp_mul(y, gel(TAU, Fl_mul(gi, H[j], f_K)), ell);
    gi = Fl_mul(gi, g_K, f_K);
    for(j=1;j<=d_chi;j++)
    {
      GEN y2 = Fp_powu(y, e_chi[(i+j-1)%d_K], ell);
      gel(z, j) = Fp_mul(gel(z, j), y2, ell);
    }
  }
  return gerepilecopy(av, z);
}

static void
imag_MLLn(long *y, GEN K, GEN p0, long d_pow, long n,
          GEN velg, GEN vellg, long j0, long flag)
{
  pari_sp av = avma;
  GEN H1data = gmael(K, 3, 2);
  long f = H1data[2], d = upowuu(itou(p0), d_pow), row = lg(vellg)-1;
  long i, j, k, lz;
  GEN g, z, M, vz_2f = vz_2f_vell(f, vellg, &M);
  for(i=1;i<=n;i++)
  {
    pari_sp av = avma;
    GEN elg = gel(velg, i);ulong el = elg[1];
    if(flag&DEBUG_BIT) err_printf("(f,el-1)=(%ld,%ld*%ld)\n", f, (el-1)/f, f);
    g = gauss_el_vell(f, elg, vellg, vz_2f);
    z = norm_chi(K, g, p0, d_pow, M, j0);
    lz = lg(z)-1;
    for(k=1;k<=lz;k++)
    {
      for(j=1;j<=row;j++)
        y[(j-1)*row+(i-1)*lz+k-1] = get_y(gel(z, k), gel(vellg, j), d);
    }
    avma = av;
  }
  avma = av;
  return ;
}

static void
imag_MLL1(long *y, GEN K, GEN p0, long d_pow, long n,
          GEN velg, GEN vellg, long j0, long flag)
{
  pari_sp av = avma;
  GEN H1data = gmael(K, 3, 2);
  long f = H1data[2], d = upowuu(itou(p0), d_pow);
  GEN elg = gel(velg, 1), ellg = gel(vellg, 1), ell = gel(ellg, 1), g, z;
  ulong el = elg[1];

  if(flag&DEBUG_BIT) err_printf("(f,el-1)=(%ld,%ld*%ld)\n", f, (el-1)/f, f);
  g = (lgefint(ell)<=3) ? gauss_Flx_mul(f, elg, ellg)
                        : gauss_ZX_mul(f, elg, ellg);
  z = norm_chi(K, g, p0, d_pow, ell, j0);
  y[0] = get_y(gel(z, 1), ellg, d);
  avma = av;
  return ;
}

static void
imag_MLL(long *y, GEN K, GEN p0, long d_pow, long n, GEN velg, GEN vellg,
         long j0, long flag)
{
  pari_sp av = avma;
  GEN H1data = gmael(K, 3, 2);
  long i, j, f = H1data[2], d = upowuu(itou(p0), d_pow), row = lg(vellg)-1;

  for(j=1;j<=row;j++)
  {
    GEN ellg = gel(vellg, j), ell = gel(ellg, 1);
    for(i=1;i<=n;i++)
    {
      pari_sp av = avma;
      GEN elg = gel(velg, i), g, z;
      ulong el = elg[1], k, lz;
      if(flag&DEBUG_BIT) err_printf("(f,el-1)=(%ld,%ld*%ld)\n", f, (el-1)/f, f);
      if ((flag&(DEBUG1+DEBUG2+DEBUG3+DEBUG4))==0)
        g = (lgefint(ell)<=3) ? gauss_Flx_mul(f, elg, ellg)
                              : gauss_ZX_mul(f, elg, ellg);
      else
      {
        if(flag&DEBUG1)
          g = gauss_Flx_mul2(f, elg, ellg);
        else if(flag&DEBUG2)
          g = gauss_ZX_mul(f, elg, ellg);
        else if(flag&DEBUG3)
          g = gauss_ZX_mul2(f, elg, ellg);
        else
          g = gauss_brute(f, elg, ellg);
      }
      z = norm_chi(K, g, p0, d_pow, ell, j0);
      lz = lg(z)-1;
      for(k=1;k<=lz;k++)
        y[(j-1)*row+(i-1)*lz+k-1] = get_y(gel(z, k), ellg, d);
      avma = av;
    }
  }
  avma = av;
  return ;
}

static void
delete_el_imag(GEN velg, long n, long f, long flag)
{
  long i, l;GEN elg;
  if (flag&DEBUG_BIT) err_printf("deleting %ld ...\n", gmael(velg ,n, 1));
  for(l=lg(velg)-1;l>=1;l--) if(gel(velg, l)!=NULL) break;
  elg = next_el_imag(gel(velg, l), f);
  for(i=n;i<l;i++) gel(velg, i) = gel(velg, i+1);
  gel(velg, i) = elg;
}

/* return an upper bound >= 0 if one was found, otherwise return -1.
*  set chi-part to be (1) if chi is Teichmuller character.
*  B_{1,omega^(-1)} is not p-adic integer. */
static GEN
cyc_imag_MLL(GEN K, GEN p0, long d_pow, long j, long flag)
{
  pari_sp av = avma;
  GEN H1data = gmael(K, 3, 2);
  long d_K=H1data[1], f = H1data[2], d_chi = gel(K, 6)[1], p = itou(p0);
  long i, n, n_el = d_pow, d = upowuu(p, d_pow), rank = n_el*d_chi;
  GEN df0, velg = const_vec(n_el, NULL), vellg = NULL;
  GEN str = nullvec(), old = mkvec2(gen_0, NULL), new = mkvec2(gen_0, NULL);
  long *y0 = pari_calloc(sizeof(long)*rank*rank);

  if (flag&DEBUG_BIT)
  err_printf("cyc_imag_MLL:p=%ld d_pow=%ld deg(K)=%ld cond(K)=%ld avma=%ld\n",
              p, d_pow, d_K, f, avma); 

  df0 = muluu(d, f%p?f:f/p);
  gel(velg, 1) = next_el_imag(mkvecsmall2(1, 1), f);
  for(n=1;n<=n_el;n++) /* loop while structure is unknown */
  {
    pari_timer ti;
    GEN y, m;long n_ell;
    vellg = set_ell_imag(velg, n, d_chi, df0);
    n_ell = lg(vellg)-1;  /* equal to n*d_chi */
    if (flag&DEBUG_BIT) err_printf("velg=%Ps\nvellg=%Ps\n", velg, vellg);
    if (DEBUGLEVEL >= 6) timer_start(&ti);
    if(n_ell==1)
      imag_MLL1(y0, K, p0, d_pow, n, velg, vellg, j, flag);
    else if(lgefint(gmael(vellg, n, 1))<=3 || (flag&SAVE_MEMORY))
      imag_MLL(y0, K, p0, d_pow, n, velg, vellg, j, flag);
    else
      imag_MLLn(y0, K, p0, d_pow, n, velg, vellg, j, flag);
    if (DEBUGLEVEL >= 6) timer_printf(&ti, "gauss sum");
    y=ary2mat(y0, n_ell);
    if (flag&DEBUG_BIT) err_printf("y=%Ps\n", y);
    y=ZM_snf(y);
    if (flag&DEBUG_BIT) err_printf("y=%Ps\n", y);
    y=make_p_part(y, p0, d_pow);
    if (flag&DEBUG_BIT)  err_printf("y=%Ps\n", y); 
    new = structure_MLL(y, p0, d_pow);
    if (flag&DEBUG_BIT)
      err_printf("d_pow=%ld d_chi=%ld old=%Ps new=%Ps\n",d_pow,d_chi,old,new); 
    if (cmpsi(d_pow*d_chi, gel(new, 1))==0) break;
    if ((m=find_del_el(old ,new, n, n_el, d_chi, flag))!=NULL)
      {delete_el_imag(velg, itou(gel(m, 1)), f, flag);
       n--;new = gel(m, 2);}
    else
       gel(velg, n+1) = next_el_imag(gel(velg, n), f);
    old=new;
  }
  pari_free(y0);
  GEN z = gel(new, 2);n = lg(vellg)-1;
  for(i=1;i<=n;i++) if (lgefint(gel(z, i))>2) str = vec_append(str, gel(z, i));
  return gerepilecopy(av, mkvec3(utoi(d_pow*d_chi), str, gen_1));
}

/* When |A_psi|=p^e, A_psi=(p^e1,...,p^er) (psi=chi^j),
*  return vec[e, [e1, ... ,er], 1].
*  If gr str is not determined, return vec[e, [], 1].
*  If |A_chi|=1, return vec[0, [], 1].
*  If |A_chi|=p, return vec[1, [1], 1].
*  If e is not determined, return vec[-1, [], 1].
*  If psi is Teichmuller, return vec[0, [], 1].
*  B_{1,omega^(-1)} is not p-adic integer.
*/
static GEN
cyc_imag(GEN K, GEN p, long j, GEN powp, long flag)
{
  GEN MinPol = gel(K, 2), H1 = gel(K, 3), Chi = gel(K, 5);
  GEN Kdata = gel(K, 6), H1data = gel(H1, 2);
  long x, d_K = H1data[1], f_K = H1data[2], d_chi = Kdata[1];
  GEN y, B1, B2;

  if (f_K == d_K+1 && equaliu(p, f_K) && j == 1) /* Teichmuller */
     return mkvec3(gen_0, nullvec(), gen_1);
  y = ber_s2(Chi, f_K, d_K, j);
  B1 = FpX_rem(gel(y, 1), MinPol, powp);
  B2 = FpX_rem(gel(y, 2), MinPol, powp);
  if (degpol(B1)<0 || degpol(B2)<0)
    return mkvec3(gen_m1, nullvec(), gen_1);/* B=0(mod p^pow) */
  x = ZX_pval(B1, p) - ZX_pval(B2, p);
  if (x<0) pari_err_BUG("pclgpsubcyclo [Bernoulli number]");
  if (x && (flag&VERBOSE_BIT)) verbose_output(K, p, x, j);
  if(x==0) return mkvec3(gen_0, onevec(0), gen_1); /* trivial */
  else if(x==1) return mkvec3(utoi(d_chi), onevec(d_chi), gen_1);
  else if((flag&USE_MLL)==0) return mkvec3(utoi(x*d_chi), onevec(0), gen_1);
  return (d_K==2)?cyc_buch(K, p, x, 1):cyc_imag_MLL(K, p, x, j, flag);
}

/* handle representatives of all injective characters,
*  d_chi=[Q_p(zeta_d):Q_p], d=d_K */
static GEN
pclgp_cyc_imag(GEN K, GEN p, long start_pow, long max_pow, long flag)
{
  pari_sp av = avma;
  GEN C = gel(K, 4), Kdata = gel(K, 6);
  long n_conj = Kdata[2], i, pow, n_done;
  GEN gr = nullvec(), Done = const_vecsmall(n_conj, 0);

  for (n_done = 0, pow = start_pow; pow<=max_pow; pow++)
  {
    GEN powp = powiu(p, pow);
    for (i = 1; i <= n_conj; i++)
    {
      GEN z;
      if (Done[i]) continue;
      z = cyc_imag(K, p, C[i], powp, flag);
      if (equalim1(gel(z, 1))) continue;
      Done[i] = 1;
      if (!isintzero(gel(z, 1))) gr = vec_append(gr, z);
      if (++n_done == n_conj) return gerepilecopy(av, gr);
    }
  }
  pari_err_BUG("pclgp_cyc_imag: max_pow is not enough");
  return gen_0; /* dummy */
}

static GEN
gather_part(GEN g, long sgn)
{
  long i, j, l = lg(g), ord = 0, flag = 1;GEN z, z2;
  for(i=1;i<l;i++)
  {
    if( !equaliu(gmael(g, i, 3), sgn) ) continue;
    ord += itou(gmael(g, i, 1));
    if (lg(gmael(g, i, 2))==1) flag = 0;
  }
  if (flag==0 || ord==0) return mkvec2(utoi(ord), nullvec());
  z = cgetg(3,t_VEC);
  gel(z, 1) = utoi(ord);
  z2 = cgetg(l, t_VEC);
  for(i=1, j=1;i<l;i++)
    if ( equaliu(gmael(g, i, 3), sgn) ) gel(z2, j++) = gmael(g, i, 2);
  setlg(z2, j);
  z2 = shallowconcat1(z2); /* needed */
  ZV_sort_inplace(z2);
  vecreverse_inplace(z2);
  gel(z, 2) = z2;
  return z;
}

#ifdef DEBUG
static void
handling(GEN K)
{
  GEN H1 = gel(K, 3), Kdata = gel(K, 6), H1data = gel(H1, 2);
  long d_K = H1data[1], f_K = H1data[2], s_K = H1data[4], g_K = H1data[5];
  long d_chi = Kdata[1];
  err_printf("  handling %s cyclic subfield K,\
deg(K)=%ld, cond(K)=%ld g_K=%ld d_chi=%ld H=%Ps\n",
s_K? "a real": "an imaginary",d_K,f_K,g_K,d_chi,vecsmall_to_vec(gmael(H1,1,1)));
}
#endif

/* HH a t_VECSMALL listing group generators
*  Aoki and Fukuda, LNCS vol.4076 (2006), 56-74.
*/
static GEN
pclgp(GEN p0, long f, GEN HH, long flag)
{
  long OrderG, OrderH, DegF, start_pow, max_pow;
  long fH, ip, lp, i, n_f, h, gen, lgH, expG;
  GEN gG, gH, gGH, cycGH, cycG, ncycGH, vChar, P, U, Ui, vH1, z, G, vData;
  GEN vp = typ(p0) == t_INT? mkvec(p0): p0;
  /* pari_sp av = avma; */
  HH = znstar_generate(f, HH); /* group elements */
  fH = znstar_conductor(HH);
  if (fH == 1) /* znstar_reduce_modulus() does not work when fH==1 */
  {
    GEN v = cgetg(1, t_VEC);
    return typ(p0) == t_INT? v: const_vec(lg(p0)-1, v);
  }
  if (fH != f)
  { 
    HH = znstar_reduce_modulus(HH, fH); f = fH;
    err_printf("   *** We use f=%ld, H=%Ps\n", f, vecsmall_to_vec(gel(HH, 1)));
  }

  OrderH = h = zv_prod(gel(HH, 2));
  OrderG = eulerphiu(f);
  DegF = OrderG / OrderH;
#ifdef DEBUG
  {long s;
  s = srh_1(HH); /* s=0,1  s==1 <==> F real */
  err_printf("F is %s, deg(F)=%ld, ", s?"real":"imaginary", DegF);
  }
#endif
  G = znstar0(utoipos(f), 1);
  gH = gel(HH, 1); lgH = lg(gH); /* gens. of H */
  P = cgetg(lgH, t_MAT);
  for (i = 1; i < lgH; i++) gel(P, i) = Zideallog(G, utoi(gH[i]));

  /* group structure of G/H */
  cycG = znstar_get_cyc(G);
  expG = itou(gel(cycG, 1));
  /* gG generators of G, gGH generators of G/H: gGH = g.Ui, g = gGH.U */
  cycGH = ZM_snf_group(hnfmodid(P, cycG), &U, &Ui);
  ncycGH = cyc_normalize(cycGH);
  gG = ZV_to_Flv(znstar_get_gen(G), f); /* gens. of G */
  /* generators of G/H */
  gGH = FlvV_factorback(gG, ZM_to_Flm(Ui, expG), f);
  vChar = chargalois(cycGH, NULL); n_f = lg(vChar)-2;
  vH1 = cgetg(n_f+1, t_VEC);
  for (i = 1; i <= n_f; i++)
  { /* skip trivial character */
    GEN chi = gel(vChar, i+1), nchi = char_normalize(chi, ncycGH);
    GEN chiG, E, H1, C = gel(nchi, 2);
    long e, he, d = itou(gel(nchi, 1));
    /* chi(prod g[i]^e[i]) = e(sum e[i]*C[i] / d), chi has order d = #(G/H1)*/
    E = Flv_extgcd(C, d); /* \sum C[i]*E[i] = 1 in Z/dZ */

    chiG = zncharlift(chi, ncycGH, U, cycG);
    H1 =  charker(cycG, chiG); /* H1 < G with G/H1 cyclic */
    e = itou( zncharconductor(G, chiG) ); /* cond H1 = cond chi */
    H1 = FlvV_factorback(gG, ZM_to_Flm(H1, expG), e);
    gen = Flv_factorback(gGH, E, e); /* G/H1 = <gen>, chi(gen) = e(1/d) */
    H1 = znstar_generate(e, H1);
    he = eulerphiu(e) / d;
    /* G/H1 = <gen> cyclic of index d, e = cond(H1) */
    gel(vH1, i) = mkvec3(H1, mkvecsmall5(d,e,he,srh_1(H1), gen), subgp2ary(H1));
  }

#ifdef DEBUG
    err_printf("cond(F)=%ld",f);
    err_printf(", G(F/Q)=%Ps\n", cycGH);
    err_printf("F has %ld cyclic subfield%s except for Q.\n", n_f,n_f>1?"s":"");
#endif

  lp = lg(vp); z = cgetg(lp, t_MAT);
  for (ip = 1; ip < lp; ip++)
  {
    pari_sp av = avma;
    long n_sub=0, n_chi=0;
    GEN gr=nullvec(), p = gel(vp, ip), zi;
    if (dvdsi(DegF, p)) pari_err(e_MISC,"p=%Pd divides [F:Q]=%ld\n", p, DegF);
    /* find conductor e of cyclic subfield K and set the subgroup HE of (Z/eZ)^*
     * corresponding to K */
    set_p_f(p, f, &start_pow, &max_pow);
    vData = const_vec(DegF, NULL);

    for (i=1; i<=n_f; i++) /* prescan. set Teichmuller */
    {
      GEN H1 = gel(vH1, i), H1data = gel(H1, 2), MinPol;
      long d_K = H1data[1], f_K = H1data[2], g_K = H1data[5];

      if (f_K == d_K+1 && equaliu(p, f_K)) /* found K=Q(zeta_p) */
      {
        pari_timer ti;
        GEN pnmax = powiu(p, max_pow), vNewton;
        MinPol = set_minpol_teich(g_K, p, max_pow);
        if (DEBUGLEVEL >= 6) timer_start(&ti);
        vNewton = FpX_Newton(MinPol, d_K+1, pnmax);
        if (DEBUGLEVEL >= 6)
          timer_printf(&ti, "FpX_Newton: teich: %ld %ld", degpol(MinPol), d_K);
        gel(vData, d_K) = mkvec2(MinPol, vNewton);
        break;
      }
    }

    for (i=1; i<=n_f; i++) /* handle all cyclic K */
    {
      GEN H1 = gel(vH1, i), H1data = gel(H1, 2), TR, MinPol, K, z1, C, Chi=NULL;
      long d_chi, n_conj, d_K = H1data[1], s_K = H1data[4];
      ulong pmodd = umodiu(p, d_K);

      if( (flag&SKIP_PROPER) && (DegF!=d_K) ) continue;

      if (!gel(vData, d_K))
      {
        pari_timer ti;
        GEN pnmax = powiu(p, max_pow), vNewton;
        MinPol = set_minpol(d_K, p, max_pow, flag);
        if (DEBUGLEVEL >= 6) timer_start(&ti);
        /*
        *  vNewton[2+i] = vNewton[2+i+d_K].
        *  We need vNewton[2+i] for 0 <= i <= d_K-1.
        *  But vNewton[2+d_K-1] may be 0 and will be deleted.
        *  So we need vNewton[2+d_K] not to delete vNewton[2+d_K-1].
        */
        vNewton = FpX_Newton(MinPol, d_K+1, pnmax);
        if (DEBUGLEVEL >= 6)
          timer_printf(&ti, "FpX_Newton: %ld %ld", degpol(MinPol), d_K);
        gel(vData, d_K) = mkvec2(MinPol, vNewton);
      }
      /* av = avma; */
      MinPol = gmael(vData, d_K, 1);
      TR = gmael(vData, d_K, 2) + 2; /* HACK: we need TR[0] = Tr(zeta^0) */

      d_chi = order_f_x(d_K, pmodd);
      n_conj = eulerphiu(d_K)/d_chi;
      n_sub++;n_chi += n_conj;
      C = set_C(pmodd, d_K, d_chi, n_conj);
      if (s_K==0) Chi = get_chi(H1);
      K = mkvecn(6, TR, MinPol, H1, C, Chi, mkvecsmall2(d_chi, n_conj));
#ifdef DEBUG
      handling(K);
#endif
      if (s_K && (flag&PLUS_PART))
        z1 = pclgp_cyc_real(K, p, max_pow, flag);
      else if ( s_K==0 && (flag&MINUS_PART))
        z1 = pclgp_cyc_imag(K, p, start_pow, max_pow, flag);
      else
        {z1 = nullvec();n_sub--;n_chi -= n_conj;}
      if(lg(z1)>1) gr = shallowconcat(gr, z1);
      /* avma = av; */
    }
    zi = mkcol(p);
    zi = vec_append(zi, (flag&PLUS_PART)?gather_part(gr, 0):nullvec());
    zi = vec_append(zi, (flag&MINUS_PART)?gather_part(gr, 1):nullvec());
    zi = shallowconcat(zi, mkcol3(cycGH, utoi(n_sub), utoi(n_chi)));
    gel(z, ip) = gerepilecopy(av, zi);
  }
  return typ(p0) == t_INT? shallowtrans(gel(z, 1)): shallowtrans(z);
}

static GEN
reduce_gcd(GEN x1, GEN x2)
{
  GEN d = gcdii(x1, x2);
  x1 = diviiexact(x1, d);
  x2 = diviiexact(x2, d);
  return mkvec2(x1, x2);
}

static GEN
ber_conj(GEN x, long k, long d)
{
  long i, l=lg(x), deg=l-3;GEN z;
  z = cgetg(d+2,t_POL);z[1] = evalsigne(1) | evalvarn(0);
  for(i=0;i<d;i++) gel(z, 2+i) = gen_0;
  for(i=0;i<=deg;i++) gel(z, 2+(i*k)%d) = gel(x, 2+i);
  return z;
}

/*
*  norm of x by g of order q
*  x^{1+g+g^2+...+g^(q-1)}
*/
static GEN
ber_norm_cyc(GEN x0, long g, long n, long d, long flag)
{
  pari_sp av = avma;
  long i, l, ei, gei, di, fi=0;GEN xi = x0;
  l = ulogint(n, 2);ei = (1L<<l);di = n/ei;
  for(i=1;i<=l;i++)
  {
    fi += (di&1)*ei;
    ei = 1L<<(l-i);gei = Fl_powu(g, ei, d);di=n/ei;
    xi = cyc_mul(xi, ber_conj(xi, gei, d), d);
    if (di&1) xi = cyc_mul(xi, ber_conj(x0, Fl_powu(g, fi, d), d), d);
  }
  return gerepilecopy(av, xi);
}

static GEN
ber_norm_by_cyc(GEN x0, long d, GEN MinPol, long flag)
{
  pari_sp av=avma;
  pari_timer ti;
  long i, r;GEN x=x0,  z, gen, gr;
  z = znstar0(utoi(d),0);
  gr = gel(z, 2);gen = gel(z, 3);r = lg(gr)-1;
  /* pari_printf("d=%ld r=%ld gen=%Ps gr=%Ps\n", d, r, gen, gr); */
  if (DEBUGLEVEL >= 6) timer_start(&ti);
  for(i=1;i<=r;i++)
    x = ber_norm_cyc(x, itou(gmael(gen, i, 2)), itou(gel(gr, i)), d, flag); 
  if (DEBUGLEVEL >= 6) timer_printf(&ti, "ber_norm_by_cyc [ber_norm_cyc]");
  if (DEBUGLEVEL >= 6) timer_start(&ti);
  x = ZX_rem(x, MinPol);  /* slow, why ? */
  if (DEBUGLEVEL >= 6) timer_printf(&ti, "ber_norm_by_cyc [ZX_rem]");
  if(lgpol(x)!=1) pari_err_BUG("rcnsubcyclo [norm of Bernoulli number]");
  return gerepilecopy(av, gel(x, 2));
}

static GEN
ber_norm_brute(GEN x0, long d, GEN MinPol)
{
  pari_sp av=avma;
  long k;GEN x=x0;
  for(k=2;k<d;k++)
    if (ugcd(k, d)==1)
      x=ZX_mod_Xnm1(ZX_mul(x, ber_conj(x0, k, d)), d);
  x = ZX_rem(x, MinPol);
  if(lgpol(x)!=1) pari_err_BUG("rcnsubcyclo [norm of Bernoulli number]");
  return gerepilecopy(av, gel(x, 2));
}

static GEN
ber_norm(GEN x0, long d, GEN MinPol, long flag)
{
  if (flag == 5) return ber_norm_brute(x0, d, MinPol);
  else return ber_norm_by_cyc(x0, d, MinPol, flag);
}

static GEN
ber_cyc1(GEN K)
{
  pari_sp av = avma;
  GEN H1 = gel(K, 1), Zeta = gel(K, 3), H1data = gel(H1, 2), Chi = gel(K, 2);
  long i, j, d = H1data[1], f = H1data[2], d2=d/2, f2 = f<<1;
  long *B = (long*)stack_malloc(sizeof(long)*d);
  GEN x;

  x = gen_1;
  for(i=1;i<=d2;i++)
  {
    pari_sp av = avma;
    GEN y = gen_0;
    if (ugcd(i, d)>1) continue;
    memset(B, 0, sizeof(long)*d);
    for(j=1;j<f;j++)
      if (Chi[j]>=0) B[(Chi[j]*i)%d] += j;
    for(j=0;j<d;j++)
      y = gadd(y, gmul(gel(Zeta, 1+j), utoi(B[j])));
    y = gerepilecopy(av, gdivgs(y, -f2));
    y = (d2>1)?cxnorm(y):absr(y);
    x = gmul(x, y);
  }
  return gerepilecopy(av, x);
}

static GEN
ber_cyc2(GEN K)
{
  pari_sp av = avma;
  GEN H1 = gel(K, 1), Zeta = gel(K, 3), H1data = gel(H1, 2), Chi = gel(K, 2);
  long i, j, d = H1data[1], f = H1data[2], d2 = d/2, f2 = f/2;
  GEN x;
  long *B = (long*)stack_malloc(sizeof(long)*d);

  x = gen_1;
  for(i=1;i<=d2;i++)
  {
    pari_sp av = avma;
    GEN y1 = gen_0, y2, y;
    if (ugcd(i, d)>1) continue;
    memset(B, 0, sizeof(long)*d);
    for(j=1;j<=f2;j++)
      if (Chi[j]>=0) B[(Chi[j]*i)%d] ++;
    for(j=0;j<d;j++)
      y1 = gadd(y1, gmul(gel(Zeta, 1+j), utoi(B[j])));
    y2=(Chi[2]>=0)?gmulgs(gsub(gel(Zeta, 1+(Chi[2]*i)%d), utoi(2)), 2):utoi(4);
    y=gerepilecopy(av, gdiv(y1, y2));
    y = (d2>1)?cxnorm(y):absr(y);
    x = gmul(x, y);
  }
  return gerepilecopy(av, x);
}

static GEN
ber_cyc3(GEN K)
{
  pari_sp av = avma;
  GEN H1 = gel(K, 1), MinPol = gel(K, 4), H1data = gel(H1, 2), Chi = gel(K, 2);
  long i, d = H1data[1], f = H1data[2], f2 = f<<1, dMinPol = degpol(MinPol);
  GEN x, x1, x2, B = const_vecsmall(d, 0);
  long *B1 = B+1;

  for(i=1;i<f;i++)
    if (Chi[i]>=0) B1[Chi[i]] += i;
  x1 = ZX_resultant(MinPol, ZX_rem(zv_to_ZX(B, 0), MinPol));
  x2 = powuu(f2, dMinPol);
  if (dMinPol&1) x2 = negi(x2);
  x = reduce_gcd(x1, x2);
  return gerepilecopy(av, x);
}

static GEN
ber_cyc4(GEN K)
{
  pari_sp av = avma;
  GEN H1 = gel(K, 1), MinPol = gel(K, 4), H1data = gel(H1, 2), Chi = gel(K, 2);
  long i, d = H1data[1], f = H1data[2], f2 = f/2;
  GEN x, x1, x2, y1, y2;
  GEN B1 = const_vecsmall(d, 0), B2 = const_vecsmall(d, 0);
  long *B11 = B1+1, *B22 = B2+1;

  for(i=1;i<=f2;i++)
    if (Chi[i]>=0) B11[Chi[i]] ++;
  y1=ZX_rem(zv_to_ZX(B1, 0), MinPol);
  x1 = ZX_resultant(MinPol, y1);
  B22[0] = 4;
  if (Chi[2]==0) B22[0] = 2;
  else if (Chi[2]>0) B22[d-Chi[2]]= -2;
  y2=ZX_rem(zv_to_ZX(B2, 0), MinPol);
  x2 = ZX_resultant(MinPol, y2);
  x = reduce_gcd(x1, x2);
  return gerepilecopy(av, x);
}

static GEN
ber_cyc5(GEN K, long flag)
{
  pari_sp av = avma;
  GEN H1 = gel(K, 1), MinPol = gel(K, 4), H1data = gel(H1, 2), Chi = gel(K, 2);
  long i, d = H1data[1], f = H1data[2], f2 = f/2, dMinPol = degpol(MinPol);
  GEN x, x1, x2, B = const_vecsmall(d, 0);
  long *B1 = B+1;

  for(i=1;i<=f2;i++)
    if (Chi[i]>=0) B1[Chi[i]] ++;

  if (dMinPol>100)
    x1 = ber_norm(zv_to_ZX(B, 0), d, MinPol, flag);
  else
    x1 = ZX_resultant(MinPol, ZX_rem(zv_to_ZX(B, 0), MinPol));

  if (Chi[2]<0)
    x2 = shifti(gen_1, 2*dMinPol);
  else if (Chi[2]==0)
    x2 = shifti(gen_1, dMinPol);
  else
  {
    long e = d/ugcd(Chi[2], d);
    x2 = powiu(poleval(polcyclo(e,0), gen_2), eulerphiu(d)/eulerphiu(e));
    x2 = shifti(x2, dMinPol);
  }
  x = reduce_gcd(x1, x2);
  return gerepilecopy(av, x);
}

static void
check_z2(GEN z2)
{
  if (dvdii(gel(z2, 1), gel(z2, 2))==0)
    pari_err_BUG("rcnsubcyclo [norm of Bernoulli number]");
  gel(z2, 1) = diviiexact(gel(z2, 1), gel(z2, 2));
}

/*
*  Hirabayashi-Yoshino, Manuscripta Math. vol.60, 423-436 (1988), Theorem 1
*
*  F is a subfield of Q(zeta_f)
*  f=p^a ==> Q=1
*  If F=Q(zeta_f), Q=1 <==> f=p^a
*  If f=4*p^a, p^a*q^b (p,q are odd primes), Q==2 <==> [Q(zeta_f):F] is odd
*/
static long
unit_index(long d, long f)
{
  ulong r, d_f;
  GEN fa = factoru(f), P = gel(fa, 1), E = gel(fa, 2);r = lg(P)-1;
  if(r==1) return 1;  /* f=P^a */
  d_f = eulerphiu_fact(fa);
  if(d==d_f) return 2;  /* F=Q(zeta_f) */
  if(r==2 && ((P[1]==2 && E[1]==2) || P[1]>2)) return ((d_f/d)&1)?2:1;
  err_printf("   *** We don't know the unit index Q. We assume Q=2.\n");
  return 2;
}

/*
*  flag = 1 ==> approx. a*chi(a), ber_cyc1()
*  flag = 2 ==> approx. chi(a), ber_cyc2()
*  flag = 3 ==> norm a*chi(a) by resultant, ber_cyc3()
*  flag = 4 ==> norm chi(a) by resultant, ber_cyc4()
*  flag = 5 ==> compute norm directly, ber_cyc5()
*  flag = 6 ==> compute norm by cyclic decomposition, ber_cyc5()
*/
static GEN
rel_class_num(long f, GEN HH, long flag, long prec)
{ 
  pari_sp av = avma;
  long OrderG, OrderH, DegF;
  long fH, i, s, n_f, h, gen, lgH, expG, W, Q;
  GEN gG, gH, gGH, cycGH, cycG, ncycGH, vChar, P, U, Ui, vH1, G, vData, Zeta;
  GEN x, z = gen_1, H0, MinPol;
  GEN z2 = mkvec2(gen_1, gen_1);

  if (flag<1 || flag>6) flag = 6;
  HH = znstar_generate(f, HH); /* group elements */
  fH = znstar_conductor(HH);
  if (fH == 1) return gen_1;
  else if (fH != f) { HH = znstar_reduce_modulus(HH, fH); f = fH; }

  OrderH = h = zv_prod(gel(HH, 2));
  OrderG = eulerphiu(f);
  DegF = OrderG / OrderH;
  s = srh_1(HH);
  if (DegF == 1 || s == 1) return gen_1;
  H0=subgr2vecsmall(subgp2ary(HH), OrderH, f);

  G = znstar0(utoipos(f), 1);
  gH = gel(HH, 1); lgH = lg(gH); /* gens. of H */
  P = cgetg(lgH, t_MAT);
  for (i = 1; i < lgH; i++) gel(P, i) = Zideallog(G, utoi(gH[i]));

  /* group structure of G/H */
  cycG = znstar_get_cyc(G);
  expG = itou(gel(cycG, 1));
  /* gG generators of G, gGH generators of G/H: gGH = g.Ui, g = gGH.U */
  cycGH = ZM_snf_group(hnfmodid(P, cycG), &U, &Ui);
  ncycGH = cyc_normalize(cycGH);
  gG = ZV_to_Flv(znstar_get_gen(G), f); /* gens. of G */
  /* generators of G/H */
  gGH = FlvV_factorback(gG, ZM_to_Flm(Ui, expG), f);
  vChar = chargalois(cycGH, NULL); n_f = lg(vChar)-2;
  vH1 = cgetg(n_f+1, t_VEC);
  for (i = 1; i <= n_f; i++)
  { /* skip trivial character */
    GEN chi = gel(vChar, i+1), nchi = char_normalize(chi, ncycGH);
    GEN chiG, E, H1, C = gel(nchi, 2);
    long e, he, d = itou(gel(nchi, 1));
    /* chi(prod g[i]^e[i]) = e(sum e[i]*C[i] / d), chi has order d = #(G/H1)*/
    E = Flv_extgcd(C, d); /* \sum C[i]*E[i] = 1 in Z/dZ */
    chiG = zncharlift(chi, ncycGH, U, cycG);
    H1 =  charker(cycG, chiG); /* H1 < G with G/H1 cyclic */
    e = itou( zncharconductor(G, chiG) ); /* cond H1 = cond chi */
    H1 = FlvV_factorback(gG, ZM_to_Flm(H1, expG), e);
    gen = Flv_factorback(gGH, E, e); /* G/H1 = <gen>, chi(gen) = e(1/d) */
    H1 = znstar_generate(e, H1);
    he = eulerphiu(e) / d;
    /* G/H1 = <gen> cyclic of index d, e = cond(H1) */
    gel(vH1, i) = mkvec3(H1, mkvecsmall5(d,e,he,srh_1(H1), gen), subgp2ary(H1));
  }

  if (flag>=1 && flag<=2 && prec==0)
  {
    double upper = 0.0;
    for (i=1; i<=n_f; i++)
    {
     GEN H1 = gel(vH1, i), H1data = gel(H1, 2);
     long j, d_K = H1data[1], f = H1data[2], s = H1data[4];
     if(s) continue;  /* F is real */
     for(j=1;j<d_K;j++)
       if(ugcd(j, d_K)==1) upper+=log10((double)(f-1)/4);
    }
    upper/=2;
    if ((prec = upper)<100) prec=100;
  }

  vData = const_vec(DegF, NULL);
  for (i=1; i<=n_f; i++)
  {
   GEN H1 = gel(vH1, i), H1data = gel(H1, 2), K, Chi;
   long d_K = H1data[1], s = H1data[4];

   if(s) continue;  /* F is real */
#ifdef DEBUG
   err_printf("  handling %s cyclic subfield K, deg(K)=%ld, cond(K)=%ld\n",
              s? "a real": "an imaginary", d_K, H1data[2]);
#endif
   if (!gel(vData, d_K))
   {
     GEN Zeta, MinPol;
     Zeta = (flag>=1 && flag<=2)? grootsof1(d_K, prec):NULL;
     MinPol = polcyclo(d_K,0);
     gel(vData, d_K) = mkvec2(Zeta, MinPol);
   }
   Zeta = gmael(vData, d_K, 1);
   MinPol = gmael(vData, d_K, 2);
   Chi = get_chi(H1);
   K = mkvec4(H1, Chi, Zeta, MinPol);

   if (flag==1)
     z = gmul(z, ber_cyc1(K));
   else if (flag==2)
     z = gmul(z, ber_cyc2(K));
   else if (flag==3)
     z = ber_cyc3(K);
   else if (flag==4)
     z = ber_cyc4(K);
   else if (flag>=5 && flag<=6)
     z = ber_cyc5(K, flag);
   if (flag>=3 && flag<=6)
   {
     gel(z2, 1) = mulii(gel(z2, 1), gel(z, 1));
     gel(z2, 2) = mulii(gel(z2, 2), gel(z, 2));
   }
  }
  Q = unit_index(DegF, f);
  W = root_of_1(H0, DegF, f);
  if (flag==1 || flag==2)
  {
    z = gmul(z, utoi(Q*W));
    x = absr(z);
  }
  else
  {
    gel(z2, 1) = mulii(gel(z2, 1), utoi(Q*W));
    check_z2(z2);
    x = gel(z2, 1);
  }
  return gerepilecopy(av, x);
}

static void
checkp(GEN p, const char *fun)
{
  if (!BPSW_psp(p)) pari_err_PRIME(fun, p);
  if (equaliu(p, 2)) pari_err_DOMAIN(fun,"p","=", gen_2, p);
}

static void
checkpf(GEN p, GEN f)
{
  if (cmpis(powiu(p, 2), 1L<<BITS_IN_HALFULONG)>=0)
    pari_err(e_MISC,"p=%Ps is too large\n", p);
  if ( cmpis(p, 2)==0)
    pari_err(e_MISC,"p=2 is not allowed\n");
  if ( !uisprime(itou(p)))
    pari_err(e_MISC,"p=%Ps is not prime\n", p);
  if (cmpis(f, 1L<<BITS_IN_HALFULONG)>=0)
    pari_err(e_MISC,"f=%Ps is too large conductor\n", f);
}

static long
check_quad_pmn(long p, long m, long n)
{
  if (n<=0)
    pari_err(e_MISC,"n=%ld is not positive.\n", n);
  if (p<=0)
    pari_err(e_MISC,"p=%ld is not positive.\n", p);
  if (p > (1L<<(BITS_IN_HALFULONG/2)))
    pari_err(e_MISC,"p=%ld is too large.\n", p);
  if (!uisprime(p))
    pari_err(e_MISC,"p=%ld is not prime.\n", p);
  if(m>=0)
    pari_err(e_MISC,"m=%ld is not negative.\n", m);
  if (labs(m) > (1L<<BITS_IN_HALFULONG))
    pari_err(e_MISC,"m=%ld is too large.\n", m);
  if(uissquarefree(-m)) return m;
  if(isfundamental(stoi(m))==0)
    pari_err(e_MISC,"m=%ld is neither square free nor fundamental discriminant.\n",m);
  return m/4;
}

GEN
pclgpsubcyclo(GEN p, GEN F, GEN H, long flag)
{
  pari_sp av = avma;
  long i, l, f;
  const char *fun = "pclgpsubcyclo";
  long flag0 = flag & (USE_BASIS | USE_FACTOR | USE_GALOIS_POL | USE_F);
  if( flag0 &&
    (flag0!=USE_BASIS) && (flag0!=USE_FACTOR) &&
    (flag0!=USE_GALOIS_POL) && (flag0!=USE_F))
    pari_err(e_MISC,"flag %ld,%ld,%ld,%ld can not be given at the same time\n",
    USE_BASIS,USE_FACTOR,USE_GALOIS_POL,USE_F);
  if (typ(F) != t_INT) pari_err_TYPE(fun, F);
  if (typ(p) == t_VEC)
  {
    l = lg(p);
    for (i = 1; i < l; i++) checkp(gel(p, i), fun);
  }
  else
    checkp(p, fun);
  f = itos(F);
  if (!H) H = cgetg(1, t_VECSMALL);
  else
  {
    if (typ(H) != t_VECSMALL) H = gtovecsmall(H);
    l = lg(H);
    for (i = 1; i < l; i++)
    {
      H[i] = smodss(H[i], f);
      if (cgcd(f, H[i]) > 1) pari_err_COPRIME("pclgpsubcyclo", F, stoi(H[i]));
    }
  }
  return gerepilecopy(av, pclgp(p, f, H, flag));
}

GEN
rcnsubcyclo(GEN F, GEN H, long flag, long prec)
{
  pari_sp av = avma;
  long i, l, f;
  const char *fun = "rcnsubcyclo";
  if (typ(F) != t_INT) pari_err_TYPE(fun, F);
  f = itos(F);
  if (!H) H = cgetg(1, t_VECSMALL);
  else
  {
    if (typ(H) != t_VECSMALL) H = gtovecsmall(H);
    l = lg(H);
    for (i = 1; i < l; i++)
    {
      H[i] = smodss(H[i], f);
      if (cgcd(f, H[i]) > 1) pari_err_COPRIME("rcnsubcyclo", F, stoi(H[i]));
    }
  }
  return gerepilecopy(av, rel_class_num(f, H, flag, prec));
}

GEN
lambdasubcyclo(GEN p, GEN F, GEN H)
{
  pari_sp av = avma;
  long i, l, f;
  const char *fun = "lambdasubcyclo";
  if (typ(F) != t_INT) pari_err_TYPE(fun, F);
  checkpf(p, F);
  f = itos(F);
  if (!H) H = cgetg(1, t_VECSMALL);
  else
  {
    if (typ(H) != t_VECSMALL) H = gtovecsmall(H);
    l = lg(H);
    for (i = 1; i < l; i++)
      if (cgcd(f, H[i]) > 1) pari_err_COPRIME("lambdasubcyclo", F, stoi(H[i]));
  }
  return gerepilecopy(av, abellambda(itos(p), f, H));
}

GEN
ipolsubcyclo(GEN p, GEN F, GEN H, GEN n)
{
  pari_sp av = avma;
  long i, l, f;
  const char *fun = "ipolsubcyclo";
  if (typ(p) != t_INT) pari_err_TYPE(fun, p);
  if (typ(F) != t_INT) pari_err_TYPE(fun, F);
  if (typ(n) != t_INT) pari_err_TYPE(fun, n);
  checkpf(p, F);
  f = itos(F);
  if (!H) H = cgetg(1, t_VECSMALL);
  else
  {
    if (typ(H) != t_VECSMALL) H = gtovecsmall(H);
    l = lg(H);
    for (i = 1; i < l; i++)
      if (cgcd(f, H[i]) > 1) pari_err_COPRIME("ipolsubcyclo", F, stoi(H[i]));
  }
  return gerepilecopy(av, abelstkpol(itos(p), f, H, itos(n)));
}

GEN
quadstkpol(long p, long m, long n, long flag)
{
  pari_sp av = avma;
  m = check_quad_pmn(p, m, n);
  return gerepilecopy(av, quadstkpol0(p, m, n, flag));
}

GEN
quadlambda(long p, long m)
{
  pari_sp av = avma;
  m = check_quad_pmn(p, m, 1);
  return gerepilecopy(av, quadlambda0(p, m));
}

static long
F2v_order(GEN H)
{
  long i, x, m, n=0, l=lg(H);
  for(i=2;i<l;i++)
  {
    for(m=0, x=H[i];x;x<<=1) if(x<0) m++;
    n+=m;
  }
  return n;
}

/*
*  y in H <==> H[y]==1
*  return n s.t. x^n in H
*  if H==0, then return 0
*/
static long
order_H_x(GEN H, long x, GEN div_H)
{
  long i, y=x, f=H[1], h = lg(div_H);
  y = Fl_powu(x, div_H[1], f);
  if(F2v_coeff(H, y)) return div_H[1];
  for(i=2;i<h;i++)
  {
    y = Fl_mul(y, Fl_powu(x, div_H[i]-div_H[i-1], f), f);
    if(F2v_coeff(H, y)) return div_H[i];
  }
  pari_err_BUG("ele_to_gen:[order_H_x]"); 
  return 0;
}

/* find y=xh (h\in H) s.t. y^o=1 */
static long
adjust_H(GEN H, long x, long o, long flag)
{
  long y, h, f=H[1];
  if (flag==0) return x; 
  y = Fl_inv(Fl_powu(x, o, f), f);
  for(h=1;h<f;h++)
    if(F2v_coeff(H, h) && y==Fl_powu(h, o, f)) return Fl_mul(x, h, f);
  pari_err_BUG("ele2gen [adjust_H]");
  return 1;
}

static long
max_order_ele(GEN H1, GEN H2, GEN div_H, long flag)
{
  long i, x, n=0, max=0, max_i=0, f=H2[1];
  n = F2v_order(H2);
  for(i=1;i<f;i++)
    {
     if (F2v_coeff(H2, i)==0) continue;
     else if ((x=order_H_x(H1, i, div_H))==n) return adjust_H(H1, i, n, flag);
     else if (x>max) {max=x;max_i=i;}
    }
  return adjust_H(H1, max_i, max, flag);
}

static void
enlarge_H(GEN H, long x, GEN div_H)
{
  pari_sp av = avma;
  long i, j, y=x, o, l=lg(H), f=H[1];GEN H1;
  if((o=order_H_x(H, x, div_H))==1) return;
  H1 = vecsmall_copy(H);
  for(i=1;i<f;i++)
     {
       if (F2v_coeff(H, i)==0) continue;
       for(y=x, j=1;j<o;j++)
       {
         F2v_set(H1, Fl_mul(i, y, f));
         y = Fl_mul(y, x, f);
       }
     }
  for(i=2;i<l;i++) H[i] = H1[i];
  avma = av;
}

/*
*       x=0011
*       y=0101
*     x^y=0110
* x&(x^y)=0010
*
*  H2 = H2 - H1
*/
static void
delete_H(GEN H2, GEN H1)
{
  long i, l=lg(H1);
  for(i=2;i<l;i++) H2[i] &= (H1[i]^H2[i]);
}

/*
*  receive subgr t_VECSMALL H of (Z/fZ)^* and return generators
*  1<= a <= f, a\in H <==> F2v_coeff(H, a)=1
*/
static GEN
ele_to_gen(GEN H, long flag)
{
  pari_sp av = avma;
  long f = H[1], g;GEN H1, H2, z = const_vecsmall(0,0), div_H;
  div_H = divisorsu(F2v_order(H));
  H1 = zero_F2v(f);F2v_set(H1, 1);
  H2 = vecsmall_copy(H);
  while((g=max_order_ele(H1, H2, div_H, flag)))
  {
    z = vecsmall_append(z, g);
    enlarge_H(H1, g, div_H);
    delete_H(H2, H1);
  }
  return gerepilecopy(av, mkvec2(utoi(f), vecsmall_to_vec(z)));
}

GEN
ele2gen(GEN H, long flag)
{
  GEN HH = NULL; /* to suppress warning */
  if(typ(H)==t_VEC) HH = ZV_to_F2v(H);
  else if(typ(H)==t_VECSMALL) HH = Flv_to_F2v(H);
  else pari_err_TYPE("ele2gen", H);
  return ele_to_gen(HH, flag);
}
